/* 

   libj - cmd.c                  Copyright (C) 2001-1998 Justin David Smith
   justins@chaos2.org            http://chaos2.org/
    
   Simplistic command library


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjcmd.h>
#include <libjstr.h>
#include <libjfile.h> 


cmd *cmd_search(cmd *c, const char *name) {

   while(c->name) {
      if(strequal(c->name, name)) return(c);
      c++;
   }
   return(NULL);

}


char *_cmd_scan_unquoted_character(char *buf, char sym) {

   char *p;
   int backslash;
   int inquotes;
   
   p = buf;
   backslash= 0;
   inquotes = 0;
   while(*p) {
      if(*p == '\"' && !backslash) inquotes = !inquotes;
      backslash = (!backslash && *p == '\\');
      if(!inquotes && *p == sym) return(p);
      p++;
   }
   return(NULL);

}


void _cmd_remove_quotes(char *buf) {

   char *p;
   int backslash;
   int inquotes;
   
   p = buf;
   backslash= 0;
   inquotes = 0;
   while(*p) {
      if(*p == '\"' && !backslash) {
         inquotes = !inquotes;
         strcopy(p, p + 1);
      } else {
         backslash = (!backslash && *p == '\\');
         p++;
      }
   }
   return;

}


int _cmd_parse_line(cmd *c, file *f, char *line, void *data) {

   char *pequals;
   char *pcomm;
   char *parg;

   /* Remove end-of-line comments */
   pcomm   = _cmd_scan_unquoted_character(line, ';');
   if(pcomm) *pcomm = 0;

   /* Scan for equals sign */
   pequals = _cmd_scan_unquoted_character(line, '=');
   if(!pequals) {
      /* No equals sign on this line */
      /*fprintf(stderr, "%s(%d) (warning):  Line is not an assignment statement\n", file_name(f), file_line(f) - 1);*/
      parg = line + strlenn(line);
   } else {
      *pequals = 0;
      parg = pequals + 1;
   }
   
   /* Parse command name */
   trim(line);
   _cmd_remove_quotes(line);
   unescape(line);
   if(!*line) return(0);
   
   /* Parse command argument */
   trim(parg);
   _cmd_remove_quotes(parg);
   unescape(parg);
   
   /* Search for command name */
   c = cmd_search(c, line);
   if(!c) {
      fprintf(stderr, "%s(%d) (warning):  Command not found \"%s\"\n", file_name(f), file_line(f) - 1, line);
      return(0);
   }
   
   /* Run command function */
   if(c->func && -1 == c->func(f, c->name, parg, c->code, c->data, data)) {
      fprintf(stderr, "%s(%d):  Function \"%s\" returned error status\n", file_name(f), file_line(f) - 1, c->name);
      return(-1);
   }
   
   return(0);
   
}



int cmd_parse_file(cmd *c, const char *filename, void *data) {

   char buf[READBUF];
   int len;
   file *f;
   
   /* Open the input file */
   if(!c) return(-1);
   if(!filename) return(-1);
   if(!(f = file_create(filename, IO_READONLY, READBUF))) return(-1);
   
   /* Read lines from input file */
   while(-1 != file_read_line(f, buf, READBUF)) {
   
      len = strlenn(buf);
      while(len + 1 < READBUF && *(buf + len - 1) == '\\') {
         if(-1 == file_read_line(f, buf + len - 1, READBUF - len + 1)) {
            fprintf(stderr, "%s(%d):  File ends after line continuation\n", file_name(f), file_line(f) - 1);
            return(-1);
         }
         len = strlenn(buf);
      }
      _cmd_parse_line(c, f, buf, data);
      
   }
   
   file_close(&f);
   return(0);
      
}


