/* 

   libj - conio_in.c             Copyright (C) 2001-1998 Justin David Smith    
   justins@chaos2.org            http://chaos2.org/
    
   Console input


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjconio.h>
#include <sys/ioctl.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <termcap.h>
#include <termios.h>
#include <unistd.h>


/* Time value manipulation */

#define  IO_TV_SET(t, s, u)         ((t).tv_sec = (s), (t).tv_usec = (u))
#define  IO_TV_CLEAR(t)             IO_TV_SET((t), 0, 0)
#define  IO_TV_IS_ZERO(t)           ((t).tv_sec == 0 && (t).tv_usec == 0)
#define  IO_TV_ADD(t, x)            (t).tv_usec += (x).tv_usec; \
                                    (t).tv_sec += (x).tv_sec; \
                                    if((t).tv_usec > 999999) { \
                                       (t).tv_sec++; \
                                       (t).tv_usec -= 1000000; \
                                    }
#define  IO_TV_SUB(t, x)            (t).tv_usec -= (x).tv_usec; \
                                    (t).tv_sec -= (x).tv_sec; \
                                    if((t).tv_usec < 0) { \
                                       (t).tv_sec--; \
                                       (t).tv_usec += 1000000; \
                                    }
#define  IO_TV_GREATER_THAN(a, b)   ((a).tv_sec > (b).tv_sec || \
                                    ((a).tv_sec == (b).tv_sec && (a).tv_usec > (b).tv_sec))
#define  IO_TV_EQUAL(a, b)          ((a).tv_sec == (b).tv_sec && \
                                    (a).tv_usec == (b).tv_usec)
#define  IO_TV_GREATER_EQUAL(a, b)  (IO_TV_GREATER_THAN((a), (b)) || IO_TV_EQUAL((a), (b)))
#define  IO_TV_LESS_THAN(a, b)      !IO_TV_GREATER_EQUAL((a), (b))
#define  IO_TV_LESS_EQUAL(a, b)     !IO_TV_GREATER_THAN((a), (b))
#define  IO_TV_UNEQUAL(a, b)        !IO_TV_EQUAL((a), (b))


/* Input */


inline int _io_char_avail(struct timeval *delay) {
/* internal-io-character-available
   
   Waits for a specified period of time for a character input from
   standard input.  If a character is available, then delay is adjusted
   to indicate the amount of time remaining, and 1 returned.  If no
   character is available in the time specified, 0 is returned and
   the delay structure is cleared.  
   
   If delay is 0, then this function returns 1 if a character is
   immediately available. If delay is NULL, a delay of zero is
   assumed. */
   
   struct timeval tstart;
   struct timeval tend;
   struct timeval *d;
   int handle;
   int sel;
   
   /* Call select to see status of stdin read handle */
   if(!delay) {
      IO_TV_CLEAR(tend);
      d = &tend;
   } else d = delay;
   handle = 1;
   gettimeofday(&tstart, NULL);
   sel = select(1, (fd_set *)&handle, (fd_set *)0, (fd_set *)0, d);
   gettimeofday(&tend, NULL);

   switch(sel) {
   case -1:
      /* Select error */
      IO_TV_CLEAR(*d);
      return(0);
      break;
   case 0:
      /* Timed out */
      IO_TV_CLEAR(*d);
      return(0);
   }
   
   /* Character(s) available */
   if(delay) {
      IO_TV_SUB(tend, tstart);
      if(IO_TV_GREATER_THAN(tend, *delay)) IO_TV_CLEAR(*delay);
      else IO_TV_SUB(*delay, tend);
   }
   return(1);
      
}


int io_char_avail() {
/* io-character-available */

   return(_io_char_avail(NULL));   

}


inline int _io_read_char(struct timeval *delay) {
/* internal-io-read-character */

   char ch;

   if(!_io_char_avail(delay)) return(-1);
   if(read(0, &ch, 1) != 1) return(-1);
   return((int)ch);

}


int io_wait_char(struct timeval *delay) {
/* io-wait-for-character 

   Waits the specified period for a character, and returns the character
   if one is found.  Otherwise, -1 is returned if the timeout expires.
   delay is modified as in _io_char_avail() */

   return(_io_read_char(delay));
   
}


int io_read_char() {
/* io-read-character 

   Like above, except returns immediately (delay = 0). */

   return(_io_read_char(NULL));

}


int _io_read_string(char *dest, unsigned int size, struct timeval *delay) {
/* internal-io-read-string */

   int ch;
   int count;
   
   if(!dest || !size) return(0);
   count = 0;
   size--;
   while(size) {
      ch = _io_read_char(delay);
      if(ch == '\n' || ch == '\r' || ch < 0) {
         *dest = 0;
         return(count);
      }
      *dest = (char)ch;
      count++;
      dest++;
      size--;
   }
   *dest = 0;
   return(count);
   
}


int io_read_string(char *dest, unsigned int size) {
/* io-read-string */

   return(_io_read_string(dest, size, NULL));
   
}


int io_wait_string(char *dest, unsigned int size, struct timeval *delay) {
/* io-wait-for-string */

   return(_io_read_string(dest, size, delay));
   
}


int io_eat() {
/* io-eat */

   while(_io_read_char(NULL) >= 0);
   return(0);
   
}


int io_idle(struct timeval *delay) {
/* io-idle */

   while(!IO_TV_IS_ZERO(*delay) && !_io_char_avail(delay));
   return(0);

}


int io_sleep(struct timeval *delay) {
/* io-sleep */

   while(!IO_TV_IS_ZERO(*delay)) _io_read_char(delay);
   return(0);

}


long _io_read_key(struct timeval *delay) {
/* internal-Io-Read-Key */

   struct timeval t;
   long key;
   char ch;
   
   IO_TV_SET(t, 0, 50000);
   ch = 0;
   
   ch = _io_read_char(delay);
   if(ch < 0) return(-1);
   key = ch;
   
   if(ch != 0x1b) return(key);

   /* Multi-character sequence -- second character.
      Could also be ESC == 0x1b only. */
   ch = _io_read_char(&t);
   if(ch < 0) return(key);
   key = (key << 8) | ch;
   if(ch != 0x5b) return(key);
   
   /* Third character -- prefix is 0x1b5b.
      0x5b is always a continuation for a 4-char sequence.
      0x3? continuation for 4- or 5-char sequence ending in 0x7e. */
   ch = _io_read_char(&t);
   if(ch < 0) {
      fprintf(stderr, "Warning:  Premature end of key sequence?  Saw %lx\n", key);
      return(key);
   }
   key = (key << 8) | ch;
   if(ch != 0x5b && (ch & 0xf0) != 0x30) return(key);

   /* Fourth character, 0x7e terminates */
   ch = _io_read_char(&t);
   if(ch < 0) {
      fprintf(stderr, "Warning:  Premature end of key sequence?  Saw %lx\n", key);
      return(key);
   }
   key = (key << 8) | ch;
   if(ch == 0x7e) return(key);
   if((key & 0xff00) == 0x5b00) return(key);
   
   /* Fifth and (hopefully) final character */
   ch = _io_read_char(&t);
   if(ch < 0) {
      fprintf(stderr, "Warning:  Premature end of key sequence?  Saw %lx\n", key);
      return(key);
   }
   if(ch != 0x7e) {
      fprintf(stderr, "Warning:  Key sequence is too long, %lx%02x\n", key, ch);
      return(key);
   }
   return(key);
   
}


long io_wait_key(struct timeval *delay) {
/* io-Wait-Key */

   return(_io_read_key(delay));
   
}


long io_read_key() {

   return(_io_read_key(NULL));

}


typedef struct _keydata {
   long key;
   char *name;
   long basekey;
   long modifier;
} keydata;

#define  __A   IO_MOD_ALT
#define  __S   IO_MOD_SHIFT
#define  __C   IO_MOD_CTRL

static keydata keys[] = 
{  {  IO_KEY_ESC,       "Escape",         IO_KEY_ESC,    0  },
   {  IO_KEY_TAB,       "Tab",            IO_KEY_TAB,    0  },
   {  IO_KEY_BKSP,      "Backspace",      IO_KEY_BKSP,   0  },
   {  IO_KEY_RETURN,    "Return",         IO_KEY_RETURN, 0  },
   {  IO_KEY_PAUSE,     "Pause",          IO_KEY_PAUSE,  0  },
   {  IO_KEY_SPACE,     "Space",          IO_KEY_SPACE,  0  },
   {  IO_KEY_ESC_A,     "Alt-Escape",     IO_KEY_ESC,    __A},
   {  IO_KEY_TAB_A,     "Alt-Tab",        IO_KEY_TAB,    __A},
   {  IO_KEY_BKSP_A,    "Alt-Backspace",  IO_KEY_BKSP,   __A},
   {  IO_KEY_RETURN_A,  "Alt-Return",     IO_KEY_RETURN, __A},
   {  IO_KEY_SPACE_A,   "Alt-Space",      IO_KEY_SPACE,  __A},
   {  IO_KEY_BKSP_C,    "Ctrl-Backspace", IO_KEY_BKSP,   __C},
   {  IO_KEY_F1,        "F1",             IO_KEY_F1,     0  },
   {  IO_KEY_F2,        "F2",             IO_KEY_F2,     0  },
   {  IO_KEY_F3,        "F3",             IO_KEY_F3,     0  },
   {  IO_KEY_F4,        "F4",             IO_KEY_F4,     0  },
   {  IO_KEY_F5,        "F5",             IO_KEY_F5,     0  },
   {  IO_KEY_F6,        "F6",             IO_KEY_F6,     0  },
   {  IO_KEY_F7,        "F7",             IO_KEY_F7,     0  },
   {  IO_KEY_F8,        "F8",             IO_KEY_F8,     0  },
   {  IO_KEY_F9,        "F9",             IO_KEY_F9,     0  },
   {  IO_KEY_F10,       "F10",            IO_KEY_F10,    0  },
   {  IO_KEY_F11,       "F11",            IO_KEY_F11,    0  },
   {  IO_KEY_F12,       "F12",            IO_KEY_F12,    0  },
   {  IO_KEY_F1_S,      "Shift-F1",       IO_KEY_F1,     __S},
   {  IO_KEY_F2_S,      "Shift-F2",       IO_KEY_F2,     __S},
   {  IO_KEY_F3_S,      "Shift-F3",       IO_KEY_F3,     __S},
   {  IO_KEY_F4_S,      "Shift-F4",       IO_KEY_F4,     __S},
   {  IO_KEY_F5_S,      "Shift-F5",       IO_KEY_F5,     __S},
   {  IO_KEY_F6_S,      "Shift-F6",       IO_KEY_F6,     __S},
   {  IO_KEY_F7_S,      "Shift-F7",       IO_KEY_F7,     __S},
   {  IO_KEY_F8_S,      "Shift-F8",       IO_KEY_F8,     __S},
   {  IO_KEY_F9_S,      "Shift-F9",       IO_KEY_F9,     __S},
   {  IO_KEY_F10_S,     "Shift-F10",      IO_KEY_F10,    __S},
   {  IO_KEY_F11_S,     "Shift-F11",      IO_KEY_F11,    __S},
   {  IO_KEY_F12_S,     "Shift-F12",      IO_KEY_F12,    __S},
   {  IO_KEY_F1_C,      "Ctrl-F1",        IO_KEY_F1,     __C},
   {  IO_KEY_F2_C,      "Ctrl-F2",        IO_KEY_F2,     __C},
   {  IO_KEY_F3_C,      "Ctrl-F3",        IO_KEY_F3,     __C},
   {  IO_KEY_F4_C,      "Ctrl-F4",        IO_KEY_F4,     __C},
   {  IO_KEY_F5_C,      "Ctrl-F5",        IO_KEY_F5,     __C},
   {  IO_KEY_F6_C,      "Ctrl-F6",        IO_KEY_F6,     __C},
   {  IO_KEY_F7_C,      "Ctrl-F7",        IO_KEY_F7,     __C},
   {  IO_KEY_F8_C,      "Ctrl-F8",        IO_KEY_F8,     __C},
   {  IO_KEY_F9_C,      "Ctrl-F9",        IO_KEY_F9,     __C},
   {  IO_KEY_F10_C,     "Ctrl-F10",       IO_KEY_F10,    __C},
   {  IO_KEY_F11_C,     "Ctrl-F11",       IO_KEY_F11,    __C},
   {  IO_KEY_F12_C,     "Ctrl-F12",       IO_KEY_F12,    __C},
   {  IO_KEY_HOME,      "Home",           IO_KEY_HOME,   0  },
   {  IO_KEY_INS,       "Insert",         IO_KEY_INS,    0  },
   {  IO_KEY_DEL,       "Delete",         IO_KEY_DEL,    0  },
   {  IO_KEY_END,       "End",            IO_KEY_END,    0  },
   {  IO_KEY_PGUP,      "Page Up",        IO_KEY_PGUP,   0  },
   {  IO_KEY_PGDN,      "Page Down",      IO_KEY_PGDN,   0  },
   {  IO_KEY_UP,        "Up Arrow",       IO_KEY_UP,     0  },
   {  IO_KEY_DOWN,      "Down Arrow",     IO_KEY_DOWN,   0  },
   {  IO_KEY_RIGHT,     "Right Arrow",    IO_KEY_RIGHT,  0  },
   {  IO_KEY_LEFT,      "Left Arrow",     IO_KEY_LEFT,   0  },
   {  IO_KEY_CENTER,    "Center",         IO_KEY_CENTER, 0  },
   {  IO_KEY_A_C,       "Ctrl-a",         'a',           __C},
   {  IO_KEY_B_C,       "Ctrl-b",         'b',           __C},
   {  IO_KEY_C_C,       "Ctrl-c",         'c',           __C},
   {  IO_KEY_D_C,       "Ctrl-d",         'd',           __C},
   {  IO_KEY_E_C,       "Ctrl-e",         'e',           __C},
   {  IO_KEY_F_C,       "Ctrl-f",         'f',           __C},
   {  IO_KEY_G_C,       "Ctrl-g",         'g',           __C},
   {  IO_KEY_H_C,       "Ctrl-h",         'h',           __C},
   {  IO_KEY_I_C,       "Ctrl-i",         'i',           __C},
   {  IO_KEY_J_C,       "Ctrl-j",         'j',           __C},
   {  IO_KEY_K_C,       "Ctrl-k",         'k',           __C},
   {  IO_KEY_L_C,       "Ctrl-l",         'l',           __C},
   {  IO_KEY_M_C,       "Ctrl-m",         'm',           __C},
   {  IO_KEY_N_C,       "Ctrl-n",         'n',           __C},
   {  IO_KEY_O_C,       "Ctrl-o",         'o',           __C},
   {  IO_KEY_P_C,       "Ctrl-p",         'p',           __C},
   {  IO_KEY_Q_C,       "Ctrl-q",         'q',           __C},
   {  IO_KEY_R_C,       "Ctrl-r",         'r',           __C},
   {  IO_KEY_S_C,       "Ctrl-s",         's',           __C},
   {  IO_KEY_T_C,       "Ctrl-t",         't',           __C},
   {  IO_KEY_U_C,       "Ctrl-u",         'u',           __C},
   {  IO_KEY_V_C,       "Ctrl-v",         'v',           __C},
   {  IO_KEY_W_C,       "Ctrl-w",         'w',           __C},
   {  IO_KEY_X_C,       "Ctrl-x",         'x',           __C},
   {  IO_KEY_Y_C,       "Ctrl-y",         'y',           __C},
   {  IO_KEY_Z_C,       "Ctrl-z",         'z',           __C},
   {  IO_KEY_3_C,       "Ctrl-3",         '3',           __C},
   {  IO_KEY_4_C,       "Ctrl-4",         '4',           __C},
   {  IO_KEY_5_C,       "Ctrl-5",         '5',           __C},
   {  IO_KEY_6_C,       "Ctrl-6",         '6',           __C},
   {  IO_KEY_7_C,       "Ctrl-7",         '7',           __C},
   {  IO_KEY_TILDE_A,   "Alt-Backquote",  '`',           __A},
   {  IO_KEY_MINUS_A,   "Alt-Minus",      '-',           __A},
   {  IO_KEY_PLUS_A,    "Alt-Plus",       '+',           __A},
   {  IO_KEY_OBRAC_A,   "Alt-Openbracket",'[',           __A},
   {  IO_KEY_CBRAC_A,   "Alt-Closebracket",']',          __A},
   {  IO_KEY_COLON_A,   "Alt-Semicolon",  ';',           __A},
   {  IO_KEY_QUOTE_A,   "Alt-Quote",      '\'',          __A},
   {  IO_KEY_COMMA_A,   "Alt-Comma",      ',',           __A},
   {  IO_KEY_DOT_A,     "Alt-Period",     '.',           __A},
   {  IO_KEY_SLASH_A,   "Alt-Forwardslash",'/',          __A},
   {  IO_KEY_PIPE_A,    "Alt-Backslash",  '\\',          __A},
   {  IO_KEY_STAR_A,    "Alt-Star",       '*',           __A},
   {  IO_KEY_0_A,       "Alt-0",          '0',           __A},
   {  IO_KEY_1_A,       "Alt-1",          '1',           __A},
   {  IO_KEY_2_A,       "Alt-2",          '2',           __A},
   {  IO_KEY_3_A,       "Alt-3",          '3',           __A},
   {  IO_KEY_4_A,       "Alt-4",          '4',           __A},
   {  IO_KEY_5_A,       "Alt-5",          '5',           __A},
   {  IO_KEY_6_A,       "Alt-6",          '6',           __A},
   {  IO_KEY_7_A,       "Alt-7",          '7',           __A},
   {  IO_KEY_8_A,       "Alt-8",          '8',           __A},
   {  IO_KEY_9_A,       "Alt-9",          '9',           __A},
   {  IO_KEY_A_A,       "Alt-a",          'a',           __A},
   {  IO_KEY_B_A,       "Alt-b",          'b',           __A},
   {  IO_KEY_C_A,       "Alt-c",          'c',           __A},
   {  IO_KEY_D_A,       "Alt-d",          'd',           __A},
   {  IO_KEY_E_A,       "Alt-e",          'e',           __A},
   {  IO_KEY_F_A,       "Alt-f",          'f',           __A},
   {  IO_KEY_G_A,       "Alt-g",          'g',           __A},
   {  IO_KEY_H_A,       "Alt-h",          'h',           __A},
   {  IO_KEY_I_A,       "Alt-i",          'i',           __A},
   {  IO_KEY_J_A,       "Alt-j",          'j',           __A},
   {  IO_KEY_K_A,       "Alt-k",          'k',           __A},
   {  IO_KEY_L_A,       "Alt-l",          'l',           __A},
   {  IO_KEY_M_A,       "Alt-m",          'm',           __A},
   {  IO_KEY_N_A,       "Alt-n",          'n',           __A},
   {  IO_KEY_O_A,       "Alt-o",          'o',           __A},
   {  IO_KEY_P_A,       "Alt-p",          'p',           __A},
   {  IO_KEY_Q_A,       "Alt-q",          'q',           __A},
   {  IO_KEY_R_A,       "Alt-r",          'r',           __A},
   {  IO_KEY_S_A,       "Alt-s",          's',           __A},
   {  IO_KEY_T_A,       "Alt-t",          't',           __A},
   {  IO_KEY_U_A,       "Alt-u",          'u',           __A},
   {  IO_KEY_V_A,       "Alt-v",          'v',           __A},
   {  IO_KEY_W_A,       "Alt-w",          'w',           __A},
   {  IO_KEY_X_A,       "Alt-x",          'x',           __A},
   {  IO_KEY_Y_A,       "Alt-y",          'y',           __A},
   {  IO_KEY_Z_A,       "Alt-z",          'z',           __A},
   {  '`',              "Backquote",      '`',           0  },
   {  '1',              "1",              '1',           0  },
   {  '2',              "2",              '2',           0  },
   {  '3',              "3",              '3',           0  },
   {  '4',              "4",              '4',           0  },
   {  '5',              "5",              '5',           0  },
   {  '6',              "6",              '6',           0  },
   {  '7',              "7",              '7',           0  },
   {  '8',              "8",              '8',           0  },
   {  '9',              "9",              '9',           0  },
   {  '0',              "0",              '0',           0  },
   {  '-',              "Minus",          '-',           0  },
   {  '=',              "Equals",         '=',           0  },
   {  'q',              "q",              'q',           0  },
   {  'w',              "w",              'w',           0  },
   {  'e',              "e",              'e',           0  },
   {  'r',              "r",              'r',           0  },
   {  't',              "t",              't',           0  },
   {  'y',              "y",              'y',           0  },
   {  'u',              "u",              'u',           0  },
   {  'i',              "i",              'i',           0  },
   {  'o',              "o",              'o',           0  },
   {  'p',              "p",              'p',           0  },
   {  '[',              "Open Bracket",   '[',           0  },
   {  ']',              "Close Bracket",  ']',           0  },
   {  'a',              "a",              'a',           0  },
   {  's',              "s",              's',           0  },
   {  'd',              "d",              'd',           0  },
   {  'f',              "f",              'f',           0  },
   {  'g',              "g",              'g',           0  },
   {  'h',              "h",              'h',           0  },
   {  'j',              "j",              'j',           0  },
   {  'k',              "k",              'k',           0  },
   {  'l',              "l",              'l',           0  },
   {  ';',              "Semicolon",      ';',           0  },
   {  '\'',             "Quote",          '\'',          0  },
   {  'z',              "z",              'z',           0  },
   {  'x',              "x",              'x',           0  },
   {  'c',              "c",              'c',           0  },
   {  'v',              "v",              'v',           0  },
   {  'b',              "b",              'b',           0  },
   {  'n',              "n",              'n',           0  },
   {  'm',              "m",              'm',           0  },
   {  ',',              "Comma",          ',',           0  },
   {  '.',              "Period",         '.',           0  },
   {  '/',              "Forward Slash",  '/',           0  },
   {  '\\',             "Backslash",      '\\',          0  },
   {  '~',              "Tilde",          '`',           __S},
   {  '!',              "Exclamation Mark",'1',          __S},
   {  '@',              "At",             '2',           __S},
   {  '#',              "Hash",           '3',           __S},
   {  '$',              "String",         '4',           __S},
   {  '%',              "Percent",        '5',           __S},
   {  '^',              "Caret",          '6',           __S},
   {  '&',              "Ampersand",      '7',           __S},
   {  '*',              "Star",           '8',           __S},
   {  '(',              "Open Parentheses",'9',          __S},
   {  ')',              "Close Parentheses",'0',         __S},
   {  '_',              "Underscore",     '-',           __S},
   {  '+',              "Plus",           '=',           __S},
   {  'Q',              "Q",              'q',           __S},
   {  'W',              "W",              'w',           __S},
   {  'E',              "E",              'e',           __S},
   {  'R',              "R",              'r',           __S},
   {  'T',              "T",              't',           __S},
   {  'Y',              "Y",              'y',           __S},
   {  'U',              "U",              'u',           __S},
   {  'I',              "I",              'i',           __S},
   {  'O',              "O",              'o',           __S},
   {  'P',              "P",              'p',           __S},
   {  '{',              "Open Brace",     '[',           __S},
   {  '}',              "Close Brace",    ']',           __S},
   {  'A',              "A",              'a',           __S},
   {  'S',              "S",              's',           __S},
   {  'D',              "D",              'd',           __S},
   {  'F',              "F",              'f',           __S},
   {  'G',              "G",              'g',           __S},
   {  'H',              "H",              'h',           __S},
   {  'J',              "J",              'j',           __S},
   {  'K',              "K",              'k',           __S},
   {  'L',              "L",              'l',           __S},
   {  ':',              "Colon",          ';',           __S},
   {  '\"',             "Doublequote",    '\'',          __S},
   {  'Z',              "Z",              'z',           __S},
   {  'X',              "X",              'x',           __S},
   {  'C',              "C",              'c',           __S},
   {  'V',              "V",              'v',           __S},
   {  'B',              "B",              'b',           __S},
   {  'N',              "N",              'n',           __S},
   {  'M',              "M",              'm',           __S},
   {  '<',              "Left Angle Bracket",',',        __S},
   {  '>',              "Right Angle Bracket",'.',       __S},
   {  '?',              "Question mark",  '/',           __S},
   {  '|',              "Pipe",           '\\',          __S},
   {  -1,               "unknown",        -1,            0  }
};


void io_lookup_key(long key, char **name, long *basekey, long *modifier) {

   keydata *p;
   long match;
   
   p = keys;
   match = 0;
   while(!match && p->key != -1) {
      if(p->key == key) match = 1;
      else p++;
   }
   if(name) *name = p->name;
   if(basekey) *basekey = p->basekey;
   if(modifier) *modifier = p->modifier;
   return;

}
