/* 

   libj - file_creat.c           Copyright (C) 2001-1998 Justin David Smith    
   justins@chaos2.org            http://chaos2.org/
    
   File creation


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjfile.h>
#include <libjstr.h>
#include <_file_macro.h>



static inline file *__file_creat(const char *n, char m, char *buf, sizea bufsize) {
/* Internal-File-Create
   This function creates a new file * structure and returns it.
   If buf is non-null, then f->buf and f->bufptr point to buf,
   f->bufsize is set, and buffer will not be released.  If buf
   is null and bufsize > 0, then f->buf is allocated, f->bufptr
   points past the end of the buffer, and f->bufsize is set.
   Otherwise, a NULL buffer is assumed. This function sets things
   up for memory buffers by default (R/W, f->f = -1, ...)   */

   file *r;
   
   if(n != NULL) {

      /* Allocate memory for file structure */
      r = (file *)malloc(sizeof(file));
      if(r != NULL) {
   
         /* Initialise values */
         r->name = (char *)malloc(strlenn(n) + 1);
         if(r->name != NULL) {

            strcopy(r->name, n);
            r->f = -1;
            r->line = 1;
            r->col = 1;
            r->byte = 0;
            r->buf = 0;
            r->bufsize = 0;
            r->bufptr = 0;
            r->close = false;
            r->release = false;
            r->eof = false;
            r->mode = __file_mode_from_char(m);
            r->func = false;
   
            /* Clear hooks */
            r->hooks.fopen = NULL;
            r->hooks.fclose= NULL;
            r->hooks.fread = NULL;
            r->hooks.fwrite= NULL;

            /* Initialise IO/memory buffer */
            if(buf != NULL) {
               r->buf = buf;
               r->bufptr = buf;
               r->bufsize = bufsize;
            } else if(bufsize > 0) {
               r->buf = (char *)malloc(bufsize);
               if(r->buf != NULL) {
                  *r->buf = '\0';
                  r->bufptr = r->buf + bufsize;
                  r->bufsize = bufsize;
                  r->release = true;
               }
            }
            if(r->buf != NULL) return(r);
            
            /* Something failed, start deallocating */
            free(r->name);
            
         } /* Name allocation? */
         
         free(r);
         
      } /* Structure allocation? */
   }
   
   /* Return failure */
   return(NULL);
   
}



static inline void __file_close(file *f) {
/* Internal-File-Close-File
   This function closes the file associated with <f>. */

   if(__file_is_file(f)) {
      /* Check for outstanding writes to the file */
      __file_write_buffer(f);
      if(f->close) {
         if(f->func) {
            if(f->hooks.fclose != NULL) f->hooks.fclose(f, f->f);
         } else close(f->f);
      }
   }

   f->f = -1;
   f->close = false;
   f->eof = true;

}



file *file_create(const char *n, char m, sizea bufsize) {
/* File-Create
   Creates a file structure based on an unopened file using the defaults
   provided. To be used on text files (well, duh). This function allocates
   memory needed for the file struct. The file will be opened and a file
   pointer assigned to it at this time. The file pointer will be closed
   when this structure is destroyed. <m> should be the file mode; 'r' for
   read or 'w' for write, or 'a' for append or 'x' for read/write. */

   file *r;
   
   /* Allocate memory */
   r = __file_creat(n, m, NULL, bufsize);
   if(r != NULL) {
   
      /* Open the file */
      r->f = open(n, r->mode, 0644);
      if(r->f >= 0) {
         if(__file_is_writable(r)) r->bufptr = r->buf;
         r->close = true;
         return(r);
      }
      
      /* Open failed */
      __file_release_buffer(r);
      free(r->name);
      free(r);
      
   }
   
   /* Return failure */
   return(NULL);
   
}



file *file_create_from_file(int f, const char *n, char m, sizea bufsize) {
/* File-Create-From-File
   Creates a file structure based on an already-open file using the defaults
   provided. This assumes the file pointer specified has already been opened.
   Note that once you start using the file I/O routines built-in, you should
   not use the native-language file I/O routines on the file since this
   structure will buffer file input as needed. To be used on text files
   (well, duh). This function allocates memory needed for the file struct. 
   The file pointer will NOT be closed when this structure is destroyed. */
   
   file *r;

   /* Allocate memory */
   r = __file_creat(n, m, NULL, bufsize);
   if(r != NULL) {
      r->f = f;
      if(__file_is_writable(r)) r->bufptr = r->buf;
      return(r);
   }
   return(NULL);
   
}
   


file *file_create_from_source(const char *n, char m, char *buf, sizea bufsize) {
/* File-Create-From-Source
   This function will simulate a rw-file in memory for use as a "fake file".
   This can be handy for reading large memory buffers as if they were files.
   The buf pointer will be copied, and the size of the buffer should be 
   indicated in bufsize: note that data will NOT be duplicated in this function, 
   and the buffer you indicate will not be released when the file pointer is 
   destroyed. */
   
   return(__file_creat(n, m, buf, bufsize));
   
}

   
   
file *file_create_from_memory(const char *n, char m, char *buf) {
/* File-Create-From-Memory
   This function will simulate a rw-file in memory for use as a "fake file".
   This can be handy for r/w'ing large memory buffers as if they were files.
   NOTE: buf should already be initialised as a null-terminated string. 
   This function will NOT duplicate the data in the buffer. */
   
   return(__file_creat(n, m, buf, strlenn(buf) + 1));

}

   
   
file *file_create_null(const char *n, char m) {
/* File-Create-Null
   Creates a null file buffer. Reads are infinite and return '\0'. Writes
   do nothing. */

   return(__file_creat(n, m, NULL, 0));
   
}


file *file_create_hooks(const char *n, char m, sizea bufsize, const filefn_hook *hooks) {
/* File-Create-Hooks */

   file *r;
   
   /* Allocate memory */
   r = __file_creat(n, m, NULL, bufsize);
   if(r != NULL) {
      /* Set hooks */
      r->hooks.fopen = hooks->fopen;
      r->hooks.fclose= hooks->fclose;
      r->hooks.fread = hooks->fread;
      r->hooks.fwrite= hooks->fwrite;
      r->func = true;
   
      /* Open the file */
      if(r->hooks.fopen != NULL) {
         r->f = r->hooks.fopen(r, n, r->mode);
         if(r->f >= 0) {
         
            /* Setup miscellaneous attributes before return */
            if(__file_is_writable(r)) r->bufptr = r->buf;
            r->close = true;
            return(r);

         }
      }         

      /* File open failed */
      __file_release_buffer(r);
      free(r->name);
      free(r);
   }
   
   return(NULL);
   
}



static inline void __file_destroy(file **f) {
/* Internal-File-Destroy
   Destroys the file pointer specified; closes files and releases buffers
   as needed. */

   if(f != NULL && *f != NULL) {
      __file_close(*f);
      __file_release_buffer(*f);
      free((*f)->name);
      free(*f);
      *f = NULL;
   }
   
}



void file_destroy(file **f) {
/* File-Destroy */

   __file_destroy(f);
   
}

      
void file_close(file **f) {
/* File-Close */

   __file_destroy(f);
   
}

      
