/* 

   libj - jfile.h                Copyright (C) 2001-1998 Justin David Smith
   justins@chaos2.org            http://chaos2.org/
    
   Main header file for jfile file I/O library


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#ifndef __libj_file_included
#define __libj_file_included


/* Standard includes */
#include <libj.h>



/* Options for read/write */
#define IO_READONLY     'r'
#define IO_WRITEONLY    'w'
#define IO_APPEND       'a'
#define IO_READWRITE    'x'


/* File structure */
typedef struct _file file;


/* Handler functions for certain actions */     /* WHAT ARE THESE */
typedef int  (*file_open_fn)  (file *, const char *, int);
typedef void (*file_close_fn) (file *, int);
typedef int  (*file_read_fn)  (file *, int, void *, int);
typedef int  (*file_write_fn) (file *, int, void *, int);


typedef struct _filefn_hook {
   file_open_fn   fopen;      /* Function hook open  file */
   file_close_fn  fclose;     /* Function hook close file */
   file_read_fn   fread;      /* Function hook read from file */
   file_write_fn  fwrite;     /* Function hook write  to file */
} filefn_hook;


struct _file {
   char *         name;       /* Filename or buffer name */
   int            f;          /* File handle or -1 for memory buffer */
   int            mode;       /* File IO mode or 0 if memory buffer */
   sizea          line;       /* Current line number (starts at 1) */
   sizea          col;        /* Current column number (starts at 1) */
   sizea          byte;       /* Number of bytes read/written */
   char *         buf;        /* IO buffer (file) or source buffer (mem) */
   sizea          bufsize;    /* Size of buf */
   char *         bufptr;     /* Pointer to current position in buf */
   bool           close;      /* Close file if true */
   bool           release;    /* Release buf if true */
   bool           eof;        /* Reached end of physical file */
   bool           func;       /* If True, use function hooks instead of IO */
   filefn_hook    hooks;      /* File I/O function hooks */
   void           *data;      /* Miscellaneous user-data */
};


/* Constructors */
file *file_create(const char *n, char m, sizea bufsize);
file *file_create_from_file(int f, const char *n, char m, sizea bufsize);
file *file_create_from_source(const char *n, char m, char *buf, sizea bufsize);
file *file_create_from_text(const char *n, char m, char *buf);
file *file_create_null(const char *n, char m);
file *file_create_hooks(const char *n, char m, sizea bufsize, const filefn_hook *hooks);


/* Destructor */
void  file_destroy(file **f);
void  file_close(file **f);


/* Informational */
bool  file_is_file(const file *f);
bool  file_is_memory(const file *f);
bool  file_is_null(const file *f);
bool  file_is_readable(const file *f);
bool  file_is_writable(const file *f);
const char *file_name(const file *f);
sizea file_line(const file *f);
sizea file_column(const file *f);
sizea file_byte(const file *f);
void  file_set_line(file *f, sizea line);
void  file_set_column(file *f, sizea col);
void  file_set_line_column(file *f, sizea line, sizea col);
void  file_set_byte(file *f, sizea b);


/* File status */
bool  file_eof(const file *f);
bool  file_flush(file *f);


/* Read */
bool  file_peek_char(file *f, char *ch);
bool  file_read_char(file *f, char *ch);
sizea file_read_line(file *f, char *rbuf, sizea size);
sizea file_read_string(file *f, char *rbuf, sizea size);
sizea file_read_buffer(file *f, char *rbuf, sizea size);
bool  file_read_word(file *f, word *ch);
bool  file_read_dword(file *f, dword *ch);
bool  file_read_sword(file *f, sword *ch);
bool  file_read_sdword(file *f, sdword *ch);


/* Write */
bool  file_write_char(file *f, char ch);
bool  file_write_line(file *f, const char *wbuf);
bool  file_write_string(file *f, const char *wbuf);
bool  file_write_buffer(file *f, const char *wbuf, sizea size);
bool  file_write_word(file *f, word ch);
bool  file_write_dword(file *f, dword ch);
bool  file_write_sword(file *f, sword ch);
bool  file_write_sdword(file *f, sdword ch);


#endif

