/* $Header: /fridge/cvs/libj/jlink/tree.c,v 1.5 2001/12/10 19:43:59 justins Exp $ */                                                                                                    
/* 

   libj - tree.c                 Copyright (C) 2001-1998 Justin David Smith
   justins (a) chaos2.org        http://chaos2.org/
    
   Functions for maintaining a binary tree


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjtree.h>
#include <libjstr.h>



/*** Tree Node Maintenance ***/


static inline bintreenode *_bintree_create_node(void *d, bintreenode *parent, bintreenode *left, bintreenode *right, const bintreecolor rb) {
/* bintree-Create-Node
   Returns a new node for the binary tree. Data is initialised according to
   the values provided. Note that new nodes should default to RED. This 
   function returns NULL on an error (mainly if the memory block could
   not be allocated) */

   bintreenode *n;
   
   n = (bintreenode *)malloc(sizeof(bintreenode));
   if(n == NULL) return(NULL);
   n->p = parent;
   n->l = left;
   n->r = right;
   n->c = rb;
   n->data = d;

   return(n);

}


static inline void _bintree_delete_node_data(void *d, const destructorfn dfunc) {
/* bintree-Delete-Node-Data
   Deletes data on a node. */
   
   if(dfunc != NULL) dfunc(d);
   
}


static inline void _bintree_delete_node(bintreenode *n, const destructorfn dfunc, const bool deletedata) {
/* bintree-Delete-Node
   This function deletes a node and the associated data with the node,
   if <dfunc> is a pointer to a function. Returns true on success; false
   if the <dfunc> call failed. */

   if(n == NULL)  return;
   if(deletedata) _bintree_delete_node_data(n->data, dfunc);
   free(n);

}


/*** Red Black Self-Balancing Code ***/


static inline bool leaf(const bintree *t, const bintreenode *n) {
/* is-Leaf
   Is this node a leaf, i.e. either NULL or sentinel? Return true if it
   is. */

   if(t == NULL) return(true);
   return(n == NULL || n == t->sent);

}


static inline bintreecolor color(const bintree *t, const bintreenode *n) {
/* node-Color
   Color of node? Get color or return BLACK if the node is actually a leaf. */

   if(t == NULL) return(BLACK);
   if(leaf(t, n)) return(n->c);
   else return(BLACK);

}


static void _bintree_rotate_left(bintree *t, bintreenode *n) {
/* bintree-Rotate-Left
   Rotates the tree to the left around node <n> */

   bintreenode *y;       /* y = New "root" node of this subintree */

   if(t == NULL || n == NULL) return;
   y = n->r;
   
   /* Move left tree to right tree. */
   n->r = y->l;
   if (!leaf(t, y->l)) y->l->p = n;

   /* Fix links to parent from new root */
   y->p = n->p;

   /* Change 'root' node to new root */
   if(!n->p) t->root = y;
   else if(n == n->p->l) n->p->l = y;
   else n->p->r = y;
   
   /* Fix new root's links to old root */
   y->l = n;
   n->p = y;

}


void _bintree_rotate_right(bintree *t, bintreenode *n) {
/* bintree-Rotate-Right
   Rotate the tree to the right around node <n> */

   bintreenode *y;       /* y = new 'root' of this subintree */
   
   if(t == NULL || n == NULL) return;
   y = n->l;
   
   /* Move right tree to left. */
   n->l = y->r;
   if (!leaf(t, y->r)) y->r->p = n;

   /* Fix new root's parentage */
   y->p = n->p;

   /* Welcome to the Enlightenment */
   if(!n->p) t->root = y;
   else if(n == n->p->r) n->p->r = y;
   else n->p->l = y;
   
   /* Fix links between new root and old root */
   y->r = n;
   n->p = y;

}


bool bintree_insert(bintree *t, void *d) {
/* bintree-Insert
   Insert a data item pointed to by <d> into the tree, then rebalance */

   /* Data */
   bintreenode *n;              /* Search for place to insert to */
   bintreenode **gonode;        /* Used in searching for insertion point */
   bintreenode *y;              /* Temporary(used in balancing) */
   bool goleft = false;         /* Which way to go? (used in insertion) */

   #ifdef LIBTREE_DEBUG_INSERT
   printf("bintreeInsert:  Entering function with tree = %8p, data = %8p\n", t, d);
   #endif

   if(t == NULL) return(false);
   n = t->root;

   /* Insertion */
   if(!leaf(t, t->root)) {
      /* Root exists; adding to an already thriving tree */
      #ifdef LIBTREE_DEBUG_INSERT
      printf("bintreeInsert:  Search for insert point [");
      fflush(stdout);
      #endif
      do {
         goleft = (t->lessthan(d, n->data));
         gonode = (goleft ? &(n->l) : &(n->r));
         if(!leaf(t, *gonode)) n = *gonode;
         #ifdef LIBTREE_DEBUG_INSERT
         printf("%c", goleft ? 'l' : 'r');
         fflush(stdout);
         #endif
      } while(!leaf(t, *gonode));
      #ifdef LIBTREE_DEBUG_INSERT
      printf("] for data %8p\n", d);
      #endif
      if(NULL == (n = (*gonode = _bintree_create_node(d, n, t->sent, t->sent, RED)))) return(false);
   } else {
      #ifdef LIBTREE_DEBUG_INSERT
      printf("bintreeInsert:  Creating the root node for data %8p\n", d);
      #endif
      if(NULL == (n = (t->root = _bintree_create_node(d, NULL, t->sent, t->sent, RED)))) return(false);
   }

   /* Red-black rebalancing begins here. */
   while(n != t->root && color(t, n->p) == RED) {
      if(n->p == n->p->p->l) {
         /* n->p EXISTS and n->p is on left side of parent */
         y = n->p->p->r;
         if(color(t, y) == RED) {
            n->p->c = BLACK;
            y->c = BLACK;
            n->p->p->c = RED;
            n = n->p->p;
         } else {
            if(n == n->p->r) {
               n = n->p;
               _bintree_rotate_left(t, n);
            }
            n->p->c = BLACK;
            n->p->p->c = RED;
            _bintree_rotate_right(t, n->p->p);
         }
      } else {
         /* n->p EXISTS and n->p is on right side of its parent */
         /* This is a mirror image of above code */
         y = n->p->p->l;
         if(color(t, y) == RED) {
            n->p->c = BLACK;
            y->c = BLACK;
            n->p->p->c = RED;
            n = n->p->p;
         } else {
            if(n == n->p->l) {
               n = n->p;
               _bintree_rotate_right(t, n);
            }
            n->p->c = BLACK;
            n->p->p->c = RED;
            _bintree_rotate_left(t, n->p->p);
         }
      }
   }
   /* Change color of root node if it is not already black. */
   t->root->c = BLACK;

   #ifdef LIBTREE_DEBUG_INSERT
   printf("bintreeInsert:  Done inserting data.\n");
   #endif
   
   return(true);

}


static bintreenode *_bintree_successor(bintree *t, bintreenode *x) {
/* bintree-Successor
   Find the successor of the given element <x>, that is, find the element
   with the smallest key greater than <x>'s key. Happy searching! */

   bintreenode *y;     /* Happy searching! */

   if(!leaf(t, x->r)) {
      /* Successor is a descendant; find the leftmost entry in the right */
      /* subintree, that'll be our successor */
      y = x->r;
      while(!leaf(t, y->l)) y = y->l;
   } else {
      /* Successor must be one of the parents of this node. */
      y = x->p;
      while(!leaf(t, y) && x == y->r) {
         x = y;
         y = y->p;
      }
   }
   return(y);

}


static inline bool _bintree_data_equal(comparefn lessthan, const void *a, const void *b) {

   return(!lessthan(a, b) && !lessthan(b, a));

}


bool bintree_delete(bintree *t, const void *d, bool deletedata) {
/* bintree-Delete
   Delete the entry matching <d>. This works in three parts; a search op,
   the splicing and then finally the rebalancing. woohoo! Note that only
   the first entry which isEqual() function returns true will be deleted,
   so use caution in your equals function >:) */
   
   /* Data */
   bintreenode *n;              /* Searching for node to whack */
   bintreenode *x;              /* Used in splice */
   bintreenode *y;              /* Used in splice- what'll be deleted */
   bintreenode *w;              /* Used in rebalancing */
   
   if(t == NULL) return(false);
   n = t->root;
   
   /* Search for matching data entry */
   while(!leaf(t, n) && !_bintree_data_equal(t->lessthan, n->data, d)) n = (t->lessthan(n->data, d) ? n->r : n->l);
   
   if(!leaf(t, n)) {
      /* A matching node was found in the tree. */
      if(leaf(t, n->l) || leaf(t, n->r)) {
         /* One or two leaves below node to be deleted. */
         y = n;
         /* at least one NULL on child */
      } else {
         /* Node to be deleted has 2 children- we don't want to drastically */
         /* change the shape of the tree, so we'll swap this node out with */
         /* its successor and delete that node instead :) */
         y = _bintree_successor(t, n);
      }
      
      /* Set X */
      if(!leaf(t, y->l)) x = y->l;
      else x = y->r;
      
      /* Setup X's parent (It is possible at this point for X to be "null", */
      /* that is for X to be a leaf. In this case we are setting up the */
      /* parent of that particular leaf. */
      x->p = y->p;

      /* Fix up pointer to Y (node that will be deleted) */
      if(!y->p) t->root = x;
      else if(y == y->p->l) y->p->l = x;
      else y->p->r = x;
      
      if(y != n) {
         /* Need to copy data stored in Y to node N-- we are swapping out */
         /* the node to delete with its successor here */
         if(deletedata) _bintree_delete_node_data(n, t->destr);
         deletedata = false;
         n->data = y->data;
         n->c = color(t, y);
      }
      
      /* Rebalancing portion of the algorithm. */
      if(color(t, y) == BLACK) {
         while(x != t->root && color(t, x) == BLACK) {
            if(x == x->p->l) {
               w = x->p->r;
               if(color(t, w) == RED) {
                  w->c = BLACK;
                  x->p->c = RED;
                  _bintree_rotate_left(t, x->p);
                  w = x->p->r;
               }
               if(color(t, w->l) == BLACK && color(t, w->r) == BLACK) {
                  w->c = RED;
                  x = x->p;
               } else {
                  if(color(t, w->r) == BLACK) {
                     w->l->c = BLACK;
                     w->c = RED;
                     _bintree_rotate_right(t, w);
                     w = x->p->r;
                   }
                  w->c = color(t, x->p);
                  x->p->c = BLACK;
                  w->r->c = BLACK;
                  _bintree_rotate_left(t, x->p);
                  x = t->root;
               }
            } else {
               w = x->p->l;
               if(color(t, w) == RED) {
                  w->c = BLACK;
                  x->p->c = RED;
                  _bintree_rotate_right(t, x->p);
                  w = x->p->l;
               }
               if(color(t, w->r) == BLACK && color(t, w->l) == BLACK) {
                  w->c = RED;
                  x = x->p;
               } else {
                  if(color(t, w->l) == BLACK) {
                     w->r->c = BLACK;
                     w->c = RED;
                     _bintree_rotate_left(t, w);
                     w = x->p->l;
                  }
                  w->c = color(t, x->p);
                  x->p->c = BLACK;
                  w->l->c = BLACK;
                  _bintree_rotate_right(t, x->p);
                  x = t->root;
               }
            }
         }
         x->c = BLACK;
      }

      _bintree_delete_node(y, t->destr, deletedata);
         
      return(true);
   } else return(false);

}


static void _bintree_delete_entire_subtree(bintree *t, bintreenode **n, bool deletedata) {
/* bintree-Delete-Entire-Subtree
   Deletes this node and all nodes in this subintree- use frequently if you
   love data loss :) */

   if(leaf(t, *n)) return;
   if(!leaf(t, (*n)->l))   _bintree_delete_entire_subtree(t, &((*n)->l), deletedata);
   if(!leaf(t, (*n)->r))   _bintree_delete_entire_subtree(t, &((*n)->r), deletedata);
   _bintree_delete_node(*n, t->destr, deletedata);
   *n = NULL;

}


/*** Some Global Creation/Destruction Functions ***/


bintree *bintree_new(comparefn lessthan, destructorfn destr) {
/* bintree-new
   This function initialises a bintree variable. This function will setup the
   required properties for a bintree. You will have to create four supplemental
   functions for the bintree to function properly, as described below. The tree
   will be initialised as an empty tree, after which you may start adding
   nodes to the tree.
   
   The function prototypes required to keep the tree in order are as follows.
   
   ltfunc: bool ltfunc(void *a, void *b)
   Returns true if the contents of <a> are strictly less than the contents
   of <b>. Several generic functions have been provided for some standard
   data types: bintreeIntLessThan, where the data are int *'s; bintreeLongLessThan,
   where the data are long *'s; bintreeDoubleLessThan, where the data are
   double *'s; bintreeStringLessThan, where the data are null-terminated strings
   (this function will do a case-sensitive comparison). Most likely, you will
   have to supply your own function for this purposes. Note also that the 
   string comparison function uses strcmp() and is hardly efficient for this
   task; it is recommended you use a custom string comparison function.
   
   dlfunc: bool dlfunc(void *a) 
   Deletes the pointer. This is called to delete data associated with a node
   when the node is deleted from the tree. Two generic functions have been 
   provided for convenience: bintreeFreeData, which uses a generic free() call
   on the pointer a, and bintreeFreeDataInert, which does NOTHING. Use the
   latter call if you do NOT want to delete data associated with a node when
   the node is destroyed (f.e, if you are using the tree to represent
   data that is stored in another structure). This function must return true;
   return false on a grave error, which will cause a cascade error that will
   eventually be returned by the function that was deleting the node: RETURNING
   FALSE WILL RESULT IN CORRUPTION OF THE DATA TREE! Be warned; just return
   true.  */
   
   bintree *t;
   
   if(lessthan == NULL) return(NULL);
   t = (bintree *)malloc(sizeof(bintree));
   if(t == NULL) return(NULL);
      
   t->sent = _bintree_create_node(NULL, NULL, NULL, NULL, RED);
   t->root = t->sent;
   t->lessthan = lessthan;
   t->destr    = destr;
   return(t);

}


static inline void _bintree_destroy(bintree **t, const bool deletedata) {
/* bintree-Destroy
   Deletes the ENTIRE tree <t>. Nodes will be destroyed and associated 
   data released with the deleter function, as needed. WARNING: AFTER 
   THIS FUNCTION IS CALED YOU MUST RECALL bintreeCreate() to start a new
   tree because the sentinel node will be destroyed! */
   
   if(t == NULL || *t == NULL) return;
   
   _bintree_delete_entire_subtree((*t), &(*t)->root, deletedata);
   _bintree_delete_node((*t)->sent, NULL, false);
   free(*t);
   *t = NULL;
   return;
   
}


void bintree_release(bintree **t) {
/* bintree-Destroy
   Deletes the ENTIRE tree <t>. Nodes will be destroyed and associated 
   data released with the deleter function, as needed. WARNING: AFTER 
   THIS FUNCTION IS CALED YOU MUST RECALL bintreeCreate() to start a new
   tree because the sentinel node will be destroyed! */

   _bintree_destroy(t, true);
   
   
}


void bintree_collapse(bintree **t) {
/* bintree-collapse
   Deletes the ENTIRE tree <t>. Nodes will be destroyed, but associated 
   data will NOT be release.  */

   _bintree_destroy(t, false);
   
}


/*** Accessor functions ***/


void bintree_walk_node(const bintree *t, const bintreenode *n, bintree_walkf f, sizea *i, void *params) {
/* bintree-Walk-Node-LNR
   Walks the subintree- internal recursive function. Walks along node <s> and
   calls function <f> for this node, with optional parameters (specified
   by the end user) specified in <param>. */

   if(!leaf(t, n->l)) bintree_walk_node(t, n->l, f, i, params);
   if(f(n->data, params)) (*i)++;
   if(!leaf(t, n->r)) bintree_walk_node(t, n->r, f, i, params);
   
}


sizea bintree_walk(const bintree *t, bintree_walkf f, void *params) {
/* bintree-Walk-LNR                                         node
   Walks the tree in the order left-node-right             /    \
   The function argument <f> is of the form:           left      right
   bool f(const void *data, void *param)
   where data is a pointer to the data stored in each node and param is
   a pointer to the parameter passed to this function. This walker will walk
   through the entire tree, It will return the number of nodes walked on which
   the function f() returned true (so you can use this as a counter for nodes
   which match some certain criterion, f.e.). A default walker function is
   provided, bintreeWalkCount, which just returns true. */

   sizea i = 0;
   if(t == NULL || f == NULL) return(0);

   if(leaf(t, t->root)) return(0);
   bintree_walk_node(t, t->root, f, &i, params);
   return(i);

}


void bintree_walk_node_rev(const bintree *t, const bintreenode *n, bintree_walkf f, sizea *i, void *params) {
/* bintree-Walk-Node-RNL
   Walks the subintree- internal recursive function. Walks along node <s> and
   calls function <f> for this node, with optional parameters (specified
   by the end user) specified in <param>. */

   if(!leaf(t, n->r)) bintree_walk_node_rev(t, n->r, f, i, params);
   if(f(n->data, params)) (*i)++;
   if(!leaf(t, n->l)) bintree_walk_node_rev(t, n->l, f, i, params);
   
}


sizea bintree_walk_rev(const bintree *t, bintree_walkf f, void *params) {
/* bintree-Walk-RNL                                         node
   Walks the tree in the order right-node-left             /    \
   The function argument <f> is of the form:           left      right
   bool f(const void *data, void *param)
   where data is a pointer to the data stored in each node and param is
   a pointer to the parameter passed to this function. This walker will walk
   through the entire tree, It will return the number of nodes walked on which
   the function f() returned true (so you can use this as a counter for nodes
   which match some certain criterion, f.e.). A default walker function is
   provided, bintreeWalkCount, which just returns true. */

   sizea i = 0;
   if(t == NULL || f == NULL) return(0);

   if(leaf(t, t->root)) return(0);
   bintree_walk_node_rev(t, t->root, f, &i, params);
   return(i);

}


bool bintree_search_node(const bintree *t, const bintreenode *s, const void *d, bintree_walkf f, void *params) {
/* Search for <d> in subintree <s>. Store result in <r> on success and return
   true. Recursive internal function */
     
   if(_bintree_data_equal(t->lessthan, s->data, d)) {
      if(f != NULL) f(d, params);
      return(true);
   } else if(t->lessthan(s->data, d)) {
      if(!leaf(t, s->r)) return(bintree_search_node(t, s->r, d, f, params));
   } else {
      if(!leaf(t, s->l)) return(bintree_search_node(t, s->l, d, f, params));
   }
   return(false);

}


bool bintree_search(const bintree *t, const void *d, bintree_walkf f, void *params) {
/* Search interface function. Returns true on success and copies data pointer
   or something to <r> if you care. */

   if(t == NULL || f == NULL) return(false);
   
   if(leaf(t, t->root)) return(false);
   return(bintree_search_node(t, t->root, d, f, params));

}


bool bintree_search_node_rev(const bintree *t, const bintreenode *s, const void *d, bintree_walkf f, void *params) {
/* Search for <d> in subintree <s>. Store result in <r> on success and return
   true. Recursive internal function */
     
   if(_bintree_data_equal(t->lessthan, s->data, d)) {
      if(f != NULL) f(d, params);
      return(true);
   } else if(t->lessthan(s->data, d)) {
      if(!leaf(t, s->r)) return(bintree_search_node_rev(t, s->r, d, f, params));
   } else {
      if(!leaf(t, s->l)) return(bintree_search_node_rev(t, s->l, d, f, params));
   }
   return(false);

}


bool bintree_search_rev(const bintree *t, const void *d, bintree_walkf f, void *params) {
/* Search interface function. Returns true on success and copies data pointer
   or something to <r> if you care. */

   if(t == NULL) return(false);
   
   if(leaf(t, t->root)) return(false);
   return(bintree_search_node_rev(t, t->root, d, f, params));

}



/*** Generic Use Functions ***/


bool bintree_int_lessthan(const void *a, const void *b) {
   return(*((int *)a) < *((int *)b));
}


bool bintree_long_lessthan(const void *a, const void *b) {
   return(*((long *)a) < *((long *)b));
}


bool bintree_double_lessthan(const void *a, const void *b) {
   return(*((double *)a) < *((double *)b));
}


bool bintree_string_lessthan(const void *a, const void *b) {
   return(strcomp((char *)a, (char *)b) < 0);
}


bool bintree_walk_count(const void *data, void *param) {
   return(true);
}



/*** Miscellaneous utilities ***/


int bintree_count(const bintree *t) {
/* Return the number of nodes in the tree */

   if(t == NULL) return(false);
   return(bintree_walk(t, bintree_walk_count, NULL));

}



/*** Node-related utilities ***/


const bintreenode *bintree_find_first_node(const bintree *t, const bintreenode *node, const void *data) {
/* Search for the first matching node */

   const bintreenode *find;
   
   /* If this node is NULL, then it cannot be a match */
   if(leaf(t, node)) return(NULL);

   if(_bintree_data_equal(t->lessthan, node->data, data)) {
      /* This node matches; make sure there isn't an 
         earlier match on the left branch, first. */
      find = bintree_find_first_node(t, node->l, data);
      if(find == NULL) return(node);
      return(find);
   }
   
   /* This node isn't a match; check which direction we should go */
   if(t->lessthan(node->data, data)) {
      return(bintree_find_first_node(t, node->r, data));
   } else {
      return(bintree_find_first_node(t, node->l, data));
   }

}


const bintreenode *bintree_find_first(const bintree *t, const void *d) {
/* Try to find the first node in the tree that matches with the
   data given, using the comparator defined in the tree.  On
   success the first matching node is returned.  */

   if(t == NULL) return(false);
   return(bintree_find_first_node(t, t->root, d));

}


const bintreenode *bintree_first_node(const bintree *t, const bintreenode *node) {
/* Try to find the first node in a subtree */

   if(node == NULL) return(NULL);
   while(!leaf(t, node->l)) node = node->l;
   return(node);
   
}


const bintreenode *bintree_last_node(const bintree *t, const bintreenode *node) {
/* Try to find the last node in a subtree */

   if(node == NULL) return(NULL);
   while(!leaf(t, node->r)) node = node->r;
   return(node);

}


const bintreenode *bintree_prev_node(const bintree *t, const bintreenode *node) {
/* Try to find the previous node in the tree */

   if(node == NULL) return(NULL);
   if(!leaf(t, node->l)) return(bintree_last_node(t, node->l));
   while(!leaf(t, node->p) && node->p->l == node) {
      node = node->p;
   }
   return(node->p);

}


const bintreenode *bintree_next_node(const bintree *t, const bintreenode *node) {
/* Try to find the next node in the tree */

   if(node == NULL) return(NULL);
   if(!leaf(t, node->r)) return(bintree_first_node(t, node->r));
   while(!leaf(t, node->p) && node->p->r == node) {
      node = node->p;
   }
   return(node->p);

}



/*** Pairs ***/


bintreepair *pair_alloc(const char *key, const char *value) {
/* Allocate a new pair */

   bintreepair *pair;
   pair = (bintreepair *)malloc(sizeof(bintreepair));
   pair->key = strdupl(key);
   pair->value = strdupl(value);
   return(pair);

}


void pair_update_value(bintreepair *pair, const char *value) {
/* Reassign the value associated with the pair */

   if(pair == NULL) return;
   free((void *)pair->value);
   pair->value = strdupl(value);

}


bool bintree_pair_lessthan(const void *a, const void *b) {
/* Ordering function on pairs; warning, both arguments must
   be the same type, therefore when searching for a key then
   it must also be packaged in a pair object. */

   if(a == NULL || b == NULL) return(a == NULL && b != NULL);
   return(bintree_string_lessthan(((bintreepair *)a)->key, ((bintreepair *)b)->key));
   
}


const bintreenode *bintree_pair_find_first(const bintree *t, const char *key) {
/* Search for the first matching key in the binary tree. */

   bintreepair pair;
   pair.key = (char *)key;
   pair.value = NULL;
   return(bintree_find_first(t, &pair));

}


bool bintree_pair_delete(bintree *t, const char *key, bool deletedata) {
/* Delete a pair from the tree */

   bintreepair pair;
   pair.key = (char *)key;
   pair.value = NULL;
   return(bintree_delete(t, &pair, deletedata));

}


void pair_free(bintreepair *pair) {
/* Release the pair */

   if(pair == NULL) return;
   free((void *)pair->key);
   free((void *)pair->value);
   free(pair);
   
}
