/* 

   libj - math.c                 Copyright (C) 2001-1998 Justin David Smith
   justins@chaos2.org            http://chaos2.org/
    
   Miscellaneous math functions


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmath.h>



qword gcd(qword a, qword b) {
/* Greatest-Common-Divisor
   This is based on Euclid's algorithm, see Algorithms. Modified so that
   this algorithm is not recursive. */
   
   qword c;
   
   while(b > 0) {
      c = a % b;
      a = b;
      b = c;
   }
   return(a);

}



qword mod_exponent(qword a, qword b, qword n) {
/* Modular-Exponentiation
   Raise a^b (mod n). See section 33.6 of Algorithms */

   qword c = 0;
   qword d = 1;
   qword k = (sizeof(qword) << 3);
   qword m = (1 << k);

   while(--k > 0) {
      c = (c << 1);
      d = (d * d) % n;
      if((b & m) != 0) {
         c++;
         d = (d * a) % n;
      }
      m = (m >> 1);
   }
   return(d);

}



bool pseudo_prime(qword n) {
/* Pseudo-Prime-Test
   See Algorithms 33.8. Returns likely prime numbers. */
   
   qword c = 0;
   qword d = 1;
   qword k = (sizeof(qword) << 3);
   qword m = (1 << k);
   qword b = (n - 1);

   while(--k > 0) {
      c = (c << 1);
      d = (d * d) % n;
      if(b & m) {
         c++;
         d = (d << 1) % n;
      }
      m = (m >> 1);
   }
   return(d == 1);

}



bool mr_prime(qword n, qword trials) {
/* Miller-Rabin-Prime */

   qword a;
   qword d;
   qword k;
   qword m;
   qword x;
   qword b = (n - 1);
   
   while(trials-- > 0) {
      a = (rand() % n) + 1;
      d = 1;
      k = (sizeof(qword) << 3);
      m = (1 << k);
      while(--k > 0) {
         x = d;
         d = (d * d) % n;
         if(d == 1 && x != 1 && x != (n - 1)) return(false);
         if(b & m) d = (d * a) % n;
         m = (m >> 1);
      }
      if(d != 1) return(false);
   }
   return(true);

}



qword pr_factor(prstate *pr, qword n, qword **r, qword runtime) {
/* Pollard-Rho-Factorization 
   runtime ~ n ^ 0.25 will print most if not all factors of n. */

   qword d;

   if(pr == NULL) return(0);   
   if(r == NULL) return(0);
   *r = 0;

   pr->pr_n = n;   
   pr->pr_i = 1;
   pr->pr_x = rand() % n;
   pr->pr_y = pr->pr_x;
   pr->pr_k = 2;
   pr->pr_rcount = 0;
   pr->pr_r = r;
   
   while(runtime-- > 0) {
      pr->pr_i++;
      pr->pr_x = ((pr->pr_x * pr->pr_x) - 1) % pr->pr_n;
      d = gcd(pr->pr_y - pr->pr_x, pr->pr_n);
      if(d != 1 && d != pr->pr_n) {
         /* d is another factor of n */
         *pr->pr_r = (qword *)realloc(*pr->pr_r, (++pr->pr_rcount) * sizeof(qword));
         *(pr->pr_r)[pr->pr_rcount - 1] = d;
      }
      if(pr->pr_i == pr->pr_k) {
         pr->pr_y = pr->pr_x;
         pr->pr_k = (pr->pr_k << 1);
      }
   }

   return(pr->pr_rcount);
   
}



bool pr_verify(prstate *pr) {
/* Pollard-Rho-Verify */

   qword i;
   qword *r;
   qword res = 0;
   
   if(pr == NULL || pr->pr_r == NULL) return(false);
   i = pr->pr_rcount;
   r = *pr->pr_r;
   
   while(i-- > 0) res *= *(r++);
   return(res == pr->pr_n);

}



qword pr_resume(prstate *pr, qword runtime) {
/* Pollard-Rho-Resume */

   qword d;
   
   if(pr == NULL || pr->pr_r == NULL) return(0);
   
   while(runtime--) {
      pr->pr_i++;
      pr->pr_x = ((pr->pr_x * pr->pr_x) - 1) % pr->pr_n;
      d = gcd(pr->pr_y - pr->pr_x, pr->pr_n);
      if(d != 1 && d != pr->pr_n) {
         // d is another factor of n
         *pr->pr_r = (qword *)realloc(*pr->pr_r, (++pr->pr_rcount) * sizeof(qword));
         *(pr->pr_r)[pr->pr_rcount - 1] = d;
      }
      if(pr->pr_i == pr->pr_k) {
         pr->pr_y = pr->pr_x;
         pr->pr_k = (pr->pr_k << 1);
      }
   }

   return(pr->pr_rcount);
   
}


