/* 

   libj - gmtx_cons.c            Copyright (C) 2001-1998 Justin David Smith
   justins@chaos2.org            http://chaos2.org/
    
   Constructors and destructors


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>
#include <libjmem.h>
#include <libjstr.h>



/*** Constructors and Destructors ***/
/* Internal Data allocation and reallocation */



static inline gmtx *__gmtx_realloc(gmtx *m, dimensiona bytesize) {
/* Reallocate-data-buffer

   If bytesize = 0, then buffer is freed.
   If bytesize > 0, buffer allocated. */

   void *p;
   
   if(!(p = realloc(m->data, bytesize))) return(NULL);
   m->data = p;
   return(m);

}


static inline void __gmtx_set_size(gmtx *m, dimensiona width, dimensiona height) {
   
   m->width  = width;                  /* Width (cells) */
   m->height = height;                 /* Height(cells) */
   m->size   = width * height;         /* Size in cells */
   m->bufsize= m->size * m->dtsize;    /* data buffer size(bytes) */
   m->rowsize= width * m->dtsize;      /* row buffer size(bytes) */

}


static inline void __gmtx_set_attr(gmtx *m, uword type, uword dtsize, dimensiona width, dimensiona height) {

   m->type   = type;                   /* Matrix type flag */
   m->dtsize = dtsize;                 /* data entry size (bytes) */
   __gmtx_set_size(m, width, height);  /* Set new size flags */
   return;

}


static inline gmtx *__gmtx_alloca(uword type, uword dtsize) {

   gmtx *m;
   
   /* Allocate matrix */
   if(!(m = (gmtx *)malloc(sizeof(gmtx)))) return(NULL);
   
   /* Initialise data values */
   __gmtx_set_attr(m, type, dtsize, 0, 0);
   m->data = NULL;
   return(m);
   
}
   

static inline void __gmtx_free(gmtx *m) {

   free(m->data); /* Release matrix contents */
   free(m);       /* Release matrix structure */
   
}


static gmtx *_gmtx_resize(gmtx *m, dimensiona width, dimensiona height) {
/* Matrix-Resize

   If either width, height are zero, then matrix will be a zero matrix
   with no data pointer, otherwise width, height are set, and data pointer
   is allocated.  Previous data is lost. */

   /* If matrix is zero, set width,height to zero and release data ptr */
   if(width <= 0 || height <= 0) width = 0, height = 0;

   /* Reallocate data buffer, modify attributes */
   if(!__gmtx_realloc(m, width * height * m->dtsize)) return(NULL);
   __gmtx_set_size(m, width, height);
   return(m);

}


static gmtx *_gmtx_resize_using(gmtx *mdst, const gmtx *msrc) {
/* Matrix-Resize-Using

   Same as above, but mdst's type and dtsize will be set to the source
   type and dtsize. All previous destination contents are lost */

   /* Attributes from src are also copied in this resize */
   if(!__gmtx_realloc(mdst, msrc->width * msrc->height * msrc->dtsize)) return(NULL);
   __gmtx_set_attr(mdst, msrc->type, msrc->dtsize, msrc->width, msrc->height);
   return(mdst);

}


static gmtx *_gmtx_reshape(gmtx *m, dimensiona width, dimensiona height) {
/* Matrix-Reshape 

   Unlike resize, current data is preserved.  width * height must
   equal old matrix's size parameter.  No data pointers are reallocated,
   so data is preserved. */

   if(width * height != m->size) return(NULL);
   __gmtx_set_size(m, width, height);
   return(m);
   
}


static gmtx *_gmtx_reshape_using(gmtx *mdst, const gmtx *msrc) {
/* Matrix-Reshape 

   Like above.  Matrices must be compatible, since type and dtsize are
   not copied and data is preserved. */

   if(MTX_ARE_COMPAT(mdst, msrc)) return(NULL);
   if(msrc->width * msrc->height != mdst->size) return(NULL);
   __gmtx_set_size(mdst, msrc->width, msrc->height);
   return(mdst);
   
}


static gmtx *_gmtx_new(uword type, uword dtsize, dimensiona width, dimensiona height) {

   gmtx *m;
   
   /* Type must include "GRID" flag */
   type |= MTX_GRID;

   /* Allocate memory for the matrix */
   if(!(m = __gmtx_alloca(type, dtsize))) return(NULL);
   
   /* Initialise matrix */
   _gmtx_resize(m, width, height);
   return(m);
   
}


static gmtxl *_gmtxl_new(dimensiona width, dimensiona height) {

   return(__gmtxlp _gmtx_new(MTX_LONG, sizeof(longdata), width, height));

}


static gmtxd *_gmtxd_new(dimensiona width, dimensiona height) {

   return(__gmtxdp _gmtx_new(MTX_DOUBLE, sizeof(doubledata), width, height));

}


static gmtx *_gmtx_new_using(const gmtx *m) {

   /* Allocate using m as a guideline */
   return(_gmtx_new(m->type, m->dtsize, m->width, m->height));
   
}


static gmtx *_gmtx_free(gmtx **m) {

   if(!m || !*m) return(NULL);
   __gmtx_free(*m);
   *m = NULL;
   return(NULL);

}


/*** Constructors and Destructors ***/
/* Internal initialiser functions */


static gmtxl *_gmtxl_const(gmtxl *m, longdata value) {

   longdata *p;
   dimensiona size;
   
   /* Initialise all values */
   MTX_LOOP_SIZEP(m, size, p);
   MTX_LOOP_ASSIGN(size, p, value);
   return(m);
   
}


static gmtxd *_gmtxd_const(gmtxd *m, doubledata value) {

   doubledata *p;
   dimensiona size;
   
   /* Initialise all values */
   MTX_LOOP_SIZEP(m, size, p);
   MTX_LOOP_ASSIGN(size, p, value);
   return(m);
   
}


static gmtxl *_gmtxl_new_const(dimensiona width, dimensiona height, longdata value) {

   gmtxl *mnew;
   
   if(!(mnew = _gmtxl_new(width, height))) return(NULL);
   return(_gmtxl_const(mnew, value));
   
}


static gmtxd *_gmtxd_new_const(dimensiona width, dimensiona height, doubledata value) {

   gmtxd *mnew;
   
   if(!(mnew = _gmtxd_new(width, height))) return(NULL);
   return(_gmtxd_const(mnew, value));
   
}


static gmtxl *_gmtxl_eyes(gmtxl *m, longdata value) {

   longdata *p;
   dimensiona width;
   dimensiona skip;
   dimensiona remainder;
   dimensiona numeyes;
   
   /* Determine number of "eyes" to put in */
   /* If zero matrix, numeyes == remainder == 0 */
   width = m->width;
   numeyes = MTX_MIN(width, m->height);
   remainder = m->size - ((numeyes - 1) * (width + 1) + 1);
   
   /* Plot "eyes" with filler zeroes */
   p = m->data;
   while(numeyes) {
      *p = value;
      p++;
      skip = width;
      MTX_LOOP_ASSIGN(skip, p, 0);
      numeyes--;
   }
   
   /* Any remaining cells should be zero */
   MTX_LOOP_ASSIGN(remainder, p, 0);
   return(m);

}


static gmtxd *_gmtxd_eyes(gmtxd *m, doubledata value) {

   doubledata *p;
   dimensiona width;
   dimensiona skip;
   dimensiona remainder;
   dimensiona numeyes;
   
   /* Determine number of "eyes" to put in */
   width = m->width;
   numeyes = MTX_MIN(width, m->height);
   remainder = m->size - ((numeyes - 1) * (width + 1) + 1);
   
   /* Plot "eyes" with filler zeroes */
   p = m->data;
   while(numeyes) {
      *p = value;
      p++;
      skip = width;
      MTX_LOOP_ASSIGN(skip, p, 0);
      numeyes--;
   }
   
   /* Any remaining cells should be zero */
   MTX_LOOP_ASSIGN(remainder, p, 0);
   return(m);

}


static gmtxl *_gmtxl_new_eyes(dimensiona width, dimensiona height, longdata value) {

   gmtxl *mnew;
   
   if(!(mnew = _gmtxl_new(width, height))) return(NULL);
   return(_gmtxl_eyes(mnew, value));
   
}


static gmtxd *_gmtxd_new_eyes(dimensiona width, dimensiona height, doubledata value) {

   gmtxd *mnew;
   
   if(!(mnew = _gmtxd_new(width, height))) return(NULL);
   return(_gmtxd_eyes(mnew, value));
   
}


static gmtx *_gmtx_copy(gmtx *mdst, const gmtx *msrc) {

   /* Check size of dest matrix; reallocate if too small */
   if(!_gmtx_resize_using(mdst, msrc)) return(NULL);

   /* Copy the matrix contents */   
   jmemcpy(mdst->data, msrc->data, msrc->bufsize);
   return(mdst);

}


static gmtx *_gmtx_new_copy(const gmtx *m) {

   gmtx *mnew;
   
   if(!(mnew = _gmtx_new_using(m))) return(NULL);
   return(_gmtx_copy(mnew, m));

}


/*** Constructors and Destructors ***/
/* Matrix destructors, both internal and interface */


gmtx *gmtx_free(gmtx **m) {

   return(_gmtx_free(m));
   
}


gmtxl *gmtxl_free(gmtxl **m) {

   return(__gmtxlp _gmtx_free(__gmtxpp m));
   
}


gmtxd *gmtxd_free(gmtxd **m) {

   return(__gmtxdp _gmtx_free(__gmtxpp m));
   
}


/*** Constructors and Destructors ***/
/* Interface constructors */


gmtxl *gmtxl_resize(gmtxl *m, dimensiona width, dimensiona height)  { return(__gmtxlp _gmtx_resize(__gmtxp m, width, height)); }
gmtxd *gmtxd_resize(gmtxd *m, dimensiona width, dimensiona height)  { return(__gmtxdp _gmtx_resize(__gmtxp m, width, height)); }
gmtx  *gmtx_resize(gmtx *m, dimensiona width, dimensiona height)    { return(_gmtx_resize(m, width, height)); }
gmtxl *gmtxl_resize_using(gmtxl *mdst, const gmtxl *msrc) { return(__gmtxlp _gmtx_resize_using(__gmtxp mdst, __gmtxp msrc)); }
gmtxd *gmtxd_resize_using(gmtxd *mdst, const gmtxd *msrc) { return(__gmtxdp _gmtx_resize_using(__gmtxp mdst, __gmtxp msrc)); }
gmtx  *gmtx_resize_using(gmtx *mdst, const gmtx *msrc)    { return(_gmtx_resize_using(mdst, msrc)); }
gmtxl *gmtxl_reshape(gmtxl *m, dimensiona width, dimensiona height) { return(__gmtxlp _gmtx_reshape(__gmtxp m, width, height)); }
gmtxd *gmtxd_reshape(gmtxd *m, dimensiona width, dimensiona height) { return(__gmtxdp _gmtx_reshape(__gmtxp m, width, height)); }
gmtx  *gmtx_reshape(gmtx *m, dimensiona width, dimensiona height)   { return(_gmtx_reshape(m, width, height)); }
gmtxl *gmtxl_reshape_using(gmtxl *mdst, const gmtxl *msrc) { return(__gmtxlp _gmtx_reshape_using(__gmtxp mdst, __gmtxp msrc)); }
gmtxd *gmtxd_reshape_using(gmtxd *mdst, const gmtxd *msrc) { return(__gmtxdp _gmtx_reshape_using(__gmtxp mdst, __gmtxp msrc)); }
gmtx  *gmtx_reshape_using(gmtx *mdst, const gmtx *msrc)    { return(_gmtx_reshape_using(mdst, msrc)); }

gmtxl *gmtxl_new(dimensiona width, dimensiona height)         { return(_gmtxl_new(width, height)); }
gmtxd *gmtxd_new(dimensiona width, dimensiona height)         { return(_gmtxd_new(width, height)); }
gmtx  *gmtx_new(uword type, uword dtsize, dimensiona width, dimensiona height) { return(_gmtx_new(type, dtsize, width, height)); }
gmtxl *gmtxl_new_using(const gmtxl *m)          { return(__gmtxlp _gmtx_new_using(__gmtxp m)); }
gmtxd *gmtxd_new_using(const gmtxd *m)          { return(__gmtxdp _gmtx_new_using(__gmtxp m)); }
gmtx  *gmtx_new_using(const gmtx *m)            { return(_gmtx_new_using(m)); }
gmtxl *gmtxl_new_zero(dimensiona width, dimensiona height)    { return(_gmtxl_new_const(width, height, 0)); }
gmtxd *gmtxd_new_zero(dimensiona width, dimensiona height)    { return(_gmtxd_new_const(width, height, 0)); }
gmtxl *gmtxl_new_ones(dimensiona width, dimensiona height)    { return(_gmtxl_new_const(width, height, 1)); }
gmtxd *gmtxd_new_ones(dimensiona width, dimensiona height)    { return(_gmtxd_new_const(width, height, 1)); }
gmtxl *gmtxl_new_ident(dimensiona width, dimensiona height)   { return(_gmtxl_new_eyes(width, height, 1)); }
gmtxd *gmtxd_new_ident(dimensiona width, dimensiona height)   { return(_gmtxd_new_eyes(width, height, 1)); }
gmtxl *gmtxl_new_const(dimensiona width, dimensiona height, longdata value)    { return(_gmtxl_new_const(width, height, value)); }
gmtxd *gmtxd_new_const(dimensiona width, dimensiona height, doubledata value)  { return(_gmtxd_new_const(width, height, value)); }
gmtxl *gmtxl_new_eyes(dimensiona width, dimensiona height, longdata value)     { return(_gmtxl_new_eyes(width, height, value)); }
gmtxd *gmtxd_new_eyes(dimensiona width, dimensiona height, doubledata value)   { return(_gmtxd_new_eyes(width, height, value)); }
gmtxl *gmtxl_new_copy(const gmtxl *m) { return(__gmtxlp _gmtx_new_copy(__gmtxp m)); }
gmtxd *gmtxd_new_copy(const gmtxd *m) { return(__gmtxdp _gmtx_new_copy(__gmtxp m)); }
gmtx  *gmtx_new_copy(const gmtx *m)   { return(_gmtx_new_copy(m)); }


/*** Constructors and Destructors ***/
/* Matrix initialisation functions */


gmtxl *gmtxl_zero(gmtxl *m)      { return(_gmtxl_const(m, 0)); }
gmtxd *gmtxd_zero(gmtxd *m)      { return(_gmtxd_const(m, 0)); }
gmtx *gmtx_zero(gmtx *m) {

   if(MTX_IS_LONG(m))         return(__gmtxp _gmtxl_const(__gmtxlp m, 0));
   else if(MTX_IS_DOUBLE(m))  return(__gmtxp _gmtxd_const(__gmtxdp m, 0));
   return(NULL);
   
}


gmtxl *gmtxl_ones(gmtxl *m)      { return(_gmtxl_const(m, 1)); }
gmtxd *gmtxd_ones(gmtxd *m)      { return(_gmtxd_const(m, 1)); }
gmtx *gmtx_ones(gmtx *m) {

   if(MTX_IS_LONG(m))         return(__gmtxp _gmtxl_const(__gmtxlp m, 1));
   else if(MTX_IS_DOUBLE(m))  return(__gmtxp _gmtxd_const(__gmtxdp m, 1));
   return(NULL);
   
}


gmtxl *gmtxl_ident(gmtxl *m)      { return(_gmtxl_eyes(m, 1)); }
gmtxd *gmtxd_ident(gmtxd *m)      { return(_gmtxd_eyes(m, 1)); }
gmtx *gmtx_ident(gmtx *m) {

   if(MTX_IS_LONG(m))         return(__gmtxp _gmtxl_eyes(__gmtxlp m, 1));
   else if(MTX_IS_DOUBLE(m))  return(__gmtxp _gmtxd_eyes(__gmtxdp m, 1));
   return(NULL);
   
}


gmtxl *gmtxl_const(gmtxl *m, longdata value)     { return(_gmtxl_const(m, value)); }
gmtxd *gmtxd_const(gmtxd *m, doubledata value)   { return(_gmtxd_const(m, value)); }
gmtxl *gmtxl_eyes(gmtxl *m, longdata value)      { return(_gmtxl_eyes(m, value)); }
gmtxd *gmtxd_eyes(gmtxd *m, doubledata value)    { return(_gmtxd_eyes(m, value)); }
gmtxl *gmtxl_copy(gmtxl *mdst, const gmtxl *msrc) { return(__gmtxlp _gmtx_copy(__gmtxp mdst, __gmtxp msrc)); }
gmtxd *gmtxd_copy(gmtxd *mdst, const gmtxd *msrc) { return(__gmtxdp _gmtx_copy(__gmtxp mdst, __gmtxp msrc)); }
gmtx  *gmtx_copy(gmtx *mdst, const gmtx *msrc)    { return(_gmtx_copy(mdst, msrc)); }
