/* 

   libj - gmtx_det.c             Copyright (C) 2001-1998 Justin David Smith
   justins@chaos2.org            http://chaos2.org/

   Matrix determinant


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>
#include <libjmem.h>
#include <libjstr.h>



/*** Matrix Determinants ***/
/* Internal and interface code for calculating a matrix minor */



static inline void _gmtx_minor_size(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona width, dimensiona height) {

   if(*x >= 0 && *x < width) *w = width - 1;
   else                      *w = width, *x = width;
   if(*y >= 0 && *y <height) *h =height - 1;
   else                      *h =height, *y =height;

}


gmtx *_gmtx_minor_copy(gmtx *mdst, const gmtx *msrc, dimensiona x, dimensiona y) {

   char *src;
   char *dst;
   dimensiona prebufsize;
   dimensiona postbufsiz;
   dimensiona prenumrows;
   dimensiona postnmrows;
   dimensiona width;
   dimensiona height;
   uword dtsize;
   dimensiona w;
   dimensiona h;

   /* Get width and height of source */
   width = msrc->width;
   height= msrc->height;

   /* If x, y out of bounds, correct width, height accordingly. */
   _gmtx_minor_size(&x, &y, &w, &h, width, height);
   
   /* If 1xn or nx1 matrix that will be annihilated, return NULL matrix */
   if(w == 0 || h == 0) {
      mdst->type  =  msrc->type;
      mdst->dtsize=msrc->dtsize;
      return(gmtx_resize(mdst, 0, 0));
   }
   
   /* If no change in width, height, then simply copy the matrix */
   /* Also look for easy boundaries we can use the submatrix stuff for */
   if(x == 0) {
      if(y == 0)        return(gmtx_submatrix_copy(mdst, w, h, msrc, 1, 1));
      else if(y >= height - 1) 
                        return(gmtx_submatrix_copy(mdst, w, h, msrc, 1, 0));
   } else if(x >= width - 1) {
      if(y == 0)        return(gmtx_submatrix_copy(mdst, w, h, msrc, 0, 1));
      else if(y >= height - 1) 
                        return(gmtx_submatrix_copy(mdst, w, h, msrc, 0, 0));
   }

   /* Resize destination buffer */
   dtsize = msrc->dtsize;
   if(!gmtx_resize_using(mdst, msrc)) return(NULL);
   
   /* Select a minor method */
   src = msrc->data;
   dst = mdst->data;

   if(w == width) {
      /* Can copy entire rows at a time, calc prenumrows and postnmrows */
      prebufsize = dtsize * width;
      prenumrows = y;
      postnmrows = h - y - 1;
      while(prenumrows) {
         jmemcpy(dst, src, prebufsize);
         src += prebufsize;
         dst += prebufsize;
         prenumrows--;
      }
      src += prebufsize;
      dst += prebufsize;
      while(postnmrows) {
         jmemcpy(dst, src, prebufsize);
         src += prebufsize;
         dst += prebufsize;
         postnmrows--;
      }
      return(mdst);
   }
   
   /* A column is missing, calc pre- and post- buffers and skips */
   prebufsize = x * dtsize;
   postbufsiz = (w - x) * dtsize;

   if(h == height) {
      /* Copy all rows without skipping any, but each row is broken. */
      while(h) {
         jmemcpy(dst, src, prebufsize);
         src += prebufsize + dtsize;
         dst += prebufsize;
         jmemcpy(dst, src, postbufsiz);
         src += postbufsiz;
         dst += postbufsiz;
         h--;
      }
      return(mdst);
   }   
   
   /* Must skip a row _and_ a column. Calculate everything */
   prenumrows = y;
   postnmrows = h - y - 1;
   while(prenumrows) {
      jmemcpy(dst, src, prebufsize);
      src += prebufsize + dtsize;
      dst += prebufsize;
      jmemcpy(dst, src, postbufsiz);
      src += postbufsiz;
      dst += postbufsiz;
      prenumrows--;
   }
   src += (w + 1) * dtsize;
   dst += w * dtsize;
   while(postnmrows) {
      jmemcpy(dst, src, prebufsize);
      src += prebufsize + dtsize;
      dst += prebufsize;
      jmemcpy(dst, src, postbufsiz);
      src += postbufsiz;
      dst += postbufsiz;
      postnmrows--;
   }
   return(mdst);
   
}


static inline gmtx *_gmtx_new_minor(const gmtx *m, dimensiona x, dimensiona y) {
/* Matrix-New-Minor */

   gmtx *mtmp;
   
   if(!(mtmp = gmtx_new_using(m))) return(NULL);
   if(!_gmtx_minor_copy(mtmp, m, x, y)) return(gmtx_free(&mtmp));
   return(mtmp);
   
}


gmtxl *gmtxl_minor_copy(gmtxl *mdst, const gmtxl *msrc, dimensiona x, dimensiona y) {
/* Matrix-Minor-Copy */

   return(__gmtxlp _gmtx_minor_copy(__gmtxp mdst, __gmtxp msrc, x, y));

}


gmtxd *gmtxd_minor_copy(gmtxd *mdst, const gmtxd *msrc, dimensiona x, dimensiona y) {
/* Matrix-Minor-Copy */

   return(__gmtxdp _gmtx_minor_copy(__gmtxp mdst, __gmtxp msrc, x, y));

}


gmtx *gmtx_minor_copy(gmtx *mdst, const gmtx *msrc, dimensiona x, dimensiona y) {
/* Matrix-Minor-Copy */

   return(_gmtx_minor_copy(mdst, msrc, x, y));

}


gmtxl *gmtxl_new_minor(const gmtxl *m, dimensiona x, dimensiona y) {
/* Matrix-New-Minor */

   return(__gmtxlp _gmtx_new_minor(__gmtxp m, x, y));

}


gmtxd *gmtxd_new_minor(const gmtxd *m, dimensiona x, dimensiona y) {
/* Matrix-New-Minor */

   return(__gmtxdp _gmtx_new_minor(__gmtxp m, x, y));

}


gmtx *gmtx_new_minor(const gmtx *m, dimensiona x, dimensiona y) {
/* Matrix-New-Minor */

   return(_gmtx_new_minor(m, x, y));

}


/*** Matrix Determinant ***/
/* Interface and internal code for calculating a determinant */


void __gmtxl_best_row_col(dimensiona *row, longdata *rval, dimensiona *col, longdata *cval, const gmtxl *m) {
/* private-Matrix-Best-Row-Or-Column */

   dimensiona *sumzero;
   dimensiona *psr;
   dimensiona *psc;
   longdata *pr;
   longdata *pc;
   dimensiona width;
   dimensiona i;
   dimensiona j;
   
   /* Default return values */
   *row = 0;
   *rval= 0;
   *col = 0;
   *cval= 0;

   /* Get dimensions of matrix */
   width = m->width;
   
   /* Allocate an array to store counters. We're going to try to find
      the row with the most zeros, and remove it.  This should simplify
      calculations considerably for us, when using large matrices. */
   if(!(sumzero = (dimensiona *)malloc(sizeof(int) * (width << 1)))) return;

   /* Search for the row with the most numerical zeros */
   pr = MTX_XY_P(m, 0, 0);
   psr= sumzero;
   psc= sumzero + width;
   for(i = 0; i < width; i++) {
      *psr = 0;
      *psc = 0;
      j = width;
      pc = MTX_XY_P(m, i, 0);
      /* Count the zeros in row and column */
      while(j) {
         if(*pr == 0) (*psr)++;
         if(*pc == 0) (*psc)++;
         pr++;
         pc += width;
         j--;
      }
      psr++;
      psc++;
   }

   /* Find the maximum of the counters */   
   psr = sumzero;
   psc = sumzero + width;
   *rval = *psr;
   *cval = *psc;
   for(i = 1; i < width; i++) {
      psr++;
      psc++;
      if(*psr > *rval) *rval = *psr, *row = i;
      if(*psc > *cval) *cval = *psc, *col = i;
   }
   free(sumzero);
   return;
   
}
   

void __gmtxd_best_row_col(dimensiona *row, doubledata *rval, dimensiona *col, doubledata *cval, const gmtxd *m) {
/* private-Matrix-Best-Row-Or-Column */

   dimensiona *sumzero;
   dimensiona *psr;
   dimensiona *psc;
   doubledata *pr;
   doubledata *pc;
   dimensiona width;
   dimensiona i;
   dimensiona j;
   
   /* Default return values */
   *row = 0;
   *rval= 0;
   *col = 0;
   *cval= 0;

   /* Get dimensions of matrix */
   width = m->width;
   
   /* Allocate an array to store counters. We're going to try to find
      the row with the most zeros, and remove it.  This should simplify
      calculations considerably for us, when using large matrices. */
   if(!(sumzero = (dimensiona *)malloc(sizeof(int) * (width << 1)))) return;

   /* Search for the row with the most numerical zeros */
   pr = MTX_XY_P(m, 0, 0);
   psr= sumzero;
   psc= sumzero + width;
   for(i = 0; i < width; i++) {
      *psr = 0;
      *psc = 0;
      j = width;
      pc = MTX_XY_P(m, i, 0);
      /* Count the zeros in row and column */
      while(j) {
         if(*pr == 0) (*psr)++;
         if(*pc == 0) (*psc)++;
         pr++;
         pc += width;
         j--;
      }
      psr++;
      psc++;
   }

   /* Find the maximum of the counters */   
   psr = sumzero;
   psc = sumzero + width;
   *rval = *psr;
   *cval = *psc;
   for(i = 1; i < width; i++) {
      psr++;
      psc++;
      if(*psr > *rval) *rval = *psr, *row = i;
      if(*psc > *cval) *cval = *psc, *col = i;
   }
   free(sumzero);
   return;
   
}
   

longdata gmtxl_det_large(const gmtxl *m) {
/* Matrix-Determinant-Large */

   gmtxl *mtmp;
   longdata *p;
   longdata rval;
   longdata cval;
   longdata det;
   dimensiona row;
   dimensiona col;
   dimensiona width;
   dimensiona i;

   /* Setup */   
   width = m->width;
   det = 0;
   
   /* If width is 0 or 1x1, then we've reached the end of recursion */
   if(!MTX_IS_SQUARE(m)) return(0);
   if(MTX_IS_SIZE_0(m)) return(0);
   if(MTX_IS_SIZE_1x1(m)) return(*m->data);

   /* Allocate temporary matrix for minor storage */
   if(!(mtmp = __gmtxlp gmtx_new_using(__gmtxp m))) return(0);

   /* Determine row,col to run on */   
   __gmtxl_best_row_col(&row, &rval, &col, &cval, m);
   if(rval < cval) {
      /* Calculating via column col */
      p = MTX_XY_P(m, col, 0);
      for(i = 0; i < width; i++) {
         if(*p != 0) {
            if(!_gmtx_minor_copy(__gmtxp mtmp, __gmtxp m, col, i)) {
               gmtxl_free(&mtmp);
               return(0);
            }
            if((i + col) & 1) det -= *p * gmtxl_det_large(mtmp);
            else              det += *p * gmtxl_det_large(mtmp);
         }
         p += width;
      }
   } else {
      /* Calculating via row. */
      p = MTX_XY_P(m, 0, row);
      for(i = 0; i < width; i++) {
         if(*p != 0) {
            if(!_gmtx_minor_copy(__gmtxp mtmp, __gmtxp m, i, row)) {
               gmtxl_free(&mtmp);
               return(0);
            }
            if((i + row) & 1) det -= *p * gmtxl_det_large(mtmp);
            else              det += *p * gmtxl_det_large(mtmp);
         }
         p++;
      }
   }
   
   /* Release and return */
   gmtxl_free(&mtmp);
   return(det);
   
}


doubledata gmtxd_det_large(const gmtxd *m) {
/* Matrix-Determinant-Large */

   gmtxd *mtmp;
   doubledata *p;
   doubledata rval;
   doubledata cval;
   doubledata det;
   dimensiona row;
   dimensiona col;
   dimensiona width;
   dimensiona i;

   /* Setup */   
   width = m->width;
   det = 0;
   
   /* If width is 0 or 1x1, then we've reached the end of recursion */
   if(!MTX_IS_SQUARE(m)) return(0);
   if(MTX_IS_SIZE_0(m)) return(0);
   if(MTX_IS_SIZE_1x1(m)) return(*m->data);

   /* Allocate temporary matrix for minor storage */
   if(!(mtmp = __gmtxdp gmtx_new_using(__gmtxp m))) return(0);

   /* Determine row,col to run on */   
   __gmtxd_best_row_col(&row, &rval, &col, &cval, m);
   if(rval < cval) {
      /* Calculating via column col */
      p = MTX_XY_P(m, col, 0);
      for(i = 0; i < width; i++) {
         if(*p != 0) {
            if(!_gmtx_minor_copy(__gmtxp mtmp, __gmtxp m, col, i)) {
               gmtxd_free(&mtmp);
               return(0);
            }
            if((i + col) & 1) det -= *p * gmtxd_det_large(mtmp);
            else              det += *p * gmtxd_det_large(mtmp);
         }
         p += width;
      }
   } else {
      /* Calculating via row. */
      p = MTX_XY_P(m, 0, row);
      for(i = 0; i < width; i++) {
         if(*p != 0) {
            if(!_gmtx_minor_copy(__gmtxp mtmp, __gmtxp m, i, row)) {
               gmtxd_free(&mtmp);
               return(0);
            }
            if((i + row) & 1) det -= *p * gmtxd_det_large(mtmp);
            else              det += *p * gmtxd_det_large(mtmp);
         }
         p++;
      }
   }
   
   /* Release and return */
   gmtxd_free(&mtmp);
   return(det);
   
}


longdata gmtxl_det(const gmtxl *m) {
/* Matrix-Determinant */

   gmtxl *mtmp;
   longdata *p;
   longdata det;
   dimensiona width;
   dimensiona i;

   /* Setup */   
   width = m->width;
   det = 0;
   
   /* If width is 0 or 1, then we've reached the end of recursion */
   if(!MTX_IS_SQUARE(m)) return(0);
   if(MTX_IS_SIZE_0(m)) return(0);
   if(MTX_IS_SIZE_1x1(m)) return(*m->data);
   
   /* Allocate temporary matrix */
   if(!(mtmp = __gmtxlp gmtx_new_using(__gmtxp m))) return(0);

   /* Calculate determinant using row 0 */
   p = m->data;
   for(i = 0; i < width; i++) {
      if(*p != 0) {
         if(!_gmtx_minor_copy(__gmtxp mtmp, __gmtxp m, i, 0)) {
            gmtxl_free(&mtmp);
            return(0);
         }
         if(i & 1) det -= gmtxl_det(mtmp);
         else      det += gmtxl_det(mtmp);
      }
      p++;
   }
   
   /* Release and return */
   gmtxl_free(&mtmp);
   return(det);
   
}


doubledata gmtxd_det(const gmtxd *m) {
/* Matrix-Determinant */

   gmtxd *mtmp;
   doubledata *p;
   doubledata det;
   dimensiona width;
   dimensiona i;

   /* Setup */   
   width = m->width;
   det = 0;
   
   /* If width is 0 or 1, then we've reached the end of recursion */
   if(!MTX_IS_SQUARE(m)) return(0);
   if(MTX_IS_SIZE_0(m)) return(0);
   if(MTX_IS_SIZE_1x1(m)) return(*m->data);
   
   /* Allocate temporary matrix */
   if(!(mtmp = __gmtxdp gmtx_new_using(__gmtxp m))) return(0);

   /* Calculate determinant using row 0 */
   p = m->data;
   for(i = 0; i < width; i++) {
      if(*p != 0) {
         if(!_gmtx_minor_copy(__gmtxp mtmp, __gmtxp m, i, 0)) {
            gmtxd_free(&mtmp);
            return(0);
         }
         if(i & 1) det -= gmtxd_det(mtmp);
         else      det += gmtxd_det(mtmp);
      }
      p++;
   }
   
   /* Release and return */
   gmtxd_free(&mtmp);
   return(det);
   
}


