/* 

   libj - gmtx_gauss.c           Copyright (C) 2001-1998 Justin David Smith    
   justins@chaos2.org            http://chaos2.org/

   Gaussian spots


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>
#include <math.h>



/*** Matrix Gaussian ***/
/* Internal functions */


static gmtxd *_gmtxd_gauss_copy(gmtxd *m, dimensiona width, dimensiona height, doubledata xctr, doubledata yctr, doubledata xsig, doubledata ysig, doubledata xscl, doubledata yscl, doubledata zmin, doubledata zmax) {
/* Matrix-Gaussian-Copy

      width    -  Width of target matrix
      height   -  Height of target matrix
      xctr,yctr-  Center coordinate.  (0,0) is center of matrix.
                  Topright of matrix is quadrant 1, bottomright is quadrant 4.
      xsig,ysig-  Sigma value in X and Y directions
      xscl,yscl-  Number of units per cell in X direction
      zmin,zman-  "zero" and "one" on the Z scale
      
*/

   doubledata *p;
   doubledata x;
   doubledata y;
   doubledata xysig;
   doubledata zscale;
   doubledata halfwidth;
   doubledata halfheight;
   doubledata prefix;
   doubledata expt;
   dimensiona cols;
   
   /* Calculate arguments */
   xysig  = xsig * ysig;
   zscale = zmax - zmin;
   halfwidth = (doubledata)width / 2;
   halfheight= (doubledata)height/ 2;
   prefix    = 1 / (M_PI * 2 * xysig) * zscale;
   expt      = 1 / xysig;

   /* Resize target matrix */
   if(!gmtxd_resize(m, width, height)) return(NULL);
   
   /* Write Gaussian matrix */
   p = m->data;
   y = yctr - halfheight * yscl;
   while(height > 0) {
      cols = width;
      x = xctr - halfwidth * xscl;
      while(cols > 0) {
         *p = zmin + prefix * exp(expt * (x * x + y * y));
         x += xscl;
         cols--;
      }
      y += yscl;
      height--;
   }
   return(m);
   
}


gmtxd *gmtxd_gauss_copy(gmtxd *m, dimensiona width, dimensiona height, doubledata xctr, doubledata yctr, doubledata xsig, doubledata ysig, doubledata xscl, doubledata yscl, doubledata zmin, doubledata zmax) {

   return(_gmtxd_gauss_copy(m, width, height, xctr, yctr, xsig, ysig, xscl, yscl, zmin, zmax));
   
}


gmtxd *gmtxd_gauss_copy_norm(gmtxd *m, dimensiona width, dimensiona height, doubledata xscl, doubledata yscl) {

   return(_gmtxd_gauss_copy(m, width, height, 0, 0, 1, 1, xscl, yscl, 0, 1));
   
}


gmtxd *gmtxd_gauss_copy_circular(gmtxd *m, dimensiona width, dimensiona height, doubledata xctr, doubledata yctr, doubledata sig, doubledata scl, doubledata zmin, doubledata zmax) {

   return(_gmtxd_gauss_copy(m, width, height, xctr, yctr, sig, sig, scl, scl, zmin, zmax));
   
}


gmtxd *gmtxd_gauss(gmtxd *m, doubledata xctr, doubledata yctr, doubledata xsig, doubledata ysig, doubledata xscl, doubledata yscl, doubledata zmin, doubledata zmax) {

   return(_gmtxd_gauss_copy(m, m->width, m->height, xctr, yctr, xsig, ysig, xscl, yscl, zmin, zmax));
   
}


gmtxd *gmtxd_gauss_norm(gmtxd *m, doubledata xscl, doubledata yscl) {

   return(_gmtxd_gauss_copy(m, m->width, m->height, 0, 0, 1, 1, xscl, yscl, 0, 1));
   
}


gmtxd *gmtxd_gauss_circular(gmtxd *m, doubledata xctr, doubledata yctr, doubledata sig, doubledata scl, doubledata zmin, doubledata zmax) {

   return(_gmtxd_gauss_copy(m, m->width, m->height, xctr, yctr, sig, sig, scl, scl, zmin, zmax));
   
}


