/* 

   libj - gmtx_inv.c             Copyright (C) 2001-1998 Justin David Smith    
   justins@chaos2.org            http://chaos2.org/
    
   Matrix inversion
   

   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>



/*** Matrix Inversion ***/
/* Internal functions for matrix inversion */



static gmtxl *_gmtxl_cof_copy(gmtxl *mdst, const gmtxl *msrc) {

   longdata *p;
   gmtxl *mtmp;
   dimensiona width;
   dimensiona height;
   dimensiona negate;
   dimensiona i;
   dimensiona j;

   /* Get width and height */
   width = msrc->width;
   height= msrc->height;
   
   /* Resize target matrix accordingly, and allocate temp matrix */
   if(!(mtmp = gmtxl_new(width - 1, height - 1))) return(NULL);
   if(!gmtx_resize_using(__gmtxp mdst, __gmtxp msrc)) return(gmtxl_free(&mtmp));
   
   /* Calculate i,j'th cofactors */
   negate = 0;
   p = mdst->data;
   for(j = 0; j < height; j++) {
      for(i = 0; i < width; i++) {
         gmtx_minor_copy(__gmtxp mtmp, __gmtxp msrc, i, j);
         if(negate & 1) *p = -gmtxl_det(mtmp);
         else           *p =  gmtxl_det(mtmp);
         negate++;
         p++;
      }
   }
   
   /* Done */
   gmtxl_free(&mtmp);
   return(mdst);

}


static gmtxd *_gmtxd_cof_copy(gmtxd *mdst, const gmtxd *msrc) {

   doubledata *p;
   gmtxd *mtmp;
   dimensiona width;
   dimensiona height;
   dimensiona negate;
   dimensiona i;
   dimensiona j;

   /* Get width and height */
   width = msrc->width;
   height= msrc->height;
   
   /* Resize target matrix accordingly, and allocate temp matrix */
   if(!(mtmp = gmtxd_new(width - 1, height - 1))) return(NULL);
   if(!gmtx_resize_using(__gmtxp mdst, __gmtxp msrc)) return(gmtxd_free(&mtmp));
   
   /* Calculate i,j'th cofactors */
   negate = 0;
   p = mdst->data;
   for(j = 0; j < height; j++) {
      for(i = 0; i < width; i++) {
         gmtx_minor_copy(__gmtxp mtmp, __gmtxp msrc, i, j);
         if(negate & 1) *p = -gmtxd_det(mtmp);
         else           *p =  gmtxd_det(mtmp);
         negate++;
         p++;
      }
   }
   
   /* Done */
   gmtxd_free(&mtmp);
   return(mdst);

}


gmtx *_gmtx_cof_copy(gmtx *mdst, const gmtx *msrc) {

   if(MTX_IS_LONG(msrc))   return(__gmtxp _gmtxl_cof_copy(__gmtxlp mdst, __gmtxlp msrc));
   if(MTX_IS_DOUBLE(msrc)) return(__gmtxp _gmtxd_cof_copy(__gmtxdp mdst, __gmtxdp msrc));
   return(NULL);

}


static inline gmtx *_gmtx_cof(gmtx *m) {

   gmtx *mtmp;
   
   if(!(mtmp = gmtx_new_copy(m))) return(NULL);
   if(!_gmtx_cof_copy(m, mtmp)) return(gmtx_free(&mtmp));
   gmtx_free(&mtmp);
   return(m);
   
}


static inline gmtx *_gmtx_new_cof(const gmtx *m) {

   gmtx *mnew;
   
   if(!(mnew = gmtx_new_using(m))) return(NULL);
   return(_gmtx_cof_copy(mnew, m));

}


static gmtxl *_gmtxl_invert_copy(gmtxl *mdst, const gmtxl *msrc) {

   longdata *p;
   longdata det;
   dimensiona size;

   /* Check for singular matrix */
   det = gmtxl_det(msrc);
   if(det == 0) return(NULL);
   
   /* Allocate memory for square destination matrix */
   if(!gmtx_resize_using(__gmtxp mdst, __gmtxp msrc)) return(NULL);
   
   /* Calculate matrix inverse "the hard way" using cof matrices */
   _gmtxl_cof_copy(mdst, msrc);
   gmtx_transpose(__gmtxp mdst);
   size = msrc->size;
   p = mdst->data;
   while(size) {
      *p /= det;
      p++;
      size--;
   }
   return(mdst);

}


static gmtxd *_gmtxd_invert_copy(gmtxd *mdst, const gmtxd *msrc) {

   doubledata det;

   /* Check for singular matrix */
   det = gmtxd_det(msrc);
   if(det == 0) return(NULL);
   
   /* Allocate memory for square destination matrix */
   if(!gmtx_resize_using(__gmtxp mdst, __gmtxp msrc)) return(NULL);
   
   /* Calculate matrix inverse "the hard way" using cof matrices */
   _gmtxd_cof_copy(mdst, msrc);
   gmtx_transpose(__gmtxp mdst);
   gmtxd_submatrix_mult_scalar(mdst, 0, 0, mdst->width, mdst->height, 1 / det);
   return(mdst);

}


gmtx *_gmtx_invert_copy(gmtx *mdst, const gmtx *msrc) {

   if(MTX_IS_LONG(msrc))   return(__gmtxp _gmtxl_invert_copy(__gmtxlp mdst, __gmtxlp msrc));
   if(MTX_IS_DOUBLE(msrc)) return(__gmtxp _gmtxd_invert_copy(__gmtxdp mdst, __gmtxdp msrc));
   return(NULL);

}


static inline gmtx *_gmtx_invert(gmtx *m) {

   gmtx *mtmp;
   
   if(!(mtmp = gmtx_new_copy(m))) return(NULL);
   if(!_gmtx_invert_copy(m, mtmp)) return(gmtx_free(&mtmp));
   gmtx_free(&mtmp);
   return(m);
   
}


static inline gmtx *_gmtx_new_invert(const gmtx *m) {

   gmtx *mnew;
   
   if(!(mnew = gmtx_new_using(m))) return(NULL);
   return(_gmtx_invert_copy(mnew, m));

}


/*** Matrix Inversion ***/
/* Interface for cofactor and inversion routines */


gmtxl *gmtxl_cof(gmtxl *m) { return(__gmtxlp _gmtx_cof(__gmtxp m)); }
gmtxd *gmtxd_cof(gmtxd *m) { return(__gmtxdp _gmtx_cof(__gmtxp m)); }
gmtx  *gmtx_cof(gmtx *m)   { return(_gmtx_cof(m)); }
gmtxl *gmtxl_new_cof(const gmtxl *m) { return(__gmtxlp _gmtx_new_cof(__gmtxp m)); }
gmtxd *gmtxd_new_cof(const gmtxd *m) { return(__gmtxdp _gmtx_new_cof(__gmtxp m)); }
gmtx  *gmtx_new_cof(const gmtx *m)   { return(_gmtx_new_cof(m)); }
gmtxl *gmtxl_cof_copy(gmtxl *mdst, const gmtxl *msrc) { return(__gmtxlp _gmtx_cof_copy(__gmtxp mdst, __gmtxp msrc)); }
gmtxd *gmtxd_cof_copy(gmtxd *mdst, const gmtxd *msrc) { return(__gmtxdp _gmtx_cof_copy(__gmtxp mdst, __gmtxp msrc)); }
gmtx  *gmtx_cof_copy(gmtx *mdst, const gmtx *msrc)    { return(_gmtx_cof_copy(mdst, msrc)); }

gmtxl *gmtxl_invert(gmtxl *m) { return(__gmtxlp _gmtx_invert(__gmtxp m)); }
gmtxd *gmtxd_invert(gmtxd *m) { return(__gmtxdp _gmtx_invert(__gmtxp m)); }
gmtx  *gmtx_invert(gmtx *m)   { return(_gmtx_invert(m)); }
gmtxl *gmtxl_new_invert(const gmtxl *m) { return(__gmtxlp _gmtx_new_invert(__gmtxp m)); }
gmtxd *gmtxd_new_invert(const gmtxd *m) { return(__gmtxdp _gmtx_new_invert(__gmtxp m)); }
gmtx  *gmtx_new_invert(const gmtx *m)   { return(_gmtx_new_invert(m)); }
gmtxl *gmtxl_invert_copy(gmtxl *mdst, const gmtxl *msrc) { return(__gmtxlp _gmtx_invert_copy(__gmtxp mdst, __gmtxp msrc)); }
gmtxd *gmtxd_invert_copy(gmtxd *mdst, const gmtxd *msrc) { return(__gmtxdp _gmtx_invert_copy(__gmtxp mdst, __gmtxp msrc)); }
gmtx  *gmtx_invert_copy(gmtx *mdst, const gmtx *msrc)    { return(_gmtx_invert_copy(mdst, msrc)); }


