/* 

   libj - gmtx_math.c            Copyright (C) 2001-1998 Justin David Smith
   justins@chaos2.org            http://chaos2.org/
    
   Mathematical manipulation of matrices


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>



/*** Matrix Math ***/
/* Internal functions for submatrix addition/subtraction */


static inline void __gmtxl_submatrix_add(longdata *pdst, dimensiona skipdst, const longdata *psrc, dimensiona skipsrc, dimensiona w, dimensiona h) {

   dimensiona cols;
   
   while(h) {
      cols = w;
      while(cols) {
         *pdst += *psrc;
         psrc++;
         pdst++;
         cols--;
      }
      psrc += skipsrc;
      pdst += skipdst;
      h--;
   }
   return;

}


static inline void __gmtxd_submatrix_add(doubledata *pdst, dimensiona skipdst, const doubledata *psrc, dimensiona skipsrc, dimensiona w, dimensiona h) {

   dimensiona cols;
   
   while(h) {
      cols = w;
      while(cols) {
         *pdst += *psrc;
         psrc++;
         pdst++;
         cols--;
      }
      psrc += skipsrc;
      pdst += skipdst;
      h--;
   }
   return;

}


static gmtx *_gmtx_submatrix_add(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {
/* Submatrix-Add 

   If msrc is zero matrix, return mdst unmodified.  Otherwise, if bounding 
   rectangle is out of range for mdst,msrc then return mdst unmodified. If
   mdst is a zero matrix then nothing is calculated.  Matrices must be
   compatible with each other, since mdst's type is not modified. */

   /* Check boundaries */
   if(!MTX_ARE_COMPAT(mdst, msrc)) return(NULL);
   if(!mtx_adjust_bounds_xywh(&sx, &sy, &w, &h, msrc->width, msrc->height)) return(mdst);
   if(!mtx_adjust_bounds_xywh(&dx, &dy, &w, &h, mdst->width, mdst->height)) return(mdst);
   
   /* Begin adding data */
   if(MTX_IS_LONG(msrc))         __gmtxl_submatrix_add((longdata *)MTX_XY_P(mdst, dx, dy), mdst->width - w, \
                                                       (longdata *)MTX_XY_P(msrc, sx, sy), msrc->width - w, \
                                                       w, h);
   else if(MTX_IS_DOUBLE(msrc))  __gmtxd_submatrix_add((doubledata *)MTX_XY_P(mdst, dx, dy), mdst->width - w, \
                                                       (doubledata *)MTX_XY_P(msrc, sx, sy), msrc->width - w, \
                                                       w, h);
   else return(NULL);
   return(mdst);

}


static inline void __gmtxl_submatrix_sub(longdata *pdst, dimensiona skipdst, const longdata *psrc, dimensiona skipsrc, dimensiona w, dimensiona h) {

   dimensiona cols;
   
   while(h) {
      cols = w;
      while(cols) {
         *pdst -= *psrc;
         psrc++;
         pdst++;
         cols--;
      }
      psrc += skipsrc;
      pdst += skipdst;
      h--;
   }
   return;

}


static inline void __gmtxd_submatrix_sub(doubledata *pdst, dimensiona skipdst, const doubledata *psrc, dimensiona skipsrc, dimensiona w, dimensiona h) {

   dimensiona cols;
   
   while(h) {
      cols = w;
      while(cols) {
         *pdst -= *psrc;
         psrc++;
         pdst++;
         cols--;
      }
      psrc += skipsrc;
      pdst += skipdst;
      h--;
   }
   return;

}


static gmtx *_gmtx_submatrix_sub(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {

   /* Check boundaries */
   if(!MTX_ARE_COMPAT(mdst, msrc)) return(NULL);
   if(!mtx_adjust_bounds_xywh(&sx, &sy, &w, &h, msrc->width, msrc->height)) return(mdst);
   if(!mtx_adjust_bounds_xywh(&dx, &dy, &w, &h, mdst->width, mdst->height)) return(mdst);
   
   /* Begin subing data */
   if(MTX_IS_LONG(msrc))         __gmtxl_submatrix_sub((longdata *)MTX_XY_P(mdst, dx, dy), mdst->width - w, \
                                                       (longdata *)MTX_XY_P(msrc, sx, sy), msrc->width - w, \
                                                       w, h);
   else if(MTX_IS_DOUBLE(msrc))  __gmtxd_submatrix_sub((doubledata *)MTX_XY_P(mdst, dx, dy), mdst->width - w, \
                                                       (doubledata *)MTX_XY_P(msrc, sx, sy), msrc->width - w, \
                                                       w, h);
   else return(NULL);
   return(mdst);

}


static gmtxl *_gmtxl_submatrix_add_times(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy, longdata times) {

   longdata *psrc;
   longdata *pdst;
   dimensiona skipsrc;
   dimensiona skipdst;
   dimensiona cols;
   
   /* Check boundaries */
   if(!MTX_ARE_COMPAT(mdst, msrc)) return(NULL);
   if(!mtx_adjust_bounds_xywh(&sx, &sy, &w, &h, msrc->width, msrc->height)) return(mdst);
   if(!mtx_adjust_bounds_xywh(&dx, &dy, &w, &h, mdst->width, mdst->height)) return(mdst);
   
   /* Begin adding data */
   psrc = MTX_XY_P(msrc, sx, sy);
   pdst = MTX_XY_P(mdst, dx, dy);
   skipsrc = msrc->width - w;
   skipdst = mdst->width - w;
   while(h) {
      cols = w;
      while(cols) {
         *pdst += times * *psrc;
         psrc++;
         pdst++;
         cols--;
      }
      psrc += skipsrc;
      pdst += skipdst;
      h--;
   }
   return(mdst);

}


static gmtxd *_gmtxd_submatrix_add_times(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy, doubledata times) {

   doubledata *psrc;
   doubledata *pdst;
   dimensiona skipsrc;
   dimensiona skipdst;
   dimensiona cols;
   
   /* Check boundaries */
   if(!MTX_ARE_COMPAT(mdst, msrc)) return(NULL);
   if(!mtx_adjust_bounds_xywh(&sx, &sy, &w, &h, msrc->width, msrc->height)) return(mdst);
   if(!mtx_adjust_bounds_xywh(&dx, &dy, &w, &h, mdst->width, mdst->height)) return(mdst);
   
   /* Begin adding data */
   psrc = MTX_XY_P(msrc, sx, sy);
   pdst = MTX_XY_P(mdst, dx, dy);
   skipsrc = msrc->width - w;
   skipdst = mdst->width - w;
   while(h) {
      cols = w;
      while(cols) {
         *pdst += times * *psrc;
         psrc++;
         pdst++;
         cols--;
      }
      psrc += skipsrc;
      pdst += skipdst;
      h--;
   }
   return(mdst);

}


static gmtxl *_gmtxl_submatrix_add_scalar(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata scalar) {

   longdata *p;
   dimensiona skip;
   dimensiona cols;
   
   /* Check boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(m);
   
   /* Begin adding data */
   p = MTX_XY_P(m, x, y);
   skip = m->width - w;
   while(h) {
      cols = w;
      while(cols) {
         *p += scalar;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(m);

}


static gmtxd *_gmtxd_submatrix_add_scalar(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata scalar) {

   doubledata *p;
   dimensiona skip;
   dimensiona cols;
   
   /* Check boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(m);
   
   /* Begin adding data */
   p = MTX_XY_P(m, x, y);
   skip = m->width - w;
   while(h) {
      cols = w;
      while(cols) {
         *p += scalar;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(m);

}


static gmtxl *_gmtxl_submatrix_mult_scalar(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata scalar) {

   longdata *p;
   dimensiona skip;
   dimensiona cols;
   
   /* Check boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(m);
   
   /* Begin multing data */
   p = MTX_XY_P(m, x, y);
   skip = m->width - w;
   while(h) {
      cols = w;
      while(cols) {
         *p *= scalar;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(m);

}


static gmtxd *_gmtxd_submatrix_mult_scalar(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata scalar) {

   doubledata *p;
   dimensiona skip;
   dimensiona cols;
   
   /* Check boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(m);
   
   /* Begin multing data */
   p = MTX_XY_P(m, x, y);
   skip = m->width - w;
   while(h) {
      cols = w;
      while(cols) {
         *p *= scalar;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(m);

}


/*** Matrix Math ***/
/* Interface functions for submatrix addition/subtraction */


gmtxl *gmtxl_add(gmtxl *mdst, const gmtxl *msrc) {

   if(!MTX_SAME_SIZE(mdst, msrc)) return(NULL);
   return(__gmtxlp _gmtx_submatrix_add(__gmtxp mdst, 0, 0, msrc->width, msrc->height, __gmtxp msrc, 0, 0));
   
}


gmtxd *gmtxd_add(gmtxd *mdst, const gmtxd *msrc) {

   if(!MTX_SAME_SIZE(mdst, msrc)) return(NULL);
   return(__gmtxdp _gmtx_submatrix_add(__gmtxp mdst, 0, 0, msrc->width, msrc->height, __gmtxp msrc, 0, 0));
   
}


gmtx *gmtx_add(gmtx *mdst, const gmtx *msrc) {

   if(!MTX_SAME_SIZE(mdst, msrc)) return(NULL);
   return(_gmtx_submatrix_add(mdst, 0, 0, msrc->width, msrc->height, msrc, 0, 0));
   
}


gmtxl *gmtxl_sub(gmtxl *mdst, const gmtxl *msrc) {

   if(!MTX_SAME_SIZE(mdst, msrc)) return(NULL);
   return(__gmtxlp _gmtx_submatrix_sub(__gmtxp mdst, 0, 0, msrc->width, msrc->height, __gmtxp msrc, 0, 0));
   
}


gmtxd *gmtxd_sub(gmtxd *mdst, const gmtxd *msrc) {

   if(!MTX_SAME_SIZE(mdst, msrc)) return(NULL);
   return(__gmtxdp _gmtx_submatrix_sub(__gmtxp mdst, 0, 0, msrc->width, msrc->height, __gmtxp msrc, 0, 0));
   
}


gmtx *gmtx_sub(gmtx *mdst, const gmtx *msrc) {

   if(!MTX_SAME_SIZE(mdst, msrc)) return(NULL);
   return(_gmtx_submatrix_sub(mdst, 0, 0, msrc->width, msrc->height, msrc, 0, 0));
   
}


gmtxl *gmtxl_add_times(gmtxl *mdst, const gmtxl *msrc, longdata times) {

   if(!MTX_SAME_SIZE(mdst, msrc)) return(NULL);
   return(_gmtxl_submatrix_add_times(mdst, 0, 0, msrc->width, msrc->height, msrc, 0, 0, times));
   
}


gmtxd *gmtxd_add_times(gmtxd *mdst, const gmtxd *msrc, doubledata times) {

   if(!MTX_SAME_SIZE(mdst, msrc)) return(NULL);
   return(_gmtxd_submatrix_add_times(mdst, 0, 0, msrc->width, msrc->height, msrc, 0, 0, times));
   
}


gmtxl *gmtxl_add_scalar(gmtxl *m, longdata scalar) {

   return(_gmtxl_submatrix_add_scalar(m, 0, 0, m->width, m->height, scalar));
   
}


gmtxd *gmtxd_add_scalar(gmtxd *m, doubledata scalar) {

   return(_gmtxd_submatrix_add_scalar(m, 0, 0, m->width, m->height, scalar));
   
}


gmtxl *gmtxl_mult_scalar(gmtxl *m, longdata scalar) {

   return(_gmtxl_submatrix_mult_scalar(m, 0, 0, m->width, m->height, scalar));
   
}


gmtxd *gmtxd_mult_scalar(gmtxd *m, doubledata scalar) {

   return(_gmtxd_submatrix_mult_scalar(m, 0, 0, m->width, m->height, scalar));
   
}


gmtxl *gmtxl_submatrix_add(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy) {

   return(__gmtxlp _gmtx_submatrix_add(__gmtxp mdst, dx, dy, w, h, __gmtxp msrc, sx, sy));
   
}


gmtxd *gmtxd_submatrix_add(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy) {

   return(__gmtxdp _gmtx_submatrix_add(__gmtxp mdst, dx, dy, w, h, __gmtxp msrc, sx, sy));
   
}


gmtx *gmtx_submatrix_add(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {

   return(_gmtx_submatrix_add(mdst, dx, dy, w, h, msrc, sx, sy));
   
}


gmtxl *gmtxl_submatrix_sub(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy) {

   return(__gmtxlp _gmtx_submatrix_sub(__gmtxp mdst, dx, dy, w, h, __gmtxp msrc, sx, sy));
   
}


gmtxd *gmtxd_submatrix_sub(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy) {

   return(__gmtxdp _gmtx_submatrix_sub(__gmtxp mdst, dx, dy, w, h, __gmtxp msrc, sx, sy));
   
}


gmtx *gmtx_submatrix_sub(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {

   return(_gmtx_submatrix_sub(mdst, dx, dy, w, h, msrc, sx, sy));
   
}


gmtxl *gmtxl_submatrix_add_times(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy, longdata times) {

   return(_gmtxl_submatrix_add_times(mdst, dx, dy, w, h, msrc, sx, sy, times));
   
}


gmtxd *gmtxd_submatrix_add_times(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy, doubledata times) {

   return(_gmtxd_submatrix_add_times(mdst, dx, dy, w, h, msrc, sx, sy, times));
   
}


gmtxl *gmtxl_submatrix_add_scalar(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata scalar) {

   return(_gmtxl_submatrix_add_scalar(m, x, y, w, h, scalar));
   
}


gmtxd *gmtxd_submatrix_add_scalar(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata scalar) {

   return(_gmtxd_submatrix_add_scalar(m, x, y, w, h, scalar));
   
}


gmtxl *gmtxl_submatrix_mult_scalar(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata scalar) {

   return(_gmtxl_submatrix_mult_scalar(m, x, y, w, h, scalar));
   
}


gmtxd *gmtxd_submatrix_mult_scalar(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata scalar) {

   return(_gmtxd_submatrix_mult_scalar(m, x, y, w, h, scalar));
   
}


/*** Matrix summation ***/
/* Interface and internal code for summing a matrix */


static longdata _gmtxl_submatrix_sum(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* internal-Matrix-Sum */

   longdata *p;
   longdata sum;
   dimensiona skip;
   dimensiona cols;

   /* Check boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(0);
   
   /* Calculate sum */
   skip = m->width - w;
   p    = MTX_XY_P(m, x, y);
   sum  = 0;
   while(h) {
      cols = w;
      while(cols) {
         sum += *p;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(sum);

}


static doubledata _gmtxd_submatrix_sum(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* internal-Matrix-Sum */

   doubledata *p;
   doubledata sum;
   dimensiona skip;
   dimensiona cols;

   /* Check boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(0);
   
   /* Calculate sum */
   skip = m->width - w;
   p    = MTX_XY_P(m, x, y);
   sum  = 0;
   while(h) {
      cols = w;
      while(cols) {
         sum += *p;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(sum);

}


longdata gmtxl_sum(const gmtxl *m) {
/* Matrix-Sum */

   return(_gmtxl_submatrix_sum(m, 0, 0, m->width, m->height));
   
}


doubledata gmtxd_sum(const gmtxd *m) {
/* Matrix-Sum */

   return(_gmtxd_submatrix_sum(m, 0, 0, m->width, m->height));
   
}


longdata gmtxl_submatrix_sum(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Submatrix-Sum */

   return(_gmtxl_submatrix_sum(m, x, y, w, h));
   
}


doubledata gmtxd_submatrix_sum(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Submatrix-Sum */

   return(_gmtxd_submatrix_sum(m, x, y, w, h));
   
}


/*** Matrix min/max/mean ***/
/* Internal and interface submatrix min/max/mean code */


static longdata _gmtxl_submatrix_min(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* internal-Submatrix-Min */

   longdata *p;
   longdata min;
   dimensiona skip;
   dimensiona cols;

   /* Adjust boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(0);

   /* Find minimum */
   p   = m->data;
   skip= m->width - w;
   min = *p;
   while(h) {
      cols = w;
      while(cols) {
         if(*p < min) min = *p;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(min);
   
}


static longdata _gmtxd_submatrix_min(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Submatrix-Min */

   doubledata *p;
   doubledata min;
   dimensiona skip;
   dimensiona cols;

   /* Adjust boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(0);

   /* Find minimum */
   p   = m->data;
   skip= m->width - w;
   min = *p;
   while(h) {
      cols = w;
      while(cols) {
         if(*p < min) min = *p;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(min);
   
}


static longdata _gmtxl_submatrix_max(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Submatrix-Max */

   longdata *p;
   longdata max;
   dimensiona skip;
   dimensiona cols;

   /* Adjust boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(0);

   /* Find minimum */
   p   = m->data;
   skip= m->width - w;
   max = *p;
   while(h) {
      cols = w;
      while(cols) {
         if(*p > max) max = *p;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(max);
   
}


static doubledata _gmtxd_submatrix_max(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Submatrix-Max */

   doubledata *p;
   doubledata max;
   dimensiona skip;
   dimensiona cols;

   /* Adjust boundaries */
   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(0);

   /* Find minimum */
   p   = m->data;
   skip= m->width - w;
   max = *p;
   while(h) {
      cols = w;
      while(cols) {
         if(*p > max) max = *p;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(max);
   
}


static inline longdata _gmtxl_submatrix_mean(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Submatrix-Mean */

   return(_gmtxl_submatrix_sum(m, x, y, w, h) / m->size);

}


static inline doubledata _gmtxd_submatrix_mean(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Submatrix-Mean */

   return(_gmtxd_submatrix_sum(m, x, y, w, h) / m->size);

}


longdata gmtxl_min(const gmtxl *m) {
/* Matrix-Min */

   return(_gmtxl_submatrix_min(m, 0, 0, m->width, m->height));
   
}


doubledata gmtxd_min(const gmtxd *m) {
/* Matrix-Min */

   return(_gmtxd_submatrix_min(m, 0, 0, m->width, m->height));
   
}


longdata gmtxl_max(const gmtxl *m) {
/* Matrix-max */

   return(_gmtxl_submatrix_max(m, 0, 0, m->width, m->height));
   
}


doubledata gmtxd_max(const gmtxd *m) {
/* Matrix-max */

   return(_gmtxd_submatrix_max(m, 0, 0, m->width, m->height));
   
}


longdata gmtxl_mean(const gmtxl *m) {
/* Matrix-mean */

   return(_gmtxl_submatrix_mean(m, 0, 0, m->width, m->height));
   
}


doubledata gmtxd_mean(const gmtxd *m) {
/* Matrix-mean */

   return(_gmtxd_submatrix_mean(m, 0, 0, m->width, m->height));
   
}


longdata gmtxl_submatrix_min(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Matrix-min */

   return(_gmtxl_submatrix_min(m, x, y, w, h));
   
}


doubledata gmtxd_submatrix_min(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Matrix-min */

   return(_gmtxd_submatrix_min(m, x, y, w, h));
   
}


longdata gmtxl_submatrix_max(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Matrix-max */

   return(_gmtxl_submatrix_max(m, x, y, w, h));
   
}


doubledata gmtxd_submatrix_max(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Matrix-max */

   return(_gmtxd_submatrix_max(m, x, y, w, h));
   
}


longdata gmtxl_submatrix_mean(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Matrix-mean */

   return(_gmtxl_submatrix_mean(m, x, y, w, h));
   
}


doubledata gmtxd_submatrix_mean(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* Matrix-mean */

   return(_gmtxd_submatrix_mean(m, x, y, w, h));
   
}


/*** Matrix Negation ***/
/* Interface & implementation */


static gmtxl *_gmtxl_submatrix_negate(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   longdata *p;
   dimensiona cols;
   dimensiona skip;

   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(m);
   
   p = MTX_XY_P(m, x, y);
   skip = m->width - w;
   while(h) {
      cols = w;
      while(cols) {
         *p = -*p;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(m);

}


static gmtxd *_gmtxd_submatrix_negate(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   doubledata *p;
   dimensiona cols;
   dimensiona skip;

   if(!mtx_adjust_bounds_xywh(&x, &y, &w, &h, m->width, m->height)) return(m);
   
   p = MTX_XY_P(m, x, y);
   skip = m->width - w;
   while(h) {
      cols = w;
      while(cols) {
         *p = -*p;
         p++;
         cols--;
      }
      p += skip;
      h--;
   }
   return(m);

}


static inline gmtx *_gmtx_submatrix_negate(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   if(MTX_IS_LONG(m))   return(__gmtxp _gmtxl_submatrix_negate(__gmtxlp m, x, y, w, h));
   if(MTX_IS_DOUBLE(m)) return(__gmtxp _gmtxd_submatrix_negate(__gmtxdp m, x, y, w, h));
   return(NULL);
   
}


gmtxl *gmtxl_negate(gmtxl *m) { return(_gmtxl_submatrix_negate(m, 0, 0, m->width, m->height)); }
gmtxd *gmtxd_negate(gmtxd *m) { return(_gmtxd_submatrix_negate(m, 0, 0, m->width, m->height)); }
gmtx  *gmtx_negate(gmtx *m)   { return(_gmtx_submatrix_negate(m, 0, 0, m->width, m->height)); }
gmtxl *gmtxl_submatrix_negate(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) { return(_gmtxl_submatrix_negate(m, x, y, w, h)); }
gmtxd *gmtxd_submatrix_negate(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) { return(_gmtxd_submatrix_negate(m, x, y, w, h)); }
gmtx  *gmtx_submatrix_negate(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h)   { return(_gmtx_submatrix_negate(m, x, y, w, h)); }
