/* 

   libj - gmtx_mult.c            Copyright (C) 2001-1998 Justin David Smith
   justins@chaos2.org            http://chaos2.org/
    
   Matrix multiplication


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>



/*** Matrix Multiplication ***/
/* Internal functions */



static inline void __gmtxl_multiply_copy_data(longdata *pc, const longdata *pa, const longdata *pb, dimensiona width, dimensiona height, dimensiona scan) {

   const longdata *pascan;
   const longdata *pbscan;
   dimensiona cols;

   /* Start initialising the new matrix */   
   while(height) {
      pascan = pa;
      pbscan = pb;
      *pc    = 0;
      cols   = scan;
      while(cols) {
         *pc += *pascan * *pbscan;
         pascan++;
         pbscan += width;
         cols--;
      }
      pa += scan;
      pb++;
      pc++;
      height--;
   }
   return;

}


static inline void __gmtxd_multiply_copy_data(doubledata *pc, const doubledata *pa, const doubledata *pb, dimensiona width, dimensiona height, dimensiona scan) {

   const doubledata *pascan;
   const doubledata *pbscan;
   dimensiona cols;

   /* Start initialising the new matrix */   
   while(height) {
      pascan = pa;
      pbscan = pb;
      *pc    = 0;
      cols   = scan;
      while(cols) {
         *pc += *pascan * *pbscan;
         pascan++;
         pbscan += width;
         cols--;
      }
      pa += scan;
      pb++;
      pc++;
      height--;
   }
   return;

}


static gmtx *_gmtx_multiply_copy(gmtx *c, const gmtx *a, const gmtx *b) {

   dimensiona width;
   dimensiona height;
   dimensiona scan;

   /* Make sure a, b are compatible */
   /* Remember, a->width == b->height */
   if(!MTX_ARE_COMPAT(a, b)) return(NULL);

   /* Get width, height of new matrix */
   width = b->width;
   height= a->height;
   scan  = a->width;
   if(scan != b->height) return(NULL);
   
   /* Reallocate target matrix to new size */
   if(!gmtx_resize(c, width, height)) return(NULL);
   
   /* Start initialising the new matrix */   
   if(MTX_IS_LONG(a))         __gmtxl_multiply_copy_data((longdata *)c->data, (longdata *)a->data, (longdata *)b->data, width, height, scan);
   else if(MTX_IS_DOUBLE(a))  __gmtxd_multiply_copy_data((doubledata *)c->data, (doubledata *)a->data, (doubledata *)b->data, width, height, scan);
   else return(NULL);
   return(c);

}


static gmtx *_gmtx_new_multiply(const gmtx *a, const gmtx *b) {

   gmtx *mtmp;
   
   if(!(mtmp = gmtx_new_using(a))) return(NULL);
   return(_gmtx_multiply_copy(mtmp, a, b));

}


static gmtx *_gmtx_multiply(gmtx *a, const gmtx *b) {

   gmtx *mtmp;
   
   if(!(mtmp = gmtx_new_copy(a))) return(NULL);
   return(_gmtx_multiply_copy(a, mtmp, b));

}


/*** Matrix Multiplication ***/
/* Interface functions */


gmtxl *gmtxl_multiply_copy(gmtxl *c, const gmtxl *a, const gmtxl *b) {

   return(__gmtxlp _gmtx_multiply_copy(__gmtxp c, __gmtxp a, __gmtxp b));
   
}


gmtxd *gmtxd_multiply_copy(gmtxd *c, const gmtxd *a, const gmtxd *b) {

   return(__gmtxdp _gmtx_multiply_copy(__gmtxp c, __gmtxp a, __gmtxp b));
   
}


gmtx *gmtx_multiply_copy(gmtx *c, const gmtx *a, const gmtx *b) {

   return(_gmtx_multiply_copy(c, a, b));
   
}


gmtxl *gmtxl_new_multiply(const gmtxl *a, const gmtxl *b) {

   return(__gmtxlp _gmtx_new_multiply(__gmtxp a, __gmtxp b));
   
}


gmtxd *gmtxd_new_multiply(const gmtxd *a, const gmtxd *b) {

   return(__gmtxdp _gmtx_new_multiply(__gmtxp a, __gmtxp b));
   
}


gmtx *gmtx_new_multiply(const gmtx *a, const gmtx *b) {

   return(_gmtx_new_multiply(a, b));
   
}


gmtxl *gmtxl_multiply(gmtxl *a, const gmtxl *b) {

   return(__gmtxlp _gmtx_multiply(__gmtxp a, __gmtxp b));
   
}


gmtxd *gmtxd_multiply(gmtxd *a, const gmtxd *b) {

   return(__gmtxdp _gmtx_multiply(__gmtxp a, __gmtxp b));
   
}


gmtx *gmtx_multiply(gmtx *a, const gmtx *b) {

   return(_gmtx_multiply(a, b));
   
}


