/* 

   libj - gmtx_subm.c            Copyright (C) 2001-1998 Justin David Smith    
   justins@chaos2.org            http://chaos2.org/
    
   Submatrix extraction


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>
#include <libjmem.h>
#include <libjstr.h>



/*** Submatrix code ***/
/* XYXY and XYWH internal handlers */


static inline void _gmtx_fix_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2) {
/* internal-Fix-XYXY */

   if(*x1 > *x2) MTX_SWAP(*x1, *x2);
   if(*y1 > *y2) MTX_SWAP(*y1, *y2);
   return;

}


static inline void _gmtx_fix_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h) {
/* internal-Fix-XYWH */
   
   if(*w < 0) *x += *w, *w = -*w;
   if(*h < 0) *y += *h, *h = -*h;
   return;
   
}


static inline void _gmtx_xyxy_to_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona x1, dimensiona y1, dimensiona x2, dimensiona y2) {
/* internal-XYXY-To-XYWH-Conversion */

   *x = x1;
   *y = y1;
   *w = x2 - x1 + 1;
   *h = y2 - y1 + 1;
   return;

}


static inline void _gmtx_xywh_to_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {
/* internal-XYWH-To-XYXY-Conversion */

   *x1 = x;
   *y1 = y;
   *x2 = x + w - 1;
   *y2 = y + h - 1;
   return;
   
}


static inline dimensiona _gmtx_adjust_bounds_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona width, dimensiona height) {
/* internal-Adjust-Bounds-XYXY */

   /* Left-top clipping */
   MTX_LEFT_CLIP_XYXY(  *x1, *x2, width);
   MTX_TOP_CLIP_XYXY(   *y1, *y2, height);
   /* Right-bottom clipping */
   MTX_RIGHT_CLIP_XYXY( *x1, *x2, width);
   MTX_BOTTOM_CLIP_XYXY(*y1, *y2, height);
   /* Return 0 if box was completely out of bounds */
   return(MTX_XYXY_IN_BOUNDS(*x1, *y1, *x2, *y2, width, height));
   
}


static inline dimensiona _gmtx_adjust_bounds_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona width, dimensiona height) {
/* internal-Adjust-Bounds-XYWH */

   /* Left-top clipping */
   MTX_LEFT_CLIP_XYWH(  *x, *w, width);
   MTX_TOP_CLIP_XYWH(   *y, *h, height);
   /* Right-bottom clipping */
   MTX_RIGHT_CLIP_XYWH( *x, *w, width);
   MTX_BOTTOM_CLIP_XYWH(*y, *h, height);
   /* Return 0 if box was completely out of bounds */
   return(MTX_XYWH_IN_BOUNDS(*x, *y, *w, *h, width, height));

}


static inline dimensiona _gmtx_shift_bounds_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona width, dimensiona height) {
/* internal-Shift-Bounds-XYXY */

   /* Shift to the right and down */
   MTX_LEFT_SHIFT_XYXY(    *x1, *x2, width);
   MTX_TOP_SHIFT_XYXY(     *y1, *y2, height);
   /* Shift to the left and up, if needed */
   MTX_RIGHT_SHIFT_XYXY(   *x1, *x2, width);
   MTX_BOTTOM_SHIFT_XYXY(  *y1, *y2, height);
   /* Clip left and top, if we're still too large.  If this is required,
      then the bounding rect was larger than the matrix */
   MTX_LEFT_CLIP_XYXY(     *x1, *x2, width);
   MTX_TOP_CLIP_XYXY(      *y1, *y2, height);
   /* Check boundaries */
   return(MTX_XYXY_IN_BOUNDS(*x1, *y1, *x2, *y2, width, height));

}


static inline dimensiona _mtx_shift_bounds_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona width, dimensiona height) {
/* internal-Shift-Bounds-XYWH */

   /* Shift to the right and down */
   MTX_LEFT_SHIFT_XYWH(    *x, *w, width);
   MTX_TOP_SHIFT_XYWH(     *y, *h, height);
   /* Shift to the left and up, if needed */
   MTX_RIGHT_SHIFT_XYWH(   *x, *w, width);
   MTX_BOTTOM_SHIFT_XYWH(  *y, *h, height);
   /* Clip left and top, if we're still too large.  If this is required,
      then the bounding rect was larger than the matrix */
   MTX_LEFT_CLIP_XYWH(     *x, *w, width);
   MTX_TOP_CLIP_XYWH(      *y, *h, height);
   /* Check boundaries */
   return(MTX_XYWH_IN_BOUNDS(*x, *y, *w, *h, width, height));

}


/*** Submatrix code ***/
/* XYXY and XYWH interface handlers */


void mtx_fix_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2) { _gmtx_fix_xyxy(x1, y1, x2, y2); }
void mtx_fix_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h)     { _gmtx_fix_xywh(x, y, w, h); }
void mtx_xyxy_to_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona x1, dimensiona y1, dimensiona x2, dimensiona y2) { _gmtx_xyxy_to_xywh(x, y, w, h, x1, y1, x2, y2); }
void mtx_xywh_to_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona x, dimensiona y, dimensiona w, dimensiona h) { _gmtx_xywh_to_xyxy(x1, y1, x2, y2, x, y, w, h); }
bool  mtx_adjust_bounds_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona width, dimensiona height) { return(_gmtx_adjust_bounds_xyxy(x1, y1, x2, y2, width, height)); }
bool  mtx_adjust_bounds_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona width, dimensiona height)     { return(_gmtx_adjust_bounds_xyxy(x, y, w, h, width, height)); }
bool  mtx_shift_bounds_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona width, dimensiona height) { return(_gmtx_shift_bounds_xyxy(x1, y1, x2, y2, width, height)); }
bool  mtx_shift_bounds_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona width, dimensiona height)     { return(_gmtx_shift_bounds_xyxy(x, y, w, h, width, height)); }


/*** Submatrix Code ***/
/* Internal copy functions */


static gmtx *__gmtx_submatrix_copy_data(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {
/* Submatrix-Copy-Data 

   Copies data in source matrix to destination matrix.  */

   void *src;
   void *dst;
   dimensiona wsrc;
   dimensiona wdst;
   dimensiona size;

   /* Copy the source submatrix to destination */
   wsrc= msrc->rowsize;
   wdst= mdst->rowsize;
   src = MTX_XY_P(msrc, sx, sy);
   dst = MTX_XY_P(mdst, dx, dy);
   size= msrc->rowsize;
   while(h) {
      jmemcpy(dst, src, size);
      msrc += wsrc;
      mdst += wdst;
      h--;
   }
   return(mdst); 

}


static gmtx *_gmtx_submatrix_copy(gmtx *mdst, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {
/* Submatrix-Copy 

   Copies submatrix of msrc into the entirety of mdst.  If msrc is NULL matrix,
   then mdst is set to NULL matrix.  If bounding rect is outsize of msrc, then
   mdst is set to NULL matrix. Destination's type and dtsize are modified. */

   /* Modify destination type, dtsize */
   mdst->type  =  msrc->type;
   mdst->dtsize=msrc->dtsize;

   /* Correct out of bound errors and NULL matrix 
      gmtx_adjust_bounds will return false automatically if msrc is
      a 0x0 matrix */
   if(!_gmtx_adjust_bounds_xywh(&sx, &sy, &w, &h, msrc->width, msrc->height)) {
      return(gmtx_resize(mdst, 0, 0));
   }
   
   /* Check size of dest matrix; reallocate if too small */
   if(!gmtx_resize(mdst, w, h)) return(NULL);
   
   /* Copy the source submatrix */
   return(__gmtx_submatrix_copy_data(mdst, 0, 0, w, h, msrc, sx, sy));

}


static gmtx *_gmtx_submatrix_copy_to(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {
/* Submatrix-Copy-To

   Copies submatrix of msrc into mdst.  If msrc is NULL matrix,
   then mdst is returned, unmodified.  The bounding rectangle is clipped
   to both msrc and mdst; mdst is NOT resized.  If bounding rectangle is
   invalid, then mdst is unmodified. Matrices must be compatible */

   /* Make sure matrices are compatible */
   if(!MTX_ARE_COMPAT(mdst, msrc)) return(NULL);

   /* Correct out of bound errors and NULL matrix */
   if(!_gmtx_adjust_bounds_xywh(&sx, &sy, &w, &h, msrc->width, msrc->height)) return(mdst);
   if(!_gmtx_adjust_bounds_xywh(&dx, &dy, &w, &h, mdst->width, mdst->height)) return(mdst);
   
   /* Copy the source submatrix into destination */
   return(__gmtx_submatrix_copy_data(mdst, dx, dy, w, h, msrc, sx, sy));

}


/*** Submatrix code ***/
/* Internal submatrix clearing functions */


static gmtxl *_gmtxl_submatrix_const(gmtxl *m, dimensiona sx, dimensiona sy, dimensiona w, dimensiona h, longdata value) {

   longdata *p;
   dimensiona cols;
   dimensiona skip;
   
   /* Check argument boundaries */
   if(!_gmtx_adjust_bounds_xywh(&sx, &sy, &w, &h, m->width, m->height)) return(m);
   
   /* Modify submatrix */
   p    = MTX_XY_P(m, sx, sy);
   skip = m->width - w;
   MTX_LOOP2_ASSIGN(h, w, p, skip, cols, 0);
   return(m);

}


static gmtxd *_gmtxd_submatrix_const(gmtxd *m, dimensiona sx, dimensiona sy, dimensiona w, dimensiona h, doubledata value) {

   doubledata *p;
   dimensiona cols;
   dimensiona skip;
   
   /* Check argument boundaries */
   if(!_gmtx_adjust_bounds_xywh(&sx, &sy, &w, &h, m->width, m->height)) return(m);
   
   /* Modify submatrix */
   p    = MTX_XY_P(m, sx, sy);
   skip = m->width - w;
   MTX_LOOP2_ASSIGN(h, w, p, skip, cols, 0);
   return(m);

}


static gmtxl *_gmtxl_submatrix_eyes(gmtxl *m, dimensiona sx, dimensiona sy, dimensiona w, dimensiona h, longdata value) {

   longdata *p;
   dimensiona skip;
   dimensiona rows;
   dimensiona cols;
   
   /* Check argument boundaries */
   if(!_gmtx_adjust_bounds_xywh(&sx, &sy, &w, &h, m->width, m->height)) return(m);
   
   /* Plot "eyes" with filler zeroes */
   p    = MTX_XY_P(m, sx, sy);
   skip = m->width - w;
   rows = 0;
   while(rows < h) { 
      cols = 0;
      while(cols < w) {
         if(rows == cols) *p = value;
         else *p = 0;
         p++;
         cols--;
      }
      p += skip;
      rows--;
   }
   return(m);

}


static gmtxd *_gmtxd_submatrix_eyes(gmtxd *m, dimensiona sx, dimensiona sy, dimensiona w, dimensiona h, doubledata value) {

   doubledata *p;
   dimensiona skip;
   dimensiona rows;
   dimensiona cols;
   
   /* Check argument boundaries */
   if(!_gmtx_adjust_bounds_xywh(&sx, &sy, &w, &h, m->width, m->height)) return(m);
   
   /* Plot "eyes" with filler zeroes */
   p    = MTX_XY_P(m, sx, sy);
   skip = m->width - w;
   rows = 0;
   while(rows < h) { 
      cols = 0;
      while(cols < w) {
         if(rows == cols) *p = value;
         else *p = 0;
         p++;
         cols--;
      }
      p += skip;
      rows--;
   }
   return(m);

}


/*** Submatrix code ***/
/* basic interface functions */


gmtxl *gmtxl_submatrix_copy(gmtxl *mdst, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy) {

   /* Copy and reallocate destination matrix */
   return(__gmtxlp _gmtx_submatrix_copy(__gmtxp mdst, w, h, __gmtxp msrc, sx, sy));

}


gmtxd *gmtxd_submatrix_copy(gmtxd *mdst, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy) {

   /* Copy and reallocate destination matrix */
   return(__gmtxdp _gmtx_submatrix_copy(__gmtxp mdst, w, h, __gmtxp msrc, sx, sy));

}


gmtx *gmtx_submatrix_copy(gmtx *mdst, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {

   /* Copy and reallocate destination matrix */
   return(_gmtx_submatrix_copy(mdst, w, h, msrc, sx, sy));

}


gmtxl *gmtxl_submatrix_copy_to(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy) {

   /* Copy into destination matrix */
   return(__gmtxlp _gmtx_submatrix_copy_to(__gmtxp mdst, dx, dy, w, h, __gmtxp msrc, sx, sy));

}


gmtxd *gmtxd_submatrix_copy_into(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy) {

   /* Copy into destination matrix */
   return(__gmtxdp _gmtx_submatrix_copy_to(__gmtxp mdst, dx, dy, w, h, __gmtxp msrc, sx, sy));

}


gmtx *gmtx_submatrix_copy_into(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy) {

   /* Copy into destination matrix */
   return(_gmtx_submatrix_copy_to(mdst, dx, dy, w, h, msrc, sx, sy));

}


/*** Submatrix constructors ***/
/* Interface and internal code for submatrix constructors */


static gmtx *_gmtx_new_submatrix(const gmtx *m, dimensiona sx, dimensiona sy, dimensiona w, dimensiona h) {
/* New-Submatrix 

   Creates and returns a new matrix, containing the submatrix of m specified.
   If the bounds are invalid, a NULL matrix is returned. */

   gmtx *mnew;

   /* Create new matrix */
   if(!(mnew = gmtx_new(m->type, m->dtsize, 0, 0))) return(NULL);

   /* Check bounds */
   if(!_gmtx_adjust_bounds_xywh(&sx, &sy, &w, &h, m->width, m->height)) return(mnew);
   
   /* Construct new matrix */
   if(!_gmtx_submatrix_copy(mnew, w, h, m, sx, sy)) return(gmtx_free(&mnew));
   return(mnew);
   
}
   

gmtxl *gmtxl_new_submatrix(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(__gmtxlp _gmtx_new_submatrix(__gmtxp m, x, y, w, h));

}


gmtxd *gmtxd_new_submatrix(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(__gmtxdp _gmtx_new_submatrix(__gmtxp m, x, y, w, h));

}


gmtx *gmtx_new_submatrix(const gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(_gmtx_new_submatrix(m, x, y, w, h));

}


/*** Submatrix code ***/
/* Clearing a submatrix */


gmtxl *gmtxl_submatrix_zero(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(_gmtxl_submatrix_const(m, x, y, w, h, 0));

}


gmtxd *gmtxd_submatrix_zero(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(_gmtxd_submatrix_const(m, x, y, w, h, 0));

}


gmtx *gmtx_submatrix_zero(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   if(MTX_IS_LONG(m))   return(__gmtxp _gmtxl_submatrix_const(__gmtxlp m, x, y, w, h, 0));
   if(MTX_IS_DOUBLE(m)) return(__gmtxp _gmtxd_submatrix_const(__gmtxdp m, x, y, w, h, 0));
   return(NULL);

}


gmtxl *gmtxl_submatrix_ones(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(_gmtxl_submatrix_const(m, x, y, w, h, 1));

}


gmtxd *gmtxd_submatrix_ones(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(_gmtxd_submatrix_const(m, x, y, w, h, 1));

}


gmtx *gmtx_submatrix_ones(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   if(MTX_IS_LONG(m))   return(__gmtxp _gmtxl_submatrix_const(__gmtxlp m, x, y, w, h, 1));
   if(MTX_IS_DOUBLE(m)) return(__gmtxp _gmtxd_submatrix_const(__gmtxdp m, x, y, w, h, 1));
   return(NULL);

}


gmtxl *gmtxl_submatrix_const(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata value) {

   return(_gmtxl_submatrix_const(m, x, y, w, h, value));

}


gmtxd *gmtxd_submatrix_const(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata value) {

   return(_gmtxd_submatrix_const(m, x, y, w, h, value));

}


gmtxl *gmtxl_submatrix_ident(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(_gmtxl_submatrix_eyes(m, x, y, w, h, 1));

}


gmtxd *gmtxd_submatrix_ident(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   return(_gmtxd_submatrix_eyes(m, x, y, w, h, 1));

}


gmtx *gmtx_submatrix_ident(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h) {

   if(MTX_IS_LONG(m))   return(__gmtxp _gmtxl_submatrix_eyes(__gmtxlp m, x, y, w, h, 1));
   if(MTX_IS_DOUBLE(m)) return(__gmtxp _gmtxd_submatrix_eyes(__gmtxdp m, x, y, w, h, 1));
   return(NULL);

}


gmtxl *gmtxl_submatrix_eyes(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata value) {

   return(_gmtxl_submatrix_eyes(m, x, y, w, h, value));

}


gmtxd *gmtxd_submatrix_eyes(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata value) {

   return(_gmtxd_submatrix_eyes(m, x, y, w, h, value));

}


