/* 

   libj - gmtx_trans.c           Copyright (C) 2001-1998 Justin David Smith    
   justins@chaos2.org            http://chaos2.org/
    
   Matrix transpose


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>



/*** Matrix Transpose ***/
/* Internal functions */


static gmtxl *__gmtxl_transpose_copy_data(gmtxl *mdst, const gmtxl *msrc, dimensiona width, dimensiona height) {
/* Matrix-Transpose-Copy-Data

   This function transposes msrc to mdst. The original matrix is dimensions
   width,height and the new matrix will be of dimensions height,width.  msrc
   and mdst may _not_ be the same pointer.  mdst should already be sized to
   the new size.  The matrix may be zero.  */

   longdata *oldxy;
   longdata *newyx;
   dimensiona i;
   dimensiona j;

   /* Our strategy here is to read each original cell, row-wise,
      and calculate its place in the new matrix */
   oldxy = msrc->data;
   j = 0;
   while(j < height) {
      i = 0;
      newyx = MTX_XY_P(mdst, j, 0);
      while(i < width) {
         /* The new cell should be on row i, column j, where each row
            is of width old-height. */
         *newyx = *oldxy;
         newyx += height;
         oldxy++;
      }
      j++;
   }
   return(mdst);

}


static gmtxd *__gmtxd_transpose_copy_data(gmtxd *mdst, const gmtxd *msrc, dimensiona width, dimensiona height) {
/* Matrix-Transpose-Copy-Data

   This function transposes msrc to mdst. The original matrix is dimensions
   width,height and the new matrix will be of dimensions height,width.  msrc
   and mdst may _not_ be the same pointer.  mdst should already be sized to
   the new size.  The matrix may be zero.  */

   doubledata *oldxy;
   doubledata *newyx;
   dimensiona i;
   dimensiona j;

   /* Our strategy here is to read each original cell, row-wise,
      and calculate its place in the new matrix */
   oldxy = msrc->data;
   j = 0;
   while(j < height) {
      i = 0;
      newyx = MTX_XY_P(mdst, j, 0);
      while(i < width) {
         /* The new cell should be on row i, column j, where each row
            is of width old-height. */
         *newyx = *oldxy;
         newyx += height;
         oldxy++;
      }
      j++;
   }
   return(mdst);

}


static gmtx *_gmtx_transpose_copy(gmtx *mdst, const gmtx *msrc) {
/* Matrix-Transpose-Copy

   mdst does not have to be compatible with msrc, msrc may be NULL matrix. */

   dimensiona height;
   dimensiona width;

   /* Get source width, height */
   width = msrc->width;
   height= msrc->height;

   /* Prepare destination */
   if(!gmtx_resize_using(mdst, msrc)) return(NULL);
   mdst->width = height;
   mdst->height= width;

   /* Copy the transpose */
   if(MTX_IS_LONG(msrc))   return(__gmtxp __gmtxl_transpose_copy_data(__gmtxlp mdst, __gmtxlp msrc, width, height));
   if(MTX_IS_DOUBLE(msrc)) return(__gmtxp __gmtxd_transpose_copy_data(__gmtxdp mdst, __gmtxdp msrc, width, height));
   return(NULL);

}


static gmtxl *__gmtxl_transpose_square(gmtxl *m) {
/* Transposes a square matrix, in place.  The matrix is assumed to be
   square.  Data copying is not necessary for square matrices, since
   two cells are always being swapped in memory. M may be the NULL matrix. */

   longdata *oldxy;
   longdata *newyx;
   dimensiona width;
   dimensiona i;
   dimensiona j;

   /* Get width and height */
   width = m->width;

   /* To rearrange the data, we note that with square matrices,
      we are always swapping the contents of two cells. Since this
      case is much simpler, we want to try it first. */
   oldxy = m->data;
   j = width - 1;
   while(j > 0) {
      i = j;
      /* Oldxy pointer follows the columns */
      oldxy += (width - j);
         
      /* Newxy pointer starts on row (width - j), 
         column (width - j - 1), thus ... */
      newyx =  m->data + m->size - (j + 1) * (width - 1);
      while(i > 0) {
         /* Swap cells */
         MTX_SWAP(*oldxy, *newyx);
         
         /* Advance old pointer, following the row, and the
            new pointer, following the column. */
         oldxy++;
         newyx += width;
         i--;
      }     
      j--;
   }    
   return(m);
   
}


static gmtxd *__gmtxd_transpose_square(gmtxd *m) {
/* Transposes a square matrix, in place.  The matrix is assumed to be
   square.  Data copying is not necessary for square matrices, since
   two cells are always being swapped in memory. M may be the NULL matrix. */

   doubledata *oldxy;
   doubledata *newyx;
   dimensiona width;
   dimensiona i;
   dimensiona j;

   /* Get width and height */
   width = m->width;

   /* To rearrange the data, we note that with square matrices,
      we are always swapping the contents of two cells. Since this
      case is much simpler, we want to try it first. */
   oldxy = m->data;
   j = width - 1;
   while(j > 0) {
      i = j;
      /* Oldxy pointer follows the columns */
      oldxy += (width - j);
         
      /* Newxy pointer starts on row (width - j), 
         column (width - j - 1), thus ... */
      newyx =  m->data + m->size - (j + 1) * (width - 1);
      while(i > 0) {
         /* Swap cells */
         MTX_SWAP(*oldxy, *newyx);
         
         /* Advance old pointer, following the row, and the
            new pointer, following the column. */
         oldxy++;
         newyx += width;
         i--;
      }     
      j--;
   }    
   return(m);
   
}


static inline gmtx *_gmtx_transpose_square(gmtx *m) {

   if(MTX_IS_LONG(m))   return(__gmtxp __gmtxl_transpose_square(__gmtxlp m));
   if(MTX_IS_DOUBLE(m)) return(__gmtxp __gmtxd_transpose_square(__gmtxdp m));
   return(NULL);
   
}


static inline gmtx *_gmtx_transpose_vector(gmtx *m) {
/* Tranpose-Vector
   
   Transposes a vector, which is really just swapping the width and height. */

   MTX_SWAP(m->height, m->width);
   return(m);
   
}


static gmtx *_gmtx_transpose(gmtx *m) {
/* Matrix-Transpose

   Since a general case in-place transpose requires memory duplication,
   we want to see if this is a square or vector first, and try for the
   easy cases :) */

   gmtx *mtmp;

   /* Attempt to transpose a vector */
   if(MTX_IS_VECTOR(m)) return(_gmtx_transpose_vector(m));

   /* Attempt to transpose a square matrix */
   if(MTX_IS_SQUARE(m)) return(_gmtx_transpose_square(m));
   
   /* Matrix is not square. We'll have to copy the old matrix in
      its entirety to a temporary buffer for this to work */
   if(!(mtmp = gmtx_new_copy(m))) return(NULL);
   if(!_gmtx_transpose_copy(m, mtmp)) return(NULL);
   gmtx_free(&mtmp);

   return(m);

}


static inline gmtx *_gmtx_new_transpose(const gmtx *m) {

   gmtx *mnew;
   
   if(!(mnew = gmtx_new_using(m))) return(NULL);
   if(!_gmtx_transpose_copy(mnew, m)) return(gmtx_free(&mnew));
   return(mnew);

}


/*** Matrix Transpose ***/
/* Interface functions */


gmtxl *gmtxl_transpose_copy(gmtxl *mdst, const gmtxl *msrc) { return(__gmtxlp _gmtx_transpose_copy(__gmtxp mdst, __gmtxp msrc)); }
gmtxd *gmtxd_transpose_copy(gmtxd *mdst, const gmtxd *msrc) { return(__gmtxdp _gmtx_transpose_copy(__gmtxp mdst, __gmtxp msrc)); }
gmtx  *gmtx_transpose_copy(gmtx *mdst, const gmtx *msrc)    { return(_gmtx_transpose_copy(mdst, msrc)); }
gmtxl *gmtxl_transpose(gmtxl *m) { return(__gmtxlp _gmtx_transpose(__gmtxp m)); }
gmtxd *gmtxd_transpose(gmtxd *m) { return(__gmtxdp _gmtx_transpose(__gmtxp m)); }
gmtx  *gmtx_transpose(gmtx *m)   { return(_gmtx_transpose(m)); }
gmtxl *gmtxl_new_transpose(const gmtxl *m) { return(__gmtxlp _gmtx_new_transpose(__gmtxp m)); }
gmtxd *gmtxd_new_transpose(const gmtxd *m) { return(__gmtxdp _gmtx_new_transpose(__gmtxp m)); }
gmtx  *gmtx_new_transpose(const gmtx *m)   { return(_gmtx_new_transpose(m)); }


