/* 

   libj - jmtx.h                 Copyright (C) 2001-1998 Justin David Smith    
   justins@chaos2.org            http://chaos2.org/
    
   Main header for matrix library


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#ifndef __libj_mtx_included
#define __libj_mtx_included



/* Global includes */
#include <libj.h>



/* Mtx types */
#define  MTX_LONG       0x01
#define  MTX_DOUBLE     0x02
#define  MTX_SPARSE     0x04
#define  MTX_GRID       0x08
#define  MTX_GLONG      (MTX_LONG | MTX_GRID)
#define  MTX_CUSTOM     0x80



/* Matrix coordinate data */
typedef sdword longdata;
typedef struct _mcoordl {
   dimensiona x;
   dimensiona y;
   longdata z;
} mcoordl;

typedef double doubledata;
typedef struct _mcoordd {
   dimensiona x;
   dimensiona y;
   doubledata z;
} mcoordd;



/* Matrix base type */
typedef struct _mtx {
   uword type;       /* Matrix type */
   uword dtsize;     /* Size of data elements */
   dimensiona width;  /* Width of matrix */
   dimensiona height; /* Height of matrix */
   dimensiona size;   /* Width * Height */
   dimensiona bufsize;/* Size of data buffer */
   dimensiona rowsize;/* Size of row buffer */
   char *data;       /* Actual matrix data */
   void *misc;       /* Miscellaneous user data */
} mtx;

/* We allow several types of matrices */
typedef mtx gmtx;
typedef mtx vmtx;
typedef mtx smtx;



/* Grid matrices */
typedef struct _gmtxl {
   uword type;
   uword dtsize;
   dimensiona width;
   dimensiona height;
   dimensiona size;
   dimensiona bufsize;
   dimensiona rowsize;
   longdata *data;
   void *misc;
} gmtxl;

typedef struct _gmtxd {
   uword type;
   uword dtsize;
   dimensiona width;
   dimensiona height;
   dimensiona size;
   dimensiona bufsize;
   dimensiona rowsize;
   doubledata *data;
   void *misc;
} gmtxd;

typedef gmtxl vmtxl;
typedef gmtxd vmtxd;



/* Matrix conversion macros */
#define  __mtx       (mtx)
#define  __mtxp      (mtx *)
#define  __mtxpp     (mtx **)
#define  __gmtx      (gmtx)   
#define  __gmtxp     (gmtx *)   
#define  __gmtxpp    (gmtx **)   
#define  __gmtxl     (gmtxl)   
#define  __gmtxlp    (gmtxl *)   
#define  __gmtxlpp   (gmtxl **)   
#define  __gmtxd     (gmtxd)   
#define  __gmtxdp    (gmtxd *)   
#define  __gmtxdpp   (gmtxd **)   

#define  __vmtx      (vmtx)
#define  __vmtxp     (vmtx *)
#define  __vmtxpp    (vmtx **)
#define  __vmtxl     (vmtxl)
#define  __vmtxlp    (vmtxl *)
#define  __vmtxlpp   (vmtxl **)
#define  __vmtxd     (vmtxd)
#define  __vmtxdp    (vmtxd *)
#define  __vmtxdpp   (vmtxd **)



/* Matrix index manipulation */

#define  MTX_X_IN_BOUNDS(m,x)    ((x) >= 0 && (x) < (m)->width)
#define  MTX_Y_IN_BOUNDS(m,y)    ((y) >= 0 && (y) < (m)->height)
#define  MTX_IN_BOUNDS(m,x,y)    (MTX_X_IN_BOUNDS(m,x) && MTX_Y_IN_BOUNDS(m,y))
#define  MTX_P_IN_BOUNDS(m,p)    ((p) >= (m)->data && (p) < (m)->data + (m)->size)
#define  MTX_XY_P(m,x,y)         ((m)->data + (x) + (y) * (m)->width)
#define  MTX_GET_XY(m,x,y)       *MTX_XY_P(m,x,y)
#define  MTX_SET_XY(m,x,y,z)     (*MTX_XY_P(m,x,y) = (z))
#define  MTX_P_X(m,p)            (((p) - (m)->data) % (m)->width)
#define  MTX_P_Y(m,p)            (((p) - (m)->data) / (m)->width)
#define  MTX_P_LEFT(m,p)         ((p)-1)
#define  MTX_P_RIGHT(m,p)        ((p)+1)
#define  MTX_P_UP(m,p)           ((p)-(m)->width)
#define  MTX_P_DOWN(m,p)         ((p)+(m)->width)


            /***  From libgmtx.c  ***/
            

bool gmtxl_is_null(const gmtxl *m); 
bool gmtxd_is_null(const gmtxd *m); 
bool gmtx_is_null(const gmtx *m); 

bool gmtxl_is_vector(const gmtxl *m);
bool gmtxd_is_vector(const gmtxd *m);
bool gmtx_is_vector(const gmtx *m);

bool gmtxl_are_equal(const gmtxl *a, const gmtxl *b); 
bool gmtxd_are_equal(const gmtxd *a, const gmtxd *b); 
bool gmtx_are_equal(const gmtx *a, const gmtx *b); 

bool gmtxl_is_square(const gmtxl *m); 
bool gmtxd_is_square(const gmtxd *m); 
bool gmtx_is_square(const gmtx *m); 


            /***  From libgmtx_cons.c  ***/
            

gmtx  *gmtx_free(gmtx **m);
gmtxl *gmtxl_free(gmtxl **m);
gmtxd *gmtxd_free(gmtxd **m);

gmtxl *gmtxl_resize(gmtxl *m, dimensiona width, dimensiona height); 
gmtxd *gmtxd_resize(gmtxd *m, dimensiona width, dimensiona height); 
gmtx  *gmtx_resize(gmtx *m, dimensiona width, dimensiona height); 
gmtxl *gmtxl_resize_using(gmtxl *mdst, const gmtxl *msrc);
gmtxd *gmtxd_resize_using(gmtxd *mdst, const gmtxd *msrc);
gmtx  *gmtx_resize_using(gmtx *mdst, const gmtx *msrc);
gmtxl *gmtxl_reshape(gmtxl *m, dimensiona width, dimensiona height); 
gmtxd *gmtxd_reshape(gmtxd *m, dimensiona width, dimensiona height); 
gmtx  *gmtx_reshape(gmtx *m, dimensiona width, dimensiona height); 

gmtxl *gmtxl_new(dimensiona width, dimensiona height); 
gmtxd *gmtxd_new(dimensiona width, dimensiona height); 
gmtx  *gmtx_new(uword type, uword dtsize, dimensiona width, dimensiona height);
gmtxl *gmtxl_new_using(const gmtxl *m);
gmtxd *gmtxd_new_using(const gmtxd *m);
gmtx  *gmtx_new_using(const gmtx *m);
gmtxl *gmtxl_new_zero(dimensiona width, dimensiona height); 
gmtxd *gmtxd_new_zero(dimensiona width, dimensiona height); 
gmtxl *gmtxl_new_ones(dimensiona width, dimensiona height); 
gmtxd *gmtxd_new_ones(dimensiona width, dimensiona height); 
gmtxl *gmtxl_new_ident(dimensiona width, dimensiona height); 
gmtxd *gmtxd_new_ident(dimensiona width, dimensiona height); 
gmtxl *gmtxl_new_const(dimensiona width, dimensiona height, longdata value); 
gmtxd *gmtxd_new_const(dimensiona width, dimensiona height, doubledata value); 
gmtxl *gmtxl_new_eyes(dimensiona width, dimensiona height, longdata value); 
gmtxd *gmtxd_new_eyes(dimensiona width, dimensiona height, doubledata value); 
gmtxl *gmtxl_new_copy(const gmtxl *m); 
gmtxd *gmtxd_new_copy(const gmtxd *m); 
gmtx  *gmtx_new_copy(const gmtx *m); 

gmtxl *gmtxl_zero(gmtxl *m); 
gmtxd *gmtxd_zero(gmtxd *m); 
gmtx  *gmtx_zero(gmtx *m);
gmtxl *gmtxl_ones(gmtxl *m); 
gmtxd *gmtxd_ones(gmtxd *m); 
gmtx  *gmtx_ones(gmtx *m);
gmtxl *gmtxl_ident(gmtxl *m); 
gmtxd *gmtxd_ident(gmtxd *m); 
gmtx  *gmtx_ident(gmtx *m);
gmtxl *gmtxl_const(gmtxl *m, longdata value); 
gmtxd *gmtxd_const(gmtxd *m, doubledata value); 
gmtxl *gmtxl_eyes(gmtxl *m, longdata value); 
gmtxd *gmtxd_eyes(gmtxd *m, doubledata value); 
gmtxl *gmtxl_copy(gmtxl *mdst, const gmtxl *msrc); 
gmtxd *gmtxd_copy(gmtxd *mdst, const gmtxd *msrc); 
gmtx  *gmtx_copy(gmtx *mdst, const gmtx *msrc); 


            /***  From libgmtx_trans.c  ***/


gmtxl *gmtxl_transpose_copy(gmtxl *mdst, const gmtxl *msrc); 
gmtxd *gmtxd_transpose_copy(gmtxd *mdst, const gmtxd *msrc); 
gmtx  *gmtx_transpose_copy(gmtx *mdst, const gmtx *msrc);
gmtxl *gmtxl_transpose(gmtxl *m); 
gmtxd *gmtxd_transpose(gmtxd *m); 
gmtx  *gmtx_transpose(gmtx *m); 
gmtxl *gmtxl_new_transpose(const gmtxl *m); 
gmtxd *gmtxd_new_transpose(const gmtxd *m); 
gmtx  *gmtx_new_transpose(const gmtx *m); 


            /***  From libgmtx_rev.c  ***/


gmtxl *gmtxl_reverse_horiz(gmtxl *m);
gmtxd *gmtxd_reverse_horiz(gmtxd *m);
gmtx  *gmtx_reverse_horiz(gmtx *m);
gmtxl *gmtxl_reverse_vert(gmtxl *m);
gmtxd *gmtxd_reverse_vert(gmtxd *m);
gmtx  *gmtx_reverse_vert(gmtx *m);
gmtxl *gmtxl_reverse(gmtxl *m);
gmtxd *gmtxd_reverse(gmtxd *m);
gmtx  *gmtx_reverse(gmtx *m);


            /***  From libgmtx_subm.c  ***/


void mtx_fix_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2); 
void mtx_fix_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h); 
void mtx_xyxy_to_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona x1, dimensiona y1, dimensiona x2, dimensiona y2); 
void mtx_xywh_to_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona x, dimensiona y, dimensiona w, dimensiona h); 
bool mtx_adjust_bounds_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona width, dimensiona height);
bool mtx_adjust_bounds_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona width, dimensiona height);
bool mtx_shift_bounds_xyxy(dimensiona *x1, dimensiona *y1, dimensiona *x2, dimensiona *y2, dimensiona width, dimensiona height); 
bool mtx_shift_bounds_xywh(dimensiona *x, dimensiona *y, dimensiona *w, dimensiona *h, dimensiona width, dimensiona height); 

gmtxl *gmtxl_submatrix_copy(gmtxl *mdst, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy);
gmtxd *gmtxd_submatrix_copy(gmtxd *mdst, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy);
gmtx  *gmtx_submatrix_copy(gmtx *mdst, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy);
gmtxl *gmtxl_submatrix_copy_to(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy);
gmtxd *gmtxd_submatrix_copy_into(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy);
gmtx  *gmtx_submatrix_copy_into(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy);

gmtxl *gmtxl_new_submatrix(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtxd *gmtxd_new_submatrix(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtx  *gmtx_new_submatrix(const gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);

gmtxl *gmtxl_submatrix_zero(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtxd *gmtxd_submatrix_zero(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtx  *gmtx_submatrix_zero(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtxl *gmtxl_submatrix_ones(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtxd *gmtxd_submatrix_ones(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtx  *gmtx_submatrix_ones(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtxl *gmtxl_submatrix_const(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata value);
gmtxd *gmtxd_submatrix_const(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata value);
gmtxl *gmtxl_submatrix_ident(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtxd *gmtxd_submatrix_ident(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtx  *gmtx_submatrix_ident(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtxl *gmtxl_submatrix_eyes(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata value);
gmtxd *gmtxd_submatrix_eyes(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata value);


            /***  From libgmtx_det.c  ***/


gmtxl *gmtxl_minor_copy(gmtxl *mdst, const gmtxl *msrc, dimensiona x, dimensiona y);
gmtxd *gmtxd_minor_copy(gmtxd *mdst, const gmtxd *msrc, dimensiona x, dimensiona y);
gmtx  *gmtx_minor_copy(gmtx *mdst, const gmtx *msrc, dimensiona x, dimensiona y);
gmtxl *gmtxl_new_minor(const gmtxl *m, dimensiona x, dimensiona y);
gmtxd *gmtxd_new_minor(const gmtxd *m, dimensiona x, dimensiona y);
gmtx  *gmtx_new_minor(const gmtx *m, dimensiona x, dimensiona y);

longdata gmtxl_det_large(const gmtxl *m);
doubledata gmtxd_det_large(const gmtxd *m);
longdata gmtxl_det(const gmtxl *m);
doubledata gmtxd_det(const gmtxd *m);


            /***  From libgmtx_math.c  ***/


gmtxl *gmtxl_add(gmtxl *mdst, const gmtxl *msrc);
gmtxd *gmtxd_add(gmtxd *mdst, const gmtxd *msrc);
gmtx  *gmtx_add(gmtx *mdst, const gmtx *msrc);
gmtxl *gmtxl_sub(gmtxl *mdst, const gmtxl *msrc);
gmtxd *gmtxd_sub(gmtxd *mdst, const gmtxd *msrc);
gmtx  *gmtx_sub(gmtx *mdst, const gmtx *msrc);
gmtxl *gmtxl_add_times(gmtxl *mdst, const gmtxl *msrc, longdata times);
gmtxd *gmtxd_add_times(gmtxd *mdst, const gmtxd *msrc, doubledata times);
gmtxl *gmtxl_add_scalar(gmtxl *m, longdata scalar);
gmtxd *gmtxd_add_scalar(gmtxd *m, doubledata scalar);
gmtxl *gmtxl_mult_scalar(gmtxl *m, longdata scalar);
gmtxd *gmtxd_mult_scalar(gmtxd *m, doubledata scalar);

gmtxl *gmtxl_submatrix_add(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy);
gmtxd *gmtxd_submatrix_add(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy);
gmtx  *gmtx_submatrix_add(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy);
gmtxl *gmtxl_submatrix_sub(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy);
gmtxd *gmtxd_submatrix_sub(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy);
gmtx  *gmtx_submatrix_sub(gmtx *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtx *msrc, dimensiona sx, dimensiona sy);
gmtxl *gmtxl_submatrix_add_times(gmtxl *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxl *msrc, dimensiona sx, dimensiona sy, longdata times);
gmtxd *gmtxd_submatrix_add_times(gmtxd *mdst, dimensiona dx, dimensiona dy, dimensiona w, dimensiona h, const gmtxd *msrc, dimensiona sx, dimensiona sy, doubledata times);
gmtxl *gmtxl_submatrix_add_scalar(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata scalar);
gmtxd *gmtxd_submatrix_add_scalar(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata scalar);
gmtxl *gmtxl_submatrix_mult_scalar(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, longdata scalar);
gmtxd *gmtxd_submatrix_mult_scalar(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h, doubledata scalar);

longdata gmtxl_sum(const gmtxl *m);
doubledata gmtxd_sum(const gmtxd *m);
longdata gmtxl_submatrix_sum(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
doubledata gmtxd_submatrix_sum(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
longdata gmtxl_min(const gmtxl *m);
doubledata gmtxd_min(const gmtxd *m);
longdata gmtxl_max(const gmtxl *m);
doubledata gmtxd_max(const gmtxd *m);
longdata gmtxl_mean(const gmtxl *m);
doubledata gmtxd_mean(const gmtxd *m);
longdata gmtxl_submatrix_min(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
doubledata gmtxd_submatrix_min(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
longdata gmtxl_submatrix_max(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
doubledata gmtxd_submatrix_max(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
longdata gmtxl_submatrix_mean(const gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
doubledata gmtxd_submatrix_mean(const gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);

gmtxl *gmtxl_negate(gmtxl *m);
gmtxd *gmtxd_negate(gmtxd *m);
gmtx  *gmtx_negate(gmtx *m);
gmtxl *gmtxl_submatrix_negate(gmtxl *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtxd *gmtxd_submatrix_negate(gmtxd *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);
gmtx  *gmtx_submatrix_negate(gmtx *m, dimensiona x, dimensiona y, dimensiona w, dimensiona h);


            /***  From libgmtx_mult.c  ***/


gmtxl *gmtxl_multiply_copy(gmtxl *c, const gmtxl *a, const gmtxl *b);
gmtxd *gmtxd_multiply_copy(gmtxd *c, const gmtxd *a, const gmtxd *b);
gmtx  *gmtx_multiply_copy(gmtx *c, const gmtx *a, const gmtx *b);
gmtxl *gmtxl_new_multiply(const gmtxl *a, const gmtxl *b);
gmtxd *gmtxd_new_multiply(const gmtxd *a, const gmtxd *b);
gmtx  *gmtx_new_multiply(const gmtx *a, const gmtx *b);
gmtxl *gmtxl_multiply(gmtxl *a, const gmtxl *b);
gmtxd *gmtxd_multiply(gmtxd *a, const gmtxd *b);
gmtx  *gmtx_multiply(gmtx *a, const gmtx *b);


            /***  From libgmtx_inv.c  ***/


gmtxl *gmtxl_cof(gmtxl *m);
gmtxd *gmtxd_cof(gmtxd *m);
gmtx  *gmtx_cof(gmtx *m);
gmtxl *gmtxl_new_cof(const gmtxl *m);
gmtxd *gmtxd_new_cof(const gmtxd *m);
gmtx  *gmtx_new_cof(const gmtx *m);
gmtxl *gmtxl_cof_copy(gmtxl *mdst, const gmtxl *msrc); 
gmtxd *gmtxd_cof_copy(gmtxd *mdst, const gmtxd *msrc);
gmtx  *gmtx_cof_copy(gmtx *mdst, const gmtx *msrc);

gmtxl *gmtxl_invert(gmtxl *m);
gmtxd *gmtxd_invert(gmtxd *m);
gmtx  *gmtx_invert(gmtx *m);
gmtxl *gmtxl_new_invert(const gmtxl *m);
gmtxd *gmtxd_new_invert(const gmtxd *m);
gmtx  *gmtx_new_invert(const gmtx *m);
gmtxl *gmtxl_invert_copy(gmtxl *mdst, const gmtxl *msrc); 
gmtxd *gmtxd_invert_copy(gmtxd *mdst, const gmtxd *msrc);
gmtx  *gmtx_invert_copy(gmtx *mdst, const gmtx *msrc);


            /***  From libvmtx.c  ***/


bool vmtxl_is_null(const vmtxl *m);
bool vmtxd_is_null(const vmtxd *m);
bool vmtx_is_null(const vmtx *m);

bool vmtxl_is_vector(const vmtxl *m);
bool vmtxd_is_vector(const vmtxd *m);
bool vmtx_is_vector(const vmtx *m);

bool vmtxl_are_equal(const vmtxl *a, const vmtxl *b);
bool vmtxd_are_equal(const vmtxd *a, const vmtxd *b);
bool vmtx_are_equal(const vmtx *a, const vmtx *b);

bool vmtxl_is_square(const vmtxl *m);
bool vmtxd_is_square(const vmtxd *m);
bool vmtx_is_square(const vmtx *m);


            /***  From libvmtx_cons.c  ***/


vmtxl *vmtxl_free(vmtxl **v);
vmtxd *vmtxd_free(vmtxd **v);
vmtx  *vmtx_free(vmtx **v);
vmtxl *vmtxl_new(dimensiona size);
vmtxd *vmtxd_new(dimensiona size);
vmtx  *vmtx_new(uword type, uword dtsize, dimensiona size);
vmtxl *vmtxl_new_using(const vmtxl *v);
vmtxd *vmtxd_new_using(const vmtxd *v);
vmtx  *vmtx_new_using(const vmtx *v);
vmtxl *vmtxl_resize(vmtxl *v, dimensiona size);
vmtxd *vmtxd_resize(vmtxd *v, dimensiona size);
vmtx  *vmtx_resize(vmtx *v, dimensiona size);
vmtxl *vmtxl_resize_using(vmtxl *vdst, const vmtxl *vsrc);
vmtxd *vmtxd_resize_using(vmtxd *vdst, const vmtxd *vsrc);
vmtx  *vmtx_resize_using(vmtx *vdst, const vmtx *vsrc);
vmtxl *vmtxl_new_zero(dimensiona size);
vmtxd *vmtxd_new_zero(dimensiona size);
vmtxl *vmtxl_new_ones(dimensiona size);
vmtxd *vmtxd_new_ones(dimensiona size);
vmtxl *vmtxl_new_const(dimensiona size, longdata value);
vmtxd *vmtxd_new_const(dimensiona size, doubledata value);

vmtxl *vmtxl_zero(vmtxl *v);
vmtxd *vmtxd_zero(vmtxd *v);
vmtx  *vmtx_zero(vmtx *v);
vmtxl *vmtxl_ones(vmtxl *v);
vmtxd *vmtxd_ones(vmtxd *v);
vmtx  *vmtx_ones(vmtx *v);
vmtxl *vmtxl_const(vmtxl *v, longdata value);
vmtxd *vmtxd_const(vmtxd *v, doubledata value);

vmtxl *vmtxl_new_copy(const vmtxl *v);
vmtxd *vmtxd_new_copy(const vmtxd *v);
vmtx  *vmtx_new_copy(const vmtx *v);
vmtxl *vmtxl_copy(vmtxl *vdst, const vmtxl *vsrc);
vmtxd *vmtxd_copy(vmtxd *vdst, const vmtxd *vsrc);
vmtx  *vmtx_copy(vmtx *vdst, const vmtx *vsrc);


            /***  From libvmtx_mult.c  ***/


vmtxl *vmtxl_gmtxl_multiply_copy(vmtxl *c, const vmtxl *a, const gmtxl *b);
vmtxd *vmtxd_gmtxd_multiply_copy(vmtxd *c, const vmtxd *a, const gmtxd *b);
vmtx  *vmtx_gmtx_multiply_copy(vmtx *c, const vmtx *a, const gmtx *b);
vmtxl *vmtxl_gmtxl_new_multiply(const vmtxl *a, const gmtxl *b);
vmtxd *vmtxd_gmtxd_new_multiply(const vmtxd *a, const gmtxd *b);
vmtx  *vmtx_gmtx_new_multiply(const vmtx *a, const gmtx *b);
vmtxl *gmtxl_vmtxl_multiply_copy(vmtxl *c, const gmtxl *a, const vmtxl *b);
vmtxd *gmtxd_vmtxd_multiply_copy(vmtxd *c, const gmtxd *a, const vmtxd *b);
vmtx  *gmtx_vmtx_multiply_copy(vmtx *c, const gmtx *a, const vmtx *b);
vmtxl *gmtxl_vmtxl_new_multiply(const gmtxl *a, const vmtxl *b);
vmtxd *gmtxd_vmtxd_new_multiply(const gmtxd *a, const vmtxd *b);
vmtx  *gmtxvmtx_new_multiply(const gmtx *a, const vmtx *b);


            /***  From libvmtx_dot.c  ***/


longdata vmtxl_dot(const vmtxl *a, const vmtxl *b);
doubledata vmtxd_dot(const vmtxd *a, const vmtxd *b);
vmtxl *vmtxl_cross_copy(vmtxl *c, const vmtxl *a, const vmtxl *b);
vmtxd *vmtxd_cross_copy(vmtxd *c, const vmtxd *a, const vmtxd *b);
vmtx  *vmtx_cross_copy(vmtx *c, const vmtx *a, const vmtx *b);
vmtxl *vmtxl_cross(vmtxl *a, const vmtxl *b);
vmtxd *vmtxd_cross(vmtxd *a, const vmtxd *b);
vmtx  *vmtx_cross(vmtx *a, const vmtx *b);
vmtxl *vmtxl_new_cross(const vmtxl *a, const vmtxl *b);
vmtxd *vmtxd_new_cross(const vmtxd *a, const vmtxd *b);
vmtx  *vmtx_new_cross(const vmtx *a, const vmtx *b);


            /***  Version  ***/

unsigned int libjmtx_version(void);


#endif

