/* 

   libj - vmtx_dot.c             Copyright (C) 2001-1998 Justin David Smith    
   justins@chaos2.org            http://chaos2.org/
    
   Dot product for vectors


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#include <libjmtx.h>
#include <_mtx_int.h>



/*** Vector Dot, Cross Products ***/
/* Internal functions */


static inline longdata _vmtxl_dot(const vmtxl *a, const vmtxl *b) {

   longdata dot;
   longdata *pa;
   longdata *pb;
   dimensiona size;   
   
   /* Make sure vectors are compatible */
   if(!MTX_ARE_COMPAT(a, b)) return(0);
   
   /* Verify size of vectors; if unequal, we will assume any missing
      elements are zero */
   size = MTX_MIN(MTX_VECTOR_SIZE(a), MTX_VECTOR_SIZE(b));
   
   /* Calculate dot product */
   dot = 0;
   pa = a->data;
   pb = b->data;
   while(size) {
      dot += (*pa) * (*pb);
      pa++;
      pb++;
      size--;
   }
   return(dot);

}


static inline doubledata _vmtxd_dot(const vmtxd *a, const vmtxd *b) {

   doubledata dot;
   doubledata *pa;
   doubledata *pb;
   dimensiona size;   
   
   /* Make sure vectors are compatible */
   if(!MTX_ARE_COMPAT(a, b)) return(0);
   
   /* Verify size of vectors; if unequal, we will assume any missing
      elements are zero */
   size = MTX_MIN(MTX_VECTOR_SIZE(a), MTX_VECTOR_SIZE(b));
   
   /* Calculate dot product */
   dot = 0;
   pa = a->data;
   pb = b->data;
   while(size) {
      dot += (*pa) * (*pb);
      pa++;
      pb++;
      size--;
   }
   return(dot);

}


static inline vmtxl *_vmtxl_cross_copy(vmtxl *c, const vmtxl *a, const vmtxl *b) {

   longdata size;
   
   /* make sure source vectors are compatible */
   if(!MTX_ARE_COMPAT(a, b)) return(NULL);

   /* Verify size of vectors; if unequal, we will assume any missing
      elements are zero */
   if(MTX_VECTOR_SIZE(a) != 3) return(NULL);
   if(MTX_VECTOR_SIZE(b) != 3) return(NULL);
   size = 3;
   
   /* Allocate memory for destination vector */
   if(!vmtx_resize(__vmtxp c, 3)) return(NULL);
   
   /* Calculate cross product */
   c->data[0] = a->data[1] * b->data[2] - b->data[1] * a->data[2];
   c->data[1] = -a->data[0] * b->data[2]+ b->data[0] * a->data[2];
   c->data[2] = a->data[0] * b->data[1] - b->data[0] * a->data[1];
   return(c);

}


static inline vmtxd *_vmtxd_cross_copy(vmtxd *c, const vmtxd *a, const vmtxd *b) {

   longdata size;

   /* make sure source vectors are compatible */
   if(!MTX_ARE_COMPAT(a, b)) return(NULL);

   /* Verify size of vectors; if unequal, we will assume any missing
      elements are zero */
   if(MTX_VECTOR_SIZE(a) != 3) return(NULL);
   if(MTX_VECTOR_SIZE(b) != 3) return(NULL);
   size = 3;
   
   /* Allocate memory for destination vector */
   if(!vmtx_resize(__vmtxp c, 3)) return(NULL);
   
   /* Calculate cross product */
   c->data[0] = a->data[1] * b->data[2] - b->data[1] * a->data[2];
   c->data[1] = -a->data[0] * b->data[2]+ b->data[0] * a->data[2];
   c->data[2] = a->data[0] * b->data[1] - b->data[0] * a->data[1];
   return(c);

}


static inline vmtx *_vmtx_cross_copy(vmtx *c, const vmtx *a, const vmtx *b) {

   if(MTX_IS_LONG(a))   return(__vmtxp _vmtxl_cross_copy(__vmtxlp c, __vmtxlp a, __vmtxlp b));
   if(MTX_IS_DOUBLE(a)) return(__vmtxp _vmtxd_cross_copy(__vmtxdp c, __vmtxdp a, __vmtxdp b));
   return(NULL);
   
}


static inline vmtx *_vmtx_cross(vmtx *a, const vmtx *b) {

   vmtx *vtmp;
   
   if(!(vtmp = vmtx_new_copy(a))) return(NULL);
   if(!_vmtx_cross_copy(a, vtmp, b)) return(vmtx_free(&vtmp));
   vmtx_free(&vtmp);
   return(a);

}


static inline vmtx *_vmtx_new_cross(const vmtx *a, const vmtx *b) {

   vmtx *vnew;
   
   if(!(vnew = vmtx_new_using(a))) return(NULL);
   if(!_vmtx_cross_copy(vnew, a, b)) return(vmtx_free(&vnew));
   return(vnew);

}


/*** Vector Dot, Cross Products ***/
/* Interface */


longdata vmtxl_dot(const vmtxl *a, const vmtxl *b) {

   return(_vmtxl_dot(a, b));

}


doubledata vmtxd_dot(const vmtxd *a, const vmtxd *b) {

   return(_vmtxd_dot(a, b));

}


vmtxl *vmtxl_cross_copy(vmtxl *c, const vmtxl *a, const vmtxl *b) { return(__vmtxlp _vmtx_cross_copy(__vmtxp c, __vmtxp a, __vmtxp b)); }
vmtxd *vmtxd_cross_copy(vmtxd *c, const vmtxd *a, const vmtxd *b) { return(__vmtxdp _vmtx_cross_copy(__vmtxp c, __vmtxp a, __vmtxp b)); }
vmtx  *vmtx_cross_copy(vmtx *c, const vmtx *a, const vmtx *b)     { return(_vmtx_cross_copy(c, a, b)); }
vmtxl *vmtxl_cross(vmtxl *a, const vmtxl *b) { return(__vmtxlp _vmtx_cross(__vmtxp a, __vmtxp b)); }
vmtxd *vmtxd_cross(vmtxd *a, const vmtxd *b) { return(__vmtxdp _vmtx_cross(__vmtxp a, __vmtxp b)); }
vmtx  *vmtx_cross(vmtx *a, const vmtx *b)    { return(_vmtx_cross(a, b)); }
vmtxl *vmtxl_new_cross(const vmtxl *a, const vmtxl *b) { return(__vmtxlp _vmtx_new_cross(__vmtxp a, __vmtxp b)); }
vmtxd *vmtxd_new_cross(const vmtxd *a, const vmtxd *b) { return(__vmtxdp _vmtx_new_cross(__vmtxp a, __vmtxp b)); }
vmtx  *vmtx_new_cross(const vmtx *a, const vmtx *b)    { return(_vmtx_new_cross(a, b)); }
