/* $Header: /fridge/cvs/libj/jstra/_stra_alloc.h,v 1.3 2001/12/10 19:44:08 justins Exp $ */                                                                                                    
/* 

   libj - _stra_alloc.c          Copyright (C) 2001-1998 Justin David Smith
   justins (a) chaos2.org        http://chaos2.org/
    
   StringA allocation and manipulations


   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this library; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
   
   This file is part of LIBJ.
   
*/
#ifndef __libj__stra_alloc_included
#define __libj__stra_alloc_included


#define STRINGA_TRACE_ALLOC   0
#define STRINGA_TRACE_COPY    0


/* Chunks are the block sizes we try to allocate in by default */
#define  CHUNK_SIZE           0x100
#define  NEXT_CHUNK(s)        ((((s) + (CHUNK_SIZE - 1)) / (CHUNK_SIZE)) * (CHUNK_SIZE))


#if STRINGA_TRACE_ALLOC && (__need_minsizea || __need_newminsizea)
   static void *MALLOC(sizea s) {
      void *p = malloc(s);
      printf("jstra: Alloc'd %p(%d)\n", p, s);
      return(p);
   }
#else /* Normal allocation pathways */
   #define MALLOC(s)       malloc(s)
#endif /* STRINGA_TRACE_ALLOC */


#if STRINGA_TRACE_ALLOC && (__need_minsizea || __need_trunca)
   static void *REALLOC(void *o, sizea s) {
      void *p = realloc(o, s);
      printf("jstra: Realloc'd %p->%p(%d)\n", o, p, s);
      return(p);
   }
#else /* Normal allocation pathways */
   #define REALLOC(p, s)   realloc((p), (s))
#endif /* STRINGA_TRACE_ALLOC */


#if STRINGA_TRACE_ALLOC && __need_freea
   static void FREE(void *p) {
      printf("jstra: Free'd %p\n", p);
      free(p);
   }
#else /* Normal allocation pathways */
   #define FREE(p)         free(p)
#endif /* STRINGA_TRACE_ALLOC */


#if __need_minsizea
   static inline sizea __strminsizea(stringa *d, sizea s) {
   /* String-Set-Minimum-Size-A
      This function sets the minimum size of the actual buffer of the StringA 
      pointed to by d. The resulting buffer will be the larger of the current
      buffer size or the size requested in <s>. The value returned is the
      new size of the buffer. The original contents of the string buffer should
      remain unchanged. 
      
      Note that we may actually allocate up to the next chunk size.
      
      Note, if <*d> is NULL, then a new buffer will be created and initialised
      to an empty string. <s> should not include the extra byte required for a
      NULL character terminating the string. */
      
      char *p;
      
      /* Account for the NULL character */
      ++s;

      /* Round up to nearest chunk size */
      s = NEXT_CHUNK(s);

      /* Create the new string buffer or resize existing buffer if needed */
      if(d->string != NULL) {
         if(s <= d->bufsize) return(d->bufsize);

         /* Expand the size of the existing string. */
         d->string  = (char *)REALLOC(d->string, s);
         d->bufsize = s;
         return(s);
      }

      /* Create a new empty string. */
      p = (char *)MALLOC(s);
      d->string  = p;
      d->length  = 0;
      d->bufsize = s;
      if(p != NULL) *p = '\0';
      return(s);

   }
#endif /* __need_minsizea */



#if __need_trunca
   static inline sizea __strtrunca(stringa *d, sizea s) {
   /* String-Truncate-A
      Truncates the string (and the buffer) to s characters (unless the buffer is
      already smaller).  A null character will be appended to the buffer.  */
      
      char *p;

      /* Create the new string buffer or resize existing buffer if needed */
      if(d->string != NULL) {
         if(s >= d->bufsize) return(d->bufsize);

         /* Truncate the size of the existing string. */
         p = (char *)REALLOC(d->string, s);
         if(p != NULL) *(p + s - 1) = '\0';
         d->string  = p;
         d->bufsize = s;
         return(s);
      }

      /* Nothing to do */
      return(0);

   }
#endif /* __need_trunca */



#if __need_newminsizea
   static inline sizea __strnewminsizea(stringa *d, sizea s) {
   /* String-New-Minimum-Size-A
      Same as above, but resulting *d is completely uninitialised. If *d used
      to point to a string and the buffer isn't large enough, the original
      string is discarded and a new buffer allocated. 
      
      The length value will be uninitialised after this function is called,
      though buffer size will be properly set. */
      
      /* Account for the NULL character */
      ++s;
      
      /* Round up to nearest chunk size */
      s = NEXT_CHUNK(s);

      /* Create the new string buffer or resize existing buffer if needed */
      if(d->string != NULL) {
         if(s <= d->bufsize) return(d->bufsize);

         /* Expand the size of the existing string. */
         FREE(d->string);
         d->string  = (char *)MALLOC(s);
         d->bufsize = s;
         return(s);
      }
      
      /* Create a new empty string. */
      d->string  = (char *)MALLOC(s);
      d->bufsize = s;
      return(s);

   }
#endif /* __need_newminsizea */



#if __need_setlena
   static inline sizea __strsetlena(stringa *d) {
   /* String-Set-Length-A
      This function updates the length of StringA referenced by <*d> by scanning
      the actual string and determining how far down the null character is. 
      The actual length of the string is returned. */
      
      return((d->length = STRLENN(d->string)));

   }
#endif /* __need_setlena */



#if __need_freea
   static inline void __strfreea(stringa *a) {
   /* This function will release a StringA string, and reset 
      the pointer to NULL. */

      FREE(a->string);
      a->string = NULL;

   }
#endif /* __need_freea */



#if __need_cpya 
   static inline stringa *__strcpya(stringa *d, const char *s) {
   /* String-Copy-A
      This is a primary interface function for StringA. StringA is a system
      of dynamically resizable strings in which idealistically, you will
      never have to allocate a string buffer again. In practice, this can
      save quite a bit of memory versus fixed string buffers for holding
      data records, etc. You have to pass these functions a pointer TO
      THE CHAR POINTER so they can unallocate any existing string at that
      pointer and reallocate a nice, new string (if the old string was not
      long enough). 
      
      This function will automatically free the *d pointer (if it is not 
      null) using a call to free, and then malloc a new pointer at that
      location that is the size of <s>, plus another character for the
      terminating null character. Then it will copy the string using a 
      memcopy call. This function will always release what was previously
      allocated to *d, regardless of what type of data was previously there.
      Thus, be careful about your old pointers. 
      
      Note that if <s> = NULL, then this will only free the <*d> pointer,
      and assign *d to NULL. */

      sizea slen;

      if(s != NULL) {

         /* Find length of source. */
         slen = STRLEN(s);

         /* Create new string buffer if needed */
         #if STRINGA_TRACE_COPY
            printf("jstra: __strcpya, Allocating %d bytes from \"%s\", for %p\n", slen, s, stra(d));
         #endif /* STRINGA_TRACE_COPY */
         __strnewminsizea(d, slen);

         /* Copy the string */
         d->length = slen;
         if(d->string != NULL) MEMCPY(d->string, s, slen + 1);
      } else {

         /* Release old pointer */
         #if STRINGA_TRACE_COPY
            printf("jstra: __strcpya, Deleting %p\n", d);
         #endif /* STRINGA_TRACE_COPY */
         __strfreea(d);
      }
      return(d);

   }
#endif /* __need_cpya */



#if __need_cpyaa
   static inline stringa *__strcpyaa(stringa *d, const stringa *s) {
   /* String-Copy-A-To-A
      Same as above, except we may make the assumption that s is a StrinA
      as well (thus we can determine the length of s with a constant-time
      function). */

      sizea slen;

      if(s != NULL && s->string != NULL) {

         /* Find length of source. */
         slen = s->length;

         /* Create new string buffer if needed */
         #if STRINGA_TRACE_COPY
            printf("jstra: __strcpyaa, Allocating %d bytes from \"%s\", for %p\n", slen, stra(s), stra(d));
         #endif /* STRINGA_TRACE_COPY */
         __strnewminsizea(d, slen);

         /* Copy the string */
         d->length = slen;
         if(d->string != NULL) MEMCPY(d->string, s->string, slen + 1);
      } else if(d->string != NULL) {

         /* Release old pointer */
         #if STRINGA_TRACE_COPY
            printf("jstra: __strcpya, Deleting %p\n", d);
         #endif /* STRINGA_TRACE_COPY */
         __strfreea(d);
      }
      return(d);

   }
#endif /* __need_strcpyaa */



#if __need_cata
   static inline stringa *__strcata(stringa *d, const char *s) {
   /* String-Cat-A
      This function behaves like strcat, except it supports the 
      StringA style of strings. Yay! This means, unfortunately,
      that life will be sheer pure hell! Why? Because we'll have
      to copy the original <d> string to the new string! So be
      warned, this isn't as efficient as strcat() -- but then,
      you can easily have buffer overflows with strcat() :) 
      Note that we don't have to do anything when <s> = NULL 
      or when <*s> = NULL :) 
      
      Warning: The length of *d must be properly set (see strsetlena()) */

      sizea slen;
      sizea dlen;

      if(d != NULL) {
         if(d->string != NULL) {
            /* Destination string isn't NULL */
            if(s != NULL) {
               /* Source string isn't NULL, either */
               slen = STRLEN(s);
               dlen = d->length;
               
               /* Create the new string buffer */
               __strminsizea(d, slen + dlen);
      
               /* Concatenate the strings on in. */
               d->length = slen + dlen;
               if(d->string != NULL) MEMCPY(d->string + dlen, s, slen + 1);
            }
         } else {
            /* Straight string copy */
            __strcpya(d, s);
         }
      }
         
      /* Return destination string */
      return(d);

   }
#endif /* __need_cata */



#if __need_cataa
   static inline stringa *__strcataa(stringa *d, const stringa *s) {
   /* String-Cat-A-To-A
      This function behaves like strcat, except it supports the 
      StringA style of strings. Yay! This means, unfortunately,
      that life will be sheer pure hell! Why? Because we'll have
      to copy the original <d> string to the new string! So be
      warned, this isn't as efficient as strcat() -- but then,
      you can easily have buffer overflows with strcat() :) 
      Note that we don't have to do anything when <s> = NULL 
      or when <*s> = NULL :) 

      This function assumes *s is a StringA string.   

      Warning: The length of *d must be properly set (see strsetlena()) */

      sizea slen;
      sizea dlen;

      if(d != NULL) {
         if(d->string != NULL) {
            /* Destination string isn't NULL */
            if(s != NULL && s->string != NULL) {
               /* Source string isn't NULL, either */
               slen = s->length;
               dlen = d->length;
               
               /* Create the new string buffer */
               __strminsizea(d, slen + dlen);
      
               /* Concatenate the strings on in. */
               d->length = slen + dlen;
               if(d->string != NULL) MEMCPY(d->string + dlen, s->string, slen + 1);
            }
         } else {
            /* Straight string copy */
            __strcpyaa(d, s);
         }
      }
         
      /* Return destination string */
      return(d);

   }
#endif /* __need_cataa */



#endif /* Already included? */
