/* LIBTOKEN                                                 Version 0.1.3
   Copyright (C) 1998   Justin Smith             justins@ugcs.caltech.edu
   ----------------------------------------------------------------------

   This  program  is free software; you can redistribute it and/or modify
   it  under  the terms of the GNU General Public License as published by
   the  Free  Software  Foundation;  either  version 2 of the License, or   
   (at your option) any later version.

   This  program  is  distributed  in  the  hope  that it will be useful,
   but  WITHOUT  ANY  WARRANTY;  without  even  the  implied  warranty of 
   MERCHANTABILITY   or  FITNESS  FOR  A  PARTICULAR  PURPOSE.   See  the  
   GNU General Public License for more details.

   You  should  have  received  a  copy of the GNU General Public License
   along  with  this  program,  in the file COPYING. If not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA

   ----------------------------------------------------------------------
*/


#ifndef __libjtoken_h_included
#define __libjtoken_h_included


// Include files required for the header declarations.
//
#include <libj.h>
#include <libjfile.h>
#include <libjlink.h>
         

// Disable debugging flags unless the generic LIBTOKEN_DEBUG
// flag has already been specified. This is so you must explicitly
// enable general debugging before libtoken starts lugging its
// debug luggage all over creation. Also, some special debugging
// options may depend on general debugging.
//
#ifndef  LIBTOKEN_DEBUG
#undef   LIBTOKEN_DEBUG_INCLUDE     // Log traversal through the input files
#undef   LIBTOKEN_DEBUG_VAR         // Variable debugging
#undef   LIBTOKEN_DEBUG_EVAL        // Evaluator debugging
#undef   LIBTOKEN_DEBUG_TRACE       // Trace all command calls
#undef   LIBTOKEN_DEBUG_TEXT        // Special state debugging
#undef   LIBTOKEN_DEBUG_DEFINE      // Command definition debugging
#undef   LIBTOKEN_DEBUG_SUICIDAL    // State-info dumps at each character
#endif   // debugging enabled?


// Special command flags to use when you add a new command using the
// tokenAddCommand() function. Each of these flags should be used with
// care and usually are only used internally.
//
#define  TOKEN_CMD_IGNORE_IF_BLOCKS    0x01  // Run cmd even if in failed block
#define  TOKEN_CMD_DO_NOT_EXEC_ARGS    0x02  // Do not auto execute arguments
#define  TOKEN_CMD_IS_CUSTOM           0x04  // Command defined using $$define;


// Special variable flags
//
#define  TOKEN_VAR_INTERNAL            0x01  // If defined, var is "constant"


// Default buffer sizes for various components of a command.
//
#define  TOKEN_CMD_LENGTH              32    // Limit on command name
#define  TOKEN_VAR_LENGTH              32    // Limit on variable name
#define  TOKEN_ARG_LENGTH              32768 // Limit on each argument size


// These are the major error codes returned by libtoken functions. Any code
// other than TOKEN_SUCCESS will usually result in a chain abort of the entire
// tokenising process. erfunc() has the power to block an error code if it
// decides the error is nonfatal. Most functions, on an error, will call
// printFileError() and then will block the error code so they can continue
// processing.
//
#define  TOKEN_SUCCESS                 0     // No errors
#define  TOKEN_ERROR_INVALID_SYMBOL    1     // Invalid symbol in var/cmd/arg
#define  TOKEN_ERROR_EOF               2     // Unexpected end of file
#define  TOKEN_ERROR_FOPEN_IN          3     // Error opening main input file
#define  TOKEN_ERROR_FOPEN_OUT         4     // Error opening main output file
#define  TOKEN_ERROR_COMMAND_NOT_FOUND 5     // Specified command not defined
#define  TOKEN_ERROR_NULL              6     // tokeniser structure is null
#define  TOKEN_ERROR_FOPEN_INCLUDE     7     // Error including a file
#define  TOKEN_ERROR_INTERNAL          8     // Redefining internal function
#define  TOKEN_ERROR_TOO_FEW_ARGS      9     // Too few arguments to command
#define  TOKEN_ERROR_TOO_MANY_ARGS     10    // Too many arguments to command
#define  TOKEN_ERROR_INVALID_ARG       11    // Invalid argument passed to cmd
#define  TOKEN_ERROR_VERSION           12    // Error setting version number


// These definitions are internal to libtoken and are not exported by default
// These are the main state positions we can be in, while reading an input
// file.
//
#ifdef   __LIBTOKEN_SOURCE
#define  TOKEN_STATE_TEXT              0     // Normal, flowing text
#define  TOKEN_STATE_CMD_NAME          1     // After $$ seq, before arguments
#define  TOKEN_STATE_ARGUMENT          2     // Searching for beginning of arg
#define  TOKEN_STATE_IN_ARG            3     // Reading a normal arg
#define  TOKEN_STATE_IN_ARGV           4     // Normal arg - parsing a variable
#define  TOKEN_STATE_IN_ARG_Q          5     // Reading a quoted arg - "arg"
#define  TOKEN_STATE_IN_ARGV_Q         6     // Quoted arg - parsing a variable
#define  TOKEN_STATE_END_ARG_Q         7     // After end quote of quoted arg
#define  TOKEN_STATE_END_ARGS          8     // After closing ), before term ;
#define  TOKEN_STATE_COMMENT           9     // In a $$< ... $$> comment block
#endif   // libtoken source?


// These definitions are internal to libtoken. These are the states the 
// boolean expression analyser can be in.
//
#ifdef   __LIBTOKEN_SOURCE
#define  EXPR_STATE_INIT               0     // Initial state
#define  EXPR_STATE_STRING             1     // Reading a normal string arg
#define  EXPR_STATE_STRING_Q           2     // Reading a quoted string arg
#define  EXPR_STATE_SYMBOL             3     // Reading a symobol( == , != ...)
#endif   // libtoken source?


// These definitions represent the possible symbols a boolean expression can
// have in it.
//
#ifdef   __LIBTOKEN_SOURCE
#define  EXPR_STRING                   0     // Return true if string not empty
#define  EXPR_EQUATE                   1     // Return T if 2 strings equals
#define  EXPR_NOT_EQUATE               2     // Return T if 2 strings not equal
#endif   // libtoken source?


// Definitions for the tokeniser version and the tokeniser copyright 
// information. When the tokeniser is initialised, the contents of these
// defines is used to setup the variables #token_copyright# and 
// #token_version#.
//
#define  TOKEN_COPYRIGHT               "Copyright(c) 1998 Justin Smith; All Rights Reserved."
#define  TOKEN_VERSION                 "Tokeniser v0.1"


// Tokeniser structure. This structure is required in virtually all 
// token functions since it contains the current command list, state variables,
// defined variables, and error information. This structure should be 
// initialised with initTokeniser and freed with freeTokeniser. 
//
typedef struct _tokeniser {

   int (*erfunc)(file *, int, const char *);    // User supplied error function
   slinknode *commands;          // All commands (compiled and custom defined)
   slinknode *variables;         // All defined variables
   slinknode *iff;               // The "if block" list.
   slinknode *customdefs;        // Track custom params for deletion
   slinknode *nestedfiles;       // Error information on which files are loaded
   int text;                     // State flag; don't echo flowing text if > 0
   int squash;                   // State flag; squash whitespace if > 0
   int sawspace;                 // Used for space squashing
   int newline;                  // State flag; remove newlines if > 0
   long version;                 // Version code for information parsing

} tokeniser;


// The following structures are internal data to libtoken functions.
//
#ifdef   __LIBTOKEN_SOURCE

// Saved state. Used by tokenSaveState() and tokenRestoreState(), usually
// surrounding a call to parseTokens() which would change state data.
//
typedef struct _tokensavedstate {
   
   int text;                     // Saved text state
   int squash;                   // Saved squash state
   int sawspace;                 // Saved sawspace flag
   int newline;                  // Saved newline state
   slinknode *iff;               // Saved if-block stack
   
} tokensavedstate;


// Token Command entry.
//
typedef struct _tokencmd {

   char *cmdname;                // <- Name of command   v-- function to exec
   int (*exfunc)(tokeniser *, file *, file *, const char *, const slinknode *, void *);
   void *param;                  // Optional parameter to pass to exfunc().
   int flags;                    // Special flags (see TOKEN_CMD_* flags above)

} tokencmd;


// Token Variable entry.
//
typedef struct _tokenvar {

   char *name;                   // Name of variable
   char *value;                  // String representation of the value
   int flags;                    // Special flags - see TOKEN_VAR_* flags.

} tokenvar;


// If-block data structure. A better reference on this is the set of
// functions tokenCmdIf(), tokenCmdElif(), tokenCmdElse(), tokenCmdEndif().
// This is used to calculate if-blocks and to determine, based on conditionals,
// what code should be executed.
//
// Valid states to be in:
//   eval   exec
//   -----  -----  ----------------------------------------------------------
//    F      F      No conditions in the if block have evaluated to true yet.
//                  Keep searching for an if/elif/else block which evaluates
//                  to true.
//    T      T      The current block has evaluated to true and is being 
//                  executed. 
//    T      F      A block in this if set has already evaluated to true, so
//                  do not waste time to evaluate the remaining blocks -or-
//                  this is an if block nested inside of a failed block.
//
typedef struct _tokeniff {

   bool eval;                    // Eval flag
   bool exec;                    // Exec flag

} tokeniff;


// This structure represents a command while it is being parsed by
// parseTokens() function. It is basically a "workspace" for the current
// command, argument, and variable that are being parsed.
//
typedef struct _parsedcommand {
   
   tokeniser *t;                 // Tokeniser variable
   file *fi;                     // Current input file
   file *fo;                     // Current output file
   char *cmdname;                // Buffer for command name
   char *cmdnamep;               // Point to last char in cmd name
   char *cmdarg;                 // Buffer for current argument
   char *cmdargp;                // Point to last char in argument
   char *cmdvar;                 // Buffer for variable being parsed
   char *cmdvarp;                // Point to last char in var name
   slinknode *cmdargs;           // Complete list of arguments for command
   int nestlevel;                // Nesting level; used by nested functions
   slinknode *states;            // linked list of (int *)'s for states
   
} parsedcommand;

// End of the libtoken source structures
//
#endif   // libtoken source?


/*** File Parsing ***/


int parseFile(tokeniser *t, const char *finname, const char *foutname);


/*** Tokeniser structure management ***/


tokeniser *initTokeniser(tokeniser *t);
tokeniser *tokenSetErrorFunction(tokeniser *t, int (*erfunc)(file *, int, const char *));
void freeTokeniser(tokeniser *t);


/*** Command management ***/


void tokenAddCommand(tokeniser *t, const char *cmdname, int (*exfunc)(tokeniser *, file *, file *, const char *, const slinknode *, void *), void *param, int flags);


/*** Variable management ***/


const char *tokenGetVariable(tokeniser *t, const char *vname);
const char *tokenSetVariable(tokeniser *t, const char *vname, const char *vvalue);
const char *tokenSetIntVariable(tokeniser *t, const char *vname, const char *vvalue);
const char *tokenUnsetVariable(tokeniser *t, const char *vname);


/*** Special States ***/


bool tokenGetText(tokeniser *t);
bool tokenSetText(tokeniser *t, bool on);
bool tokenResetText(tokeniser *t);
bool tokenGetSquash(tokeniser *t);
bool tokenSetSquash(tokeniser *t, bool on);
bool tokenResetSquash(tokeniser *t);
bool tokenGetNewline(tokeniser *t);
bool tokenSetNewline(tokeniser *t, bool on);
bool tokenResetNewline(tokeniser *t);


/*** Version Information ***/


bool tokenSetVersion(tokeniser *t, long version);
long tokenVersionSupported(long version);
long tokenVersion(tokeniser *t);
long tokenRealVersion();
long tokenVersionFromStr(const char *version);
char *tokenStrFromVersion(char *r, long version);


unsigned int libjtoken_version(void);


#endif
