/* LIBTOKEN                                                 Version 0.1.3
   Copyright (C) 1998   Justin Smith             justins@ugcs.caltech.edu
   ----------------------------------------------------------------------

   This  program  is free software; you can redistribute it and/or modify
   it  under  the terms of the GNU General Public License as published by
   the  Free  Software  Foundation;  either  version 2 of the License, or   
   (at your option) any later version.

   This  program  is  distributed  in  the  hope  that it will be useful,
   but  WITHOUT  ANY  WARRANTY;  without  even  the  implied  warranty of 
   MERCHANTABILITY   or  FITNESS  FOR  A  PARTICULAR  PURPOSE.   See  the  
   GNU General Public License for more details.

   You  should  have  received  a  copy of the GNU General Public License
   along  with  this  program,  in the file COPYING. If not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA

   ----------------------------------------------------------------------

   INFORMATION
   ============
   
   libtoken is a simple tokenising library designed to parse "command
   strings" embedded withing ASCII text files. These tokens should be
   delimited in the form:
   
   $$command;                       Simple command
   $$command();                      "
   $$command(argument);             Command with a text argument
   $$command(argument, argument);   Command with multiple text arguments
   
   Argument strings may be enclosed in double-quotes, and within quotes
   standard ANSI escape sequences are recognized (outside the quotes,
   ANSI escape sequences are not recognised). 
   
   The libtoken concept originates from the makehtml compiler, which has
   been discontinued in favor of this project.
   

   COMMANDS
   ---------
   
   $$; or $$();            Replace this sequence with the "$$" sequence.
   
   $$echo(argument);       Echoes the text specified. Multiple arguments
   $$echo(arg, arg, arg);  may be specified. Any argument may be a variable
                           substitution string, of the form #varname#.
                           Commands embedded in the echo command will NOT
                           be executed. 
                        
   $$exec(argument);       Executes the argument specified. Similar in
                           function to the $$echo command, except that
                           commands embedded in the argument WILL be
                           executed.
                           
   $$get(varname);         Echoes the contents of the specified variable
   $$get(varname, varname);name to the output file.
                        
   $$set(varname, value);  Sets the variable name to the specified value.
   $$set(varname, value,   varname should NOT be enclosed in #'s.
         varname, value);  
                           
   $$unset(varname);       Unsets the variable specified. varname should
   $$unset(vname, vname);  NOT be enclosed in #'s. 
   
   $$getenv(varname);      Echoes the contents of the specified environment
   $$getenv(vname, vname); variable(s) to the output file. This is currently
                           the only way to access environment variables 
                           (there is no #varname#-equivalent notation).
   
   $$setenv(varname, val); Sets an environment variable to the specified
   $$setenv(varname, val,  value. 
            varname, val);
            
   $$unsetenv(varname);    Unsets the specified environment variable(s).
   $$unsetenv(vnam, vnam);
   
   $$include(filename);    Includes the specified file(s). Each file will
                           also be parsed.
                           
   $$include_text(fname);  Includes the specified file(s), but the files
                           read in will be written verbatim, and will not
                           be considered tokenised files.
                           
   $$iif(expr,true,false); Evaluate the expression, print out the contents
                           of "true" if the expression is true and "false"
                           if the expression is false.
                           
   $$comment(blah);        Comment
   
   $$<                     Begins a comment block. Text inside the comment
   $$>                     block will not be written to the output file,
                           until an $$> is encountered. Commands inside the
                           comment block will not be executed. Comment blocks
                           terminate automatically at the end of the file.
                           Nested comment blocks are allowed.
                           
   $$text(on);             Defaults to whether normal text is echoed or
   $$text(off);            not. If "on", normal text is written to the output
   $$text(reset);          file. If "off", only command output is written to
                           the output file. This option is "on" by default
                           and can be changed with the tokenSetText() function.
                           This command nests itself; that is, each call to
                           $$text(off) will increment a counter and each call
                           to $$text(on) decrements the counter. Use 
                           $$text(reset) to reset the counter and force
                           text to be displayed (this is equivalent to the
                           tokenResetText() function). Text echo status will
                           NOT carry over the end of the current file.
                           
   $$squash(on);           Defaults to off. If "on", then whitespace in
   $$squash(off);          normal flowing text will be squashed into smaller
   $$squash(reset);        whitespace. Multiple spaces and tabs in a row
                           will be squashed to a single space character.
                           This does not affect command output, and does
                           not carry across files.
   
   $$newline(on);          Defaults to on. If "off", newlines in normal
   $$newline(off);         flowing text will be removed. This does not
   $$newline(reset);       affect command output, and does not carry.
                           
   $$if(expr);             Begins an if block. If any of the expressions
   $$if(expr, expr, expr); specified return true, then the block of text
                           after the if statement is executed (see below).
                           Note that if this condition fails, the text after
                           the if statement will NOT be written to the
                           output file, and commands will still be parsed
                           for syntax but will NOT be executed. The "block
                           of text" is defined as the block after the $$if
                           command but before the $$else, $$elif, or $$endif
                           statement at the appropriate level, or until the
                           end of the file. $$if; and $$if(); evaluate always
                           as false. If blocks may be nested. If blocks
                           always terminate at the end of the current file.
                           
   $$elif(expr);           Similar to the $$if statement above, but only
   $$elif(expr, expr);     evaluated if the matching $$if and any preceding
                           $$elif statements evaluate as false. If an $$elif
                           is encountered without a matching $$if, then it
                           is treated as an $$if statement. Not required.
                           
   $$else;                 Only evaluated if all preceding $$if/$$elif
                           blocks return false. Not required.
                           
   $$endif;                Required to terminate the if block.
   
   $$define(fnname, dfn);  Defines a function (macro). Once a function is  
   $$define(fnname, dfn,   defined, you can invoke it with $$fnname();
            fnname, dfn);  The sequence will be replaced by the text "dfn",
                           and then dfn will be evaluated as normal text.
                           Argument strings are recognised with $$.arg1;,
                           $$.arg2;, ... $$.arg9;, and the command name 
                           is stored in $$.arg0;. $$.args; will print all
                           of the arguments separated by spaces. $$.argx;
                           will print all unreferenced arguments, separated
                           by spaces. Note that you can also use the special
                           variables #.arg0#, #.arg1#, ... to reference
                           arguments, each method has implications. The
                           $$.arg*; commands will be completely reprocessed
                           while the #.arg*# variables will not. Unless you
                           know what you are doing you probably want to
                           use the #.arg*$ variables.
                           
   $$undef(fnname);        Undefines a function. The function cannot be an
   $$undef(fnname, fnname);internal function (one internal to the tokeniser
                           or a function compiled in by the parent program)
                           and must have been previously declared with 
                           $$define.
                           
   $$dump_commands;        Program debugging commands.
   $$dump_variables;
   $$dump_state;
   $$dump_versions;
   
   $$add(val1, val2);      Basic math operations. Invalid values will
   $$subtract(val1, val2); evaluate to zero, but in the future I will
   $$multiply(val1, val2); have things set up so they produce an error
   $$divide(val1, val2);   instead. These are floating point functions.
   $$negate(val);
   $$exp(val);
   $$ln(val);
   
   Note that command names must be at most TOKEN_CMD_LENGTH characters
   long. Command names cannot have the following symbols:  ( ) $ " ' , ;
   Whitespace characters appearing anywhere in the command name will be
   dropped (including space, tab, newline).


   VARIABLES
   ----------
   
   Variables may be indicated in argument strings with the notation
   #varname#. For example, the following code (ignore the newlines)
   will emit:
   
   code                             result
   -----                            -------
   $$set(hello, "Hello, world!");   :) Hello, World! :)
   $$echo(":) #hello# :)");
   
   Note the special variable ## is defined to be equal to a single #.
   Variable expansion occurs before command execution.
   

   EXPRESSIONS
   ------------
   
   Expressions must follow one of the following forms:
   
   string                     Returns true if the string is nonempty and
   "string"                   not all whitespace.

   string1 == string2         Returns true if string1 equals string2,
   "string1" == "string2"     excepting leading and trailing whitespace
                              on each string.
                              
   string1 != string2         Returns true if string1 does not equal string2,
   "string1" != "string2"     excepting leading and trailing whitespace
                              on each string.
   

   VERSION HISTORY
   ================
   
   pre 0.1        This is basically the equivalent of the last known
                  version of makehtml, somewhere around version 2 or so.
                  A few sources from makehtml were salvaged for the 
                  tokeniser, but most of the tokeniser was written from
                  scratch.
                  
   0.1            The tokeniser is written
   
   0.1.1          Version information added.
                  Nested commands are handled completely different now -
                    f.e., escape sequences in a nested command are not
                    expanded. 
                  Processing of $$define'd commands is drastically
                    different, using variable arguments to prevent
                    escaping.
       
   0.1.2          The single, 107K source file is split into several
                  source files so the damn thing will compile faster.                 
                  
   0.1.3          Fixed to work with libfile v2                  


   SYMBOLIC DEFINES
   =================
   
   __cplusplus             Use C++ where available.
   
   X86_ASM                 Use X86 assembly if available.
   
   NO_MATH_LIB             Disable use of a math library entirely.
                           (disables all math functions as well)
   
   READBUF                 Sets the default file buffer size.

   LIBTOKEN_NO_OUTPUT      Supress printing of error messages to stderr.
   
   LIBTOKEN_DEBUG          Enable libtoken debugging.
   LIBTOKEN_DEBUG_INCLUDE
   LIBTOKEN_DEBUG_VAR
   LIBTOKEN_DEBUG_EVAL
   LIBTOKEN_DEBUG_TEXT
   LIBTOKEN_DEBUG_TRACE
   LIBTOKEN_DEBUG_SUICIDAL
   
   ----------------------------------------------------------------------
*/


#define __LIBTOKEN_SOURCE
#include <libjtoken.h>
#include <libtoken_cmd.h>
#include <libtoken_parser.h>
#include <libtoken_supp.h>
#include <libjlink.h>
#include <libjmem.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef NO_MATH_LIB
#include <math.h>
#endif  // math lib?


static unsigned long tokenVersions[] = {
// MMmmppll (M)ajor (m)inor (p)atch (l)etter
   0x00010200,          // 0.1.2    - sources split
   0x00010100,          // 0.1.1    - development - version info added
   0x00010000,          // 0.1      - development
   0
};


static char erstr[32768];
static int ercode;


/*** Internal Variable Accessors ***/


char *get_erstr() {
/* Get-erstr */

   return(erstr);
   
}


int *get_ercode() {
/* Get-ercode */

   return(&ercode);
   
}


unsigned long *get_token_versions() {
/* Get-Token-Versions */

   return(tokenVersions);
   
}


/*** Major Interface Functions ***/


tokeniser *initTokeniser(tokeniser *t) {
/* Init-Tokeniser */

   if(!t) return(NULL);
   t->erfunc = NULL;
   t->commands = NULL;
   t->variables = NULL;
   t->iff = NULL;
   t->customdefs = NULL;
   t->nestedfiles = NULL;
   t->version = *tokenVersions;
   
   // Basic commands
   tokenAddCommand(t, "", &tokenCmdEmpty, NULL, 0);
   tokenAddCommand(t, "echo", &tokenCmdEcho, NULL, TOKEN_CMD_DO_NOT_EXEC_ARGS);
   tokenAddCommand(t, "exec", &tokenCmdExec, NULL, 0);
   tokenAddCommand(t, "comment", &tokenCmdComment, NULL, TOKEN_CMD_DO_NOT_EXEC_ARGS);

   // Variables
   tokenAddCommand(t, "get", &tokenCmdGet, NULL, 0);
   tokenAddCommand(t, "set", &tokenCmdSet, NULL, 0);
   tokenAddCommand(t, "unset", &tokenCmdUnset, NULL, 0);
   tokenAddCommand(t, "getenv", &tokenCmdGetEnv, NULL, 0);
   tokenAddCommand(t, "setenv", &tokenCmdSetEnv, NULL, 0);
   tokenAddCommand(t, "unsetenv", &tokenCmdUnsetEnv, NULL, 0);

   // File inclusion
   tokenAddCommand(t, "include", &tokenCmdInclude, NULL, 0);
   tokenAddCommand(t, "include_text", &tokenCmdIncludeText, NULL, 0);

   // Flow control
   tokenAddCommand(t, "iif", &tokenCmdIif, NULL, TOKEN_CMD_DO_NOT_EXEC_ARGS);
   tokenAddCommand(t, "if", &tokenCmdIf, NULL, TOKEN_CMD_IGNORE_IF_BLOCKS);
   tokenAddCommand(t, "elif", &tokenCmdElif, NULL, TOKEN_CMD_IGNORE_IF_BLOCKS);
   tokenAddCommand(t, "else", &tokenCmdElse, NULL, TOKEN_CMD_IGNORE_IF_BLOCKS);
   tokenAddCommand(t, "endif", &tokenCmdEndif, NULL, TOKEN_CMD_IGNORE_IF_BLOCKS);
   
   // Special state
   tokenAddCommand(t, "text", &tokenCmdText, NULL, 0);
   tokenAddCommand(t, "squash", &tokenCmdSquash, NULL, 0);
   tokenAddCommand(t, "newline", &tokenCmdNewline, NULL, 0);
   
   // Command definition
   tokenAddCommand(t, "define", &tokenCmdDefine, NULL, TOKEN_CMD_DO_NOT_EXEC_ARGS);
   tokenAddCommand(t, "undef", &tokenCmdUndef, NULL, TOKEN_CMD_DO_NOT_EXEC_ARGS);

   // Runtime debugging
   tokenAddCommand(t, "dump_commands", &tokenCmdDumpCmd, NULL, TOKEN_CMD_IGNORE_IF_BLOCKS);
   tokenAddCommand(t, "dump_variables", &tokenCmdDumpVar, NULL, TOKEN_CMD_IGNORE_IF_BLOCKS);
   tokenAddCommand(t, "dump_state", &tokenCmdDumpState, NULL, TOKEN_CMD_IGNORE_IF_BLOCKS);
   tokenAddCommand(t, "dump_versions", &tokenCmdDumpVersions, NULL, TOKEN_CMD_IGNORE_IF_BLOCKS);

   // Math commands
   #ifndef NO_MATH_LIB
   tokenAddCommand(t, "add", &tokenCmdAdd, NULL, 0);
   tokenAddCommand(t, "subtract", &tokenCmdSubtract, NULL, 0);
   tokenAddCommand(t, "multiply", &tokenCmdMultiply, NULL, 0);
   tokenAddCommand(t, "divide", &tokenCmdDivide, NULL, 0);
   tokenAddCommand(t, "negate", &tokenCmdNegate, NULL, 0);
   tokenAddCommand(t, "exp", &tokenCmdExp, NULL, 0);
   tokenAddCommand(t, "ln", &tokenCmdLn, NULL, 0);
   #endif 
   
   // Version support
   tokenAddCommand(t, "get_version", &tokenCmdGetVersion, NULL, 0);
   tokenAddCommand(t, "set_version", &tokenCmdSetVersion, NULL, 0);
   tokenAddCommand(t, "supported_version", &tokenCmdSupportedVersion, NULL, 0);
   tokenAddCommand(t, "get_real_version", &tokenCmdGetRealVersion, NULL, 0);
   tokenAddCommand(t, "version_id", &tokenCmdVersionID, NULL, 0);
   
   // Basic variables
   tokenSetIntVariable(t, "", "#");
   tokenSetIntVariable(t, "token_version", TOKEN_VERSION);
   tokenSetIntVariable(t, "token_copyright", TOKEN_COPYRIGHT);

   // Math constants
   #ifndef NO_MATH_LIB
   tokenSetIntVariable(t, "M_E",          "2.7182818284590452354");  // e
   tokenSetIntVariable(t, "M_LOG2E",      "1.4426950408889634074");  // log2 e
   tokenSetIntVariable(t, "M_LOG10E",     "0.43429448190325182765"); // log10 e
   tokenSetIntVariable(t, "M_LN2",        "0.69314718055994530942"); // ln 2
   tokenSetIntVariable(t, "M_LN10",       "2.30258509299404568402"); // ln 10
   tokenSetIntVariable(t, "M_PI",         "3.14159265358979323846"); // pi
   tokenSetIntVariable(t, "M_PI_2",       "1.57079632679489661923"); // pi/2
   tokenSetIntVariable(t, "M_PI_4",       "0.78539816339744830962"); // pi/4
   tokenSetIntVariable(t, "M_1_PI",       "0.31830988618379067154"); // 1/pi
   tokenSetIntVariable(t, "M_2_PI",       "0.63661977236758134308"); // 2/pi
   tokenSetIntVariable(t, "M_2_SQRTPI",   "1.12837916709551257390"); // 2/sqrt(pi)
   tokenSetIntVariable(t, "M_SQRT2",      "1.41421356237309504880"); // sqrt(2)
   tokenSetIntVariable(t, "M_SQRT1_2",    "0.70710678118654752440"); // 1/sqrt(2)
   #endif
   
   // Special argument processing commands
   tokenAddCommand(t, ".arg0", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg1", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg2", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg3", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg4", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg5", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg6", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg7", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg8", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".arg9", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".args", &tokenCmdArg, NULL, 0);
   tokenAddCommand(t, ".argx", &tokenCmdArg, NULL, 0);
   
   // Special argument processing variables
   tokenSetIntVariable(t, ".arg0", "#.arg0#");
   tokenSetIntVariable(t, ".arg1", "#.arg1#");
   tokenSetIntVariable(t, ".arg2", "#.arg2#");
   tokenSetIntVariable(t, ".arg3", "#.arg3#");
   tokenSetIntVariable(t, ".arg4", "#.arg4#");
   tokenSetIntVariable(t, ".arg5", "#.arg5#");
   tokenSetIntVariable(t, ".arg6", "#.arg6#");
   tokenSetIntVariable(t, ".arg7", "#.arg7#");
   tokenSetIntVariable(t, ".arg8", "#.arg8#");
   tokenSetIntVariable(t, ".arg9", "#.arg9#");
   tokenSetIntVariable(t, ".args", "#.args#");
   tokenSetIntVariable(t, ".argx", "#.argx#");

   // Initialise state
   tokenResetText(t);
   tokenResetSquash(t);
   tokenResetNewline(t);
   return(t);

}


tokeniser *tokenSetErrorFunction(tokeniser *t, int (*erfunc)(file *, int, const char *)) {
/* Token-Set-Error-Function */

   if(!t) return(NULL);
   t->erfunc = erfunc;
   return(t);

}


void freeTokeniser(tokeniser *t) {
/* Free-Tokeniser */

   if(!t) return;
   slink_release(&t->commands);
   slink_release(&t->variables);
   slink_release(&t->iff);
   slink_release(&t->customdefs);
   slink_release(&t->nestedfiles);

}


int parseFile(tokeniser *t, const char *finname, const char *foutname) {
/* Parse-File 
   Entry-point function for file parsing. */
   
   file *fi;
   file *fo;
   
   if(!t || !finname) {
      #ifndef LIBTOKEN_NO_OUTPUT
      fprintf(stderr, "token: null tokeniser, or filename variable.\n");
      #endif
      return(TOKEN_ERROR_NULL);
   }
   
   // Open each file
   if(!(fi = file_create(finname, 'r', READBUF))) {
      sprintf(erstr, "Cannot open input file %s", finname);
      ercode = TOKEN_ERROR_FOPEN_IN;
      printFileError(t);
      return(TOKEN_ERROR_FOPEN_IN);
   }
   if(foutname) {
      if(!(fo = file_create(foutname, 'w', READBUF))) {
         sprintf(erstr, "Cannot open output file %s", foutname);
         ercode = TOKEN_ERROR_FOPEN_OUT;
         printFileError(t);
         file_destroy(&fi);
         return(TOKEN_ERROR_FOPEN_OUT);
      }   
   } else {
      if(!(fo = file_create_null("(null)", IO_WRITEONLY))) {
         sprintf(erstr, "Cannot create null output file");
         ercode = TOKEN_ERROR_FOPEN_OUT;
         printFileError(t);
         file_destroy(&fi);
         return(TOKEN_ERROR_FOPEN_OUT);
      }   
   }
         
   // Parse tokens in the input file, write to the output file.
   if((ercode = parseTokens(t, fi, fo))) {
      file_destroy(&fi);
      file_destroy(&fo);
      return(ercode);
   }
   
   // Close each file
   file_destroy(&fi);
   file_destroy(&fo);
   return(TOKEN_SUCCESS);
   
}


/*** Command Interface ***/


void tokenAddCommand(tokeniser *t, const char *cmdname, int (*exfunc)(tokeniser *, file *, file *, const char *, const slinknode *, void *), void *param, int flags) {
/* Token-Add-Command */

   char cmd[TOKEN_CMD_LENGTH + 1];
   tokencmd *c;
   
   if(!t) return;
   
   // Remove invalid symbols from the command name.
   tokenCommandValidate(cmd, cmdname);
   cmdname = cmd;
   
   // Add the command
   c = (tokencmd *)cmdExists(t, cmdname);
   if(c) {
      // Command is already in the linked list. Update the command
      // instead of adding a duplicate entry :)
      c->exfunc = exfunc;
      c->param = param;
      c->flags = flags;
      return;
   }

   // Creating a new command from scratch.  
   c = (tokencmd *)malloc(sizeof(tokencmd));
   c->cmdname = (char *)malloc(sscan_null(cmdname) + 1);
   scopy(c->cmdname, cmdname);
   c->exfunc = exfunc;
   c->param = param;
   c->flags = flags;
   slink_prepend_head(&t->commands, c, &tokenDestroyCommand);
   return;

}


/*** Variable Interface ***/


const char *tokenGetVariable(tokeniser *t, const char *vname) {
/* Token-Get-Variable */

   char vnam[TOKEN_VAR_LENGTH + 1];
   slinknode *cur;
   tokenvar *v;
   
   if(!t || !vname) return(NULL);

   // Remove invalid symbols from the variable name.
   tokenCommandValidate(vnam, vname);
   vname = vnam;
   
   // Search for the specified variable
   cur = t->variables;
   while(cur) {
      v = (tokenvar *)cur->data;
      if(!(scmp(v->name, vname) ^ 0xffffffff)) {
         // Variable found.
         #ifdef LIBTOKEN_DEBUG_VAR
         printf("get: '%s' = '%s'\n", v->name, v->value);
         #endif // debug
         return(v->value);
      }
      cur = cur->next;
   }
   
   // Variable not found.
   return("");

}


const char *tokenSetVariable(tokeniser *t, const char *vname, const char *vvalue) {
/* Token-Set-Variable */

   char vnam[TOKEN_VAR_LENGTH + 1];
   tokenvar *v;
   
   if(!t || !vname) return(NULL);

   // Remove invalid symbols from the variable name.
   tokenCommandValidate(vnam, vname);
   vname = vnam;
   
   // Search for the specified variable
   v = varExists(t, vname);
   if(v) {
      // Variable found.
      free(v->value);
      v->value = (char *)malloc(sscan_null(vvalue) + 1);
      scopy(v->value, vvalue);
      v->flags = 0;
      #ifdef LIBTOKEN_DEBUG_VAR
      printf("set: '%s' = '%s'\n", v->name, v->value);
      #endif // debug
      return(v->value);
   }
   
   // Variable not found.
   v = (tokenvar *)malloc(sizeof(tokenvar));
   v->name = (char *)malloc(sscan_null(vname) + 1);
   scopy(v->name, vname);
   v->value = (char *)malloc(sscan_null(vvalue) + 1);
   scopy(v->value, vvalue);
   v->flags = 0;
   slink_prepend_head(&t->variables, v, &tokenDestroyVariable);
   #ifdef LIBTOKEN_DEBUG_VAR
   printf("set: '%s' = '%s'\n", v->name, v->value);
   #endif // debug
   return(v->value);

}


const char *tokenSetIntVariable(tokeniser *t, const char *vname, const char *vvalue) {
/* Token-Set-Internal-Variable */

   char vnam[TOKEN_VAR_LENGTH + 1];
   tokenvar *v;
   
   if(!t || !vname) return(NULL);

   // Remove invalid symbols from the variable name.
   tokenCommandValidate(vnam, vname);
   vname = vnam;
   
   // Search for the specified variable
   v = varExists(t, vname);
   if(v) {
      // Variable found.
      free(v->value);
      v->value = (char *)malloc(sscan_null(vvalue) + 1);
      scopy(v->value, vvalue);
      v->flags = TOKEN_VAR_INTERNAL;
      #ifdef LIBTOKEN_DEBUG_VAR
      printf("set: '%s' = '%s'\n", v->name, v->value);
      #endif // debug
      return(v->value);
   }
   
   // Variable not found.
   v = (tokenvar *)malloc(sizeof(tokenvar));
   v->name = (char *)malloc(sscan_null(vname) + 1);
   scopy(v->name, vname);
   v->value = (char *)malloc(sscan_null(vvalue) + 1);
   scopy(v->value, vvalue);
   v->flags = TOKEN_VAR_INTERNAL;
   slink_prepend_head(&t->variables, v, &tokenDestroyVariable);
   #ifdef LIBTOKEN_DEBUG_VAR
   printf("set: '%s' = '%s'\n", v->name, v->value);
   #endif // debug
   return(v->value);

}


const char *tokenUnsetVariable(tokeniser *t, const char *vname) {
/* Token-Unset-Variable */

   char vnam[TOKEN_VAR_LENGTH + 1];
   slinknode *cur;
   tokenvar *v;
   
   if(!t || !vname) return(NULL);

   // Remove invalid symbols from the variable name.
   tokenCommandValidate(vnam, vname);
   vname = vnam;
   
   // Search for the specified variable
   cur = t->variables;
   while(cur) {
      v = (tokenvar *)cur->data;
      if(!(scmp(v->name, vname) ^ 0xffffffff)) {
         // Variable found.
         free(v->value);
         v->value = (char *)malloc(1);
         *v->value = 0;
         #ifdef LIBTOKEN_DEBUG_VAR
         printf("unset: '%s' = '%s'\n", v->name, v->value);
         #endif // debug
         return(v->value);
      }
      cur = cur->next;
   }
   
   // Variable not found.
   return("");

}


/*** Special Options ***/


bool tokenGetText(tokeniser *t) {
/* Token-Get-Text */ 
   
   if(!t) return(false);
   return(t->text <= 0);

}


bool tokenSetText(tokeniser *t, bool on) {
/* Token-Set-Text 
   If on is true, decrement text counter. Otherwise, increment the counter.
   Returns whether text will be echoed. */
   
   if(!t) return(false);
   if(on) t->text--;
   else t->text++;
   return(t->text <= 0);

}


bool tokenResetText(tokeniser *t) {
/* Token-Reset-Text */
   
   if(!t) return(false);
   t->text = 0;
   return(t->text <= 0);

}


bool tokenGetSquash(tokeniser *t) {
/* Token-Get-Squash */ 
   
   if(!t) return(false);
   return(t->squash <= 0);

}


bool tokenSetSquash(tokeniser *t, bool on) {
/* Token-Set-Squash 
   If on is true, decrement squash counter. Otherwise, increment the counter.
   Returns whether squash will be echoed. */
   
   if(!t) return(false);
   if(on) t->squash++;
   else t->squash--;
   return(t->squash <= 0);

}


bool tokenResetSquash(tokeniser *t) {
/* Token-Reset-Squash */
   
   if(!t) return(false);
   t->squash = 0;
   t->sawspace = 0;
   return(t->squash <= 0);

}


bool tokenGetNewline(tokeniser *t) {
/* Token-Get-Newline */ 
   
   if(!t) return(false);
   return(t->newline <= 0);

}


bool tokenSetNewline(tokeniser *t, bool on) {
/* Token-Set-Newline 
   If on is true, decrement newline counter. Otherwise, increment the counter.
   Returns whether newline will be echoed. */
   
   if(!t) return(false);
   if(on) t->newline--;
   else t->newline++;
   return(t->newline <= 0);

}


bool tokenResetNewline(tokeniser *t) {
/* Token-Reset-Newline */
   
   if(!t) return(false);
   t->newline = 0;
   return(t->newline <= 0);

}


/*** Version Information ***/


bool tokenSetVersion(tokeniser *t, long version) {
/* Token-Set-Version */

   long *tv = tokenVersions;
   
   if(!t) return(false);
   
   while(*tv) {
      if(*tv == version) {
         t->version = version;
         return(true);
      }
      tv++;
   }

   return(false);

}


long tokenVersionSupported(long version) {
/* Token-Version-Supported */

   long *tv = tokenVersions;
   
   while(*tv && *tv > version) tv++;
   return(*tv);

}


long tokenVersion(tokeniser *t) {
/* Token-Version */

   if(t) return(t->version);
   return(0);
   
}


long tokenRealVersion() {
/* Token-Real-Version */

   return(*tokenVersions);
   
}


long tokenVersionFromStr(const char *version) {
/* Token-Version-From-String */

   long r = 0;
   int state = 0;
   char *pi;
   char *ps;
   char *p;

   if(!version) return(0);
   pi = (char *)malloc(sscan_null(version) + 1);
   scopy(pi, version);
   ps = pi;
   p = pi;

   while(*p) {
      if(*p == '.') {
         *p = 0;
         switch(state) {
            case 0:
               r = r | ((atoi(ps) & 0xff) << 24);               
               break;
            case 1:
               r = r | ((atoi(ps) & 0xff) << 16);
               break;
            case 2:
               r = r | ((atoi(ps) & 0xff) << 8);
               break;
         }
         state++;
         ps = p + 1;
      }
      p++;
   }
   if(p != ps) {
      switch(state) {
         case 0:
            r = r | ((atoi(ps) & 0xff) << 24);               
            break;
         case 1:
            r = r | ((atoi(ps) & 0xff) << 16);
            break;
         case 2:
            r = r | ((atoi(ps) & 0xff) << 8);
            break;
      }
   }
   
   free(pi);
   return(r);
   
}


char *tokenStrFromVersion(char *r, long version) {
/* Token-String-From-Version */

   sprintf(r, "%ld.%ld.%ld", ((version >> 24) & 0xff), ((version >> 16) & 0xff), ((version >> 8) & 0xff));
   return(r);

}



unsigned int libjtoken_version(void) {

   return(__libj_version);

}
