/* LIBTOKEN                                                 Version 0.1.2
   Copyright (C) 1998   Justin Smith             justins@ugcs.caltech.edu
   ----------------------------------------------------------------------

   This  program  is free software; you can redistribute it and/or modify
   it  under  the terms of the GNU General Public License as published by
   the  Free  Software  Foundation;  either  version 2 of the License, or   
   (at your option) any later version.

   This  program  is  distributed  in  the  hope  that it will be useful,
   but  WITHOUT  ANY  WARRANTY;  without  even  the  implied  warranty of 
   MERCHANTABILITY   or  FITNESS  FOR  A  PARTICULAR  PURPOSE.   See  the  
   GNU General Public License for more details.

   You  should  have  received  a  copy of the GNU General Public License
   along  with  this  program,  in the file COPYING. If not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA

   ----------------------------------------------------------------------
*/


#define __LIBTOKEN_SOURCE
#include <libjtoken.h>
#include <libtoken_int.h>
#include <libtoken_supp.h>
#include <libtoken_parser.h>
#include <libtoken_expr.h>
#include <libjlink.h>
#include <libjmem.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef NO_MATH_LIB
#include <math.h>
#endif  // math lib?


/*** Internal Commands - Fundamental Set ***/


int tokenCmdEmpty(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-$$ */

   file_write_line(fo, "$$");
   return(TOKEN_SUCCESS);

}


int tokenCmdEcho(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Echo */

   const slinknode *cur = cmdargs;
   while(cur) {
      file_write_line(fo, (char *)cur->data);
      cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdExec(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Exec */

   const slinknode *cur = cmdargs;
   while(cur) {
      file_write_line(fo, (char *)cur->data);
      cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdComment(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Comment */

   return(TOKEN_SUCCESS);

}


/*** Commands: Variable manipulation ***/


int tokenCmdGet(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Get */

   const slinknode *cur = cmdargs;
   while(cur) {
      file_write_line(fo, tokenGetVariable(t, (char *)cur->data));
      cur = cur->next;
      if(cur) file_write_line(fo, " ");
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdSet(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Set */

   const slinknode *cur = cmdargs;
   tokenvar *v;
   while(cur) {
      v = varExists(t, (char *)cur->data);
      if(!v || !intVar(v)) {
         if(!cur->next) tokenSetVariable(t, (char *)cur->data, "");
         else tokenSetVariable(t, (char *)cur->data, (char *)cur->next->data);
      } else if(v) {
         // Variable is internal
         snprintf(get_erstr(), 32768, "\"%s\": Cannot redefine internal variable \"%s\"", cmdname, (char *)cur->data);
         *get_ercode() = TOKEN_ERROR_INTERNAL;
         printFileError(t);
         if(*get_ercode()) return(*get_ercode());
      }
      cur = cur->next;
      if(cur) cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdUnset(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Unset */

   const slinknode *cur = cmdargs;
   tokenvar *v;
   while(cur) {
      v = varExists(t, (char *)cur->data);
      if(!v || !intVar(v)) tokenUnsetVariable(t, (char *)cur->data);
      else if(v) {
         // Variable is internal
         snprintf(get_erstr(), 32768, "\"%s\": Cannot undefine internal variable \"%s\"", cmdname, (char *)cur->data);
         *get_ercode() = TOKEN_ERROR_INTERNAL;
         printFileError(t);
         if(*get_ercode()) return(*get_ercode());
      }
      cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdGetEnv(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Get-Environment */

   const slinknode *cur = cmdargs;
   const char *evar;
   while(cur) {
      evar = getenv((char *)cur->data);
      if(evar) file_write_line(fo, evar);
      cur = cur->next;
      if(cur) file_write_line(fo, " ");
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdSetEnv(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Set-Environment */

   const slinknode *cur = cmdargs;
   while(cur) {
      if(!cur->next) setenv((char *)cur->data, "", true);
      else setenv((char *)cur->data, (char *)cur->next->data, true);
      cur = cur->next;
      if(cur) cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdUnsetEnv(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Unset-Environment */

   const slinknode *cur = cmdargs;
   while(cur) {
      unsetenv((char *)cur->data);
      cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


/*** Commands: File Inclusion ***/


int tokenCmdInclude(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Include */

   const slinknode *cur = cmdargs;
   file *fin;
   tokensavedstate ss;
   
   while(cur) {
      #ifdef LIBTOKEN_DEBUG_INCLUDE
      printf("Including file %s ...\n", (char *)cur->data);
      #endif // debug
      // Open new (temporary) input file
      if(!(fin = file_create((char *)cur->data, 'r', READBUF))) {
         snprintf(get_erstr(), 32768, "Cannot include token file %s", (char *)cur->data);
         *get_ercode() = TOKEN_ERROR_FOPEN_INCLUDE;
         printFileError(t);
         if(*get_ercode()) return(*get_ercode());
      } else {
         // Parse tokens in the input file, write to the output file.
         tokenSaveState(&ss, t);
         if((*get_ercode() = parseTokens(t, fin, fo))) {
            tokenRestoreState(t, &ss);
            file_destroy(&fin);
            return(*get_ercode());
         }
         tokenRestoreState(t, &ss);
      
         // Close each file
         file_destroy(&fin);
      }
      
      cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdIncludeText(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Include-Text */

   const slinknode *cur = cmdargs;
   file *fin;
   char ch;
   
   while(cur) {
      #ifdef LIBTOKEN_DEBUG_INCLUDE
      printf("Including text file %s ...\n", (char *)cur->data);
      #endif // debug
      // Open new (temporary) input file
      if(!(fin = file_create((char *)cur->data, 'r', READBUF))) {
         snprintf(get_erstr(), 32768, "Cannot include text file %s", (char *)cur->data);
         *get_ercode() = TOKEN_ERROR_FOPEN_INCLUDE;
         printFileError(t);
         if(*get_ercode()) return(*get_ercode());
      } else {
         // Parse tokens in the input file, write to the output file.
         while(file_read_char(fin, &ch)) file_write_char(fo, ch);
      
         // Close each file
         file_destroy(&fin);
      }
      
      cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


/*** Commands: Conditional processing ***/


int tokenCmdIif(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Iif */

   file *fbufi;
   tokensavedstate ss;

   switch(slink_count(cmdargs)) {
      case 0:
      case 1:
         // Too few arguments. Oh well...
         snprintf(get_erstr(), 32768, "%s requires 2 or 3 arguments; %d argument was specified", cmdname, slink_count(cmdargs));
         *get_ercode() = TOKEN_ERROR_TOO_FEW_ARGS;
         if((*get_ercode() = printFileError(t))) return(*get_ercode());
         return(TOKEN_SUCCESS);
         break;
         
      case 2:
      case 3:
         // Only handle the "true" case.
         if(evalExpression((char *)cmdargs->data)) {
            fbufi = file_create_from_source(cmdname, 'r', (char *)cmdargs->next->data, sscan_null(cmdargs->next->data));
            tokenSaveState(&ss, t);
            if((*get_ercode() = parseTokens(t, fbufi, fo))) {
               file_destroy(&fbufi);
               tokenRestoreState(t, &ss);
               return(*get_ercode());
            }
            tokenRestoreState(t, &ss);
            file_destroy(&fbufi);
         } else if(slink_count(cmdargs) >= 3) {
            fbufi = file_create_from_source(cmdname, 'r', (char *)cmdargs->next->next->data, sscan_null(cmdargs->next->next->data));
            tokenSaveState(&ss, t);
            if((*get_ercode() = parseTokens(t, fbufi, fo))) {
               file_destroy(&fbufi);
               tokenRestoreState(t, &ss);
               return(*get_ercode());
            }
            tokenRestoreState(t, &ss);
            file_destroy(&fbufi);
         }
         return(TOKEN_SUCCESS);
         break;
         
      default:
         // Too many arguments
         snprintf(get_erstr(), 32768, "%s requires 2 or 3 arguments; %d argument was specified", cmdname, slink_count(cmdargs));
         *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
         if((*get_ercode() = printFileError(t))) return(*get_ercode());
         return(TOKEN_SUCCESS);
         break;

   }

}


int tokenCmdIf(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-If */

   tokeniff *i = NULL;

   // We are starting a new if block. If we are executable here, then
   // create the block and evaluate the first expressions. Otherwise, we
   // are nested in a failed block so we will create the new if block but
   // will not evaluate the expressions.
   i = (tokeniff *)malloc(sizeof(tokeniff));
   if(ifExec(t)) {
      // New if block is executable
      i->eval = false;
      while(cmdargs && !i->eval) {
         if(evalExpression((char *)cmdargs->data)) i->eval = true;
         cmdargs = cmdargs->next;
      }
      i->exec = i->eval;
   } else {
      // New if block nested in failed block
      i->exec = false;
      i->eval = true;
   }
   slink_prepend_head(&t->iff, i, free);
   return(TOKEN_SUCCESS);

}


int tokenCmdElif(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Else-If */

   tokeniff *i = NULL;

   // We are continuing the current if block. If we are evaluable here, then
   // evaluate the current expressions. Otherwise, we are either nested in a
   // failed block or we have already had a successful block in this if.
   // Be prepared to create a new if block if one does not exist.
   if(!t->iff) {
      // Brand new if block
      i = (tokeniff *)malloc(sizeof(tokeniff));
      i->eval = false;
      while(cmdargs && !i->eval) {
         if(evalExpression((char *)cmdargs->data)) i->eval = true;
         cmdargs = cmdargs->next;
      }
      i->exec = i->eval;
      slink_prepend_head(&t->iff, i, free);
   } else {
      i = (tokeniff *)t->iff->data;
      if(!i->eval) {
         // Continuing a block that has failed up until now.
         while(cmdargs && !i->eval) {
            if(evalExpression((char *)cmdargs->data)) i->eval = true;
            cmdargs = cmdargs->next;
         }
         i->exec = i->eval;
      } else {
         // Have already successfully evaluated. Note that the previous
         // block may have been successful, so we need to turn off the
         // executable flag.
         i->exec = false;
      }
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdElse(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Else */

   tokeniff *i = NULL;

   if(cmdargs && *(char *)cmdargs->data) {
      snprintf(get_erstr(), 32768, "%s does not accept any arguments.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   
   // We are continuing the current if block. If we are evaluable here, then
   // execute this block. Otherwise, we are either nested in a
   // failed block or we have already had a successful block in this if.
   // Be prepared to create a new if block if one does not exist.
   if(!t->iff) {
      // Brand new if block
      i = (tokeniff *)malloc(sizeof(tokeniff));
      i->eval = true;
      i->exec = true;
      slink_prepend_head(&t->iff, i, free);
   } else {
      i = (tokeniff *)t->iff->data;
      if(!i->eval) {
         // Continuing a block that has failed up until now.
         i->eval = true;
         i->exec = true;
      } else {
         // Have already successfully evaluated. Note that the previous
         // block may have been successful, so we need to turn off the
         // executable flag.
         i->exec = false;
      }
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdEndif(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-End-If */

   if(cmdargs && *(char *)cmdargs->data) {
      snprintf(get_erstr(), 32768, "%s does not accept any arguments.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   
   // Close out the current if block.
   if(t->iff) slink_free(&t->iff, true);
   return(TOKEN_SUCCESS);

}


/*** Commands: Special state options ***/


int tokenCmdText(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Text */

   if(!cmdargs) {
      snprintf(get_erstr(), 32768, "%s: No arguments specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_FEW_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
      return(TOKEN_SUCCESS);
   }
   if(cmdargs->next) {
      snprintf(get_erstr(), 32768, "%s: Too many arguments specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   if(!(scmp(cmdargs->data, "on") ^ 0xffffffff)) tokenSetText(t, true);
   else if(!(scmp(cmdargs->data, "off") ^ 0xffffffff)) tokenSetText(t, false);
   else if(!(scmp(cmdargs->data, "reset") ^ 0xffffffff)) tokenResetText(t);
   else {
      snprintf(get_erstr(), 32768, "%s: Invalid option, \"%s\"", cmdname, (char *)cmdargs->data);
      *get_ercode() = TOKEN_ERROR_INVALID_ARG;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   #ifdef LIBTOKEN_DEBUG_TEXT
   printf("%s %d '%s'\n", tokenGetText(t) ? "visible" : "invisible", t->text, (char *)cmdargs->data);
   #endif
   return(TOKEN_SUCCESS);

}


int tokenCmdSquash(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Squash */

   if(!cmdargs) {
      snprintf(get_erstr(), 32768, "%s: No arguments specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_FEW_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
      return(TOKEN_SUCCESS);
   }
   if(cmdargs->next) {
      snprintf(get_erstr(), 32768, "%s: Too many arguments specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   if(!(scmp(cmdargs->data, "on") ^ 0xffffffff)) tokenSetSquash(t, true);
   else if(!(scmp(cmdargs->data, "off") ^ 0xffffffff)) tokenSetSquash(t, false);
   else if(!(scmp(cmdargs->data, "reset") ^ 0xffffffff)) tokenResetSquash(t);
   else {
      snprintf(get_erstr(), 32768, "%s: Invalid option, \"%s\"", cmdname, (char *)cmdargs->data);
      *get_ercode() = TOKEN_ERROR_INVALID_ARG;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   #ifdef LIBTOKEN_DEBUG_TEXT
   printf("%s %d '%s'\n", tokenGetSquash(t) ? "visible" : "invisible", t->squash, (char *)cmdargs->data);
   #endif
   return(TOKEN_SUCCESS);

}


int tokenCmdNewline(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Newline */

   if(!cmdargs) {
      snprintf(get_erstr(), 32768, "%s: No arguments specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_FEW_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
      return(TOKEN_SUCCESS);
   }
   if(cmdargs->next) {
      snprintf(get_erstr(), 32768, "%s: Too many arguments specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   if(!(scmp(cmdargs->data, "on") ^ 0xffffffff)) tokenSetNewline(t, true);
   else if(!(scmp(cmdargs->data, "off") ^ 0xffffffff)) tokenSetNewline(t, false);
   else if(!(scmp(cmdargs->data, "reset") ^ 0xffffffff)) tokenResetNewline(t);
   else {
      snprintf(get_erstr(), 32768, "%s: Invalid option, \"%s\"", cmdname, (char *)cmdargs->data);
      *get_ercode() = TOKEN_ERROR_INVALID_ARG;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   #ifdef LIBTOKEN_DEBUG_TEXT
   printf("%s %d '%s'\n", tokenGetNewline(t) ? "visible" : "invisible", t->newline, (char *)cmdargs->data);
   #endif
   return(TOKEN_SUCCESS);

}


/*** Commands: Custom ($$define;) commands ***/


int tokenCmdCustom(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Custom-Function */

   file *fbufi;
   char  carg[TOKEN_ARG_LENGTH + 1];
   char *cargp = carg;
   const char *p = (const char *)param;
   const char *pl = NULL;
   const slinknode *cur = NULL;
   int n;
   int referenced = 0;
   tokensavedstate ss;

   #ifdef LIBTOKEN_DEBUG_DEFINE
   printf("%s(", cmdname);
   cur = cmdargs;
   while(cur) {
      printf("%s", (char *)cur->data);
      cur = cur->next;
      if(cur) printf(", ");
   }
   printf(")\n");
   #endif

   // Check to see what arguments are not referenced.
   while(*p) {
      if(!(scmp_size(p, "$$.arg", 6) ^ 0xffffffff) && *(p + 6)) {
         // Possibly an argument
         if((*(p + 6) >= '0' && *(p + 6) <= '9') && *(p + 7) == ';') {
            // It's an argument! Yay...
            n = *(p + 6) - '0';
            if(n) referenced = referenced | (1 << (n - 1));
            p += 8;
         } else p++;
      } else if(!(scmp_size(p, "#.arg", 5) ^ 0xffffffff) && *(p + 5)) {
         // Possibly an argument
         if((*(p + 5) >= '0' && *(p + 5) <= '9') && *(p + 6) == '#') {
            // It's an argument! Yay...
            n = *(p + 5) - '0';
            if(n) referenced = referenced | (1 << (n - 1));
            p += 7;
         } else p++;
      } else p++;
   }

   // Argument expansion.
   p = (const char *)param;
   while(*p) {
      if(!(scmp_size(p, "$$.args;", 8) ^ 0xffffffff)) {
         // Print all arguments, separated by spaces
         cur = cmdargs;
         #ifdef LIBTOKEN_DEBUG_DEFINE
         printf("  %d: ", slink_count(cur));
         #endif
         while(cur) {
            pl = (const char *)cur->data;
            while(*pl && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(pl++);
            cur = cur->next;
            if(cur && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = ' ';
         }
         #ifdef LIBTOKEN_DEBUG_DEFINE
         printf("\n");
         #endif
         p += 8;
      } else if(!(scmp_size(p, "$$.argx;", 8) ^ 0xffffffff)) {
         // Print all unreferenced arguments; separate by spaces.
         cur = cmdargs;
         #ifdef LIBTOKEN_DEBUG_DEFINE
         printf("  %d: ", slink_count(cur));
         #endif
         n = 0;
         while(cur) {
            if(n > 9 || !((1 << (n)) & referenced)) {
               // Print this argument.
               pl = (const char *)cur->data;
               while(*pl && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(pl++);
               cur = cur->next;
               if(cur && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = ' ';
            } else cur = cur->next;
            n++;
         }
         #ifdef LIBTOKEN_DEBUG_DEFINE
         printf("\n");
         #endif
         p += 8;
      } else if(!(scmp_size(p, "$$.arg", 6) ^ 0xffffffff) && *(p + 6)) {
         // Possibly an argument
         if((*(p + 6) >= '0' && *(p + 6) <= '9') && *(p + 7) == ';') {
            // It's an argument! Yay...
            n = *(p + 6) - '0';
            if(n) {
               cur = cmdargs;
               while(--n && cur) cur = cur->next;
               if(cur) {
                  // Argument has a value!
                  pl = (const char *)cur->data;
                  while(*pl && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(pl++);
               }
            } else {
               // Insert command name
               pl = cmdname;
               while(*pl && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(pl++);
            }
            p += 8;
         } else {
            // Not an argument.
            if(cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(p++);
            else p++;
         }
      } else if(!(scmp_size(p, "#.args#", 7) ^ 0xffffffff)) {
         // Print all arguments, separated by spaces
         cur = cmdargs;
         #ifdef LIBTOKEN_DEBUG_DEFINE
         printf("  %d: ", slink_count(cur));
         #endif
         while(cur) {
            pl = (const char *)cur->data;
            while(*pl && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(pl++);
            cur = cur->next;
            if(cur && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = ' ';
         }
         #ifdef LIBTOKEN_DEBUG_DEFINE
         printf("\n");
         #endif
         p += 7;
      } else if(!(scmp_size(p, "#.argx#", 7) ^ 0xffffffff)) {
         // Print all unreferenced arguments; separate by spaces.
         cur = cmdargs;
         #ifdef LIBTOKEN_DEBUG_DEFINE
         printf("  %d: ", slink_count(cur));
         #endif
         n = 0;
         while(cur) {
            if(n > 9 || !((1 << (n)) & referenced)) {
               // Print this argument.
               pl = (const char *)cur->data;
               while(*pl && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(pl++);
               cur = cur->next;
               if(cur && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = ' ';
            } else cur = cur->next;
            n++;
         }
         #ifdef LIBTOKEN_DEBUG_DEFINE
         printf("\n");
         #endif
         p += 7;
      } else if(!(scmp_size(p, "#.arg", 5) ^ 0xffffffff) && *(p + 5)) {
         // Possibly an argument
         if((*(p + 5) >= '0' && *(p + 5) <= '9') && *(p + 6) == '#') {
            // It's an argument! Yay...
            n = *(p + 5) - '0';
            if(n) {
               cur = cmdargs;
               while(--n && cur) cur = cur->next;
               if(cur) {
                  // Argument has a value!
                  pl = (const char *)cur->data;
                  while(*pl && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(pl++);
               }
            } else {
               // Insert command name
               pl = cmdname;
               while(*pl && cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(pl++);
            }
            p += 7;
         } else {
            // Not an argument.
            if(cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(p++);
            else p++;
         }
      } else {
         if(cargp - carg < TOKEN_ARG_LENGTH) *(cargp++) = *(p++);
         else p++;
      }
   }
   *cargp = 0;

   #ifdef LIBTOKEN_DEBUG_DEFINE
   printf("  %s\n", carg);
   #endif

   // Execute the command.
   if(*carg) {
      tokenSaveState(&ss, t);
      fbufi = file_create_from_source(cmdname, 'r', carg, sscan_null(carg));
      if((*get_ercode() = parseTokens(t, fbufi, fo))) {
         tokenRestoreState(t, &ss);
         file_destroy(&fbufi);
         return(*get_ercode());
      }
      file_destroy(&fbufi);
      tokenRestoreState(t, &ss);
   }
   
   // Finished.
   return(TOKEN_SUCCESS);

}


int tokenCmdDefine(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Define */

   const slinknode *cur = cmdargs;
   char *s;
   tokencmd *c;
   while(cur) {
      c = cmdExists(t, (char *)cur->data);
      if(!c || cmdIsCustom(c)) {
         if(cur->next) {
            s = (char *)malloc(sscan_null(cur->next->data) + 1);
            scopy(s, cur->next->data);
            slink_prepend_head(&t->customdefs, s, free);
            tokenAddCommand(t, (char *)cur->data, &tokenCmdCustom, s, TOKEN_CMD_IS_CUSTOM);
         } else tokenAddCommand(t, (char *)cur->data, &tokenCmdCustom, "", TOKEN_CMD_IS_CUSTOM);
      } else if(c) {
         // Command is an internal function, cannot redefine
         snprintf(get_erstr(), 32768, "\"%s\": Cannot redefine internal function \"%s\"", cmdname, (char *)cur->data);
         *get_ercode() = TOKEN_ERROR_INTERNAL;
         printFileError(t);
         if(*get_ercode()) return(*get_ercode());
      }
      cur = cur->next;
      if(cur) cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdUndef(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Undefine */

   const slinknode *cur = cmdargs;
   tokencmd *c;
   while(cur) {
      c = cmdExists(t, (char *)cur->data);
      if(cmdIsCustom(c)) tokenAddCommand(t, (char *)cur->data, &tokenCmdCustom, "", TOKEN_CMD_IS_CUSTOM);
      else if(c) {
         // Command is an internal function, cannot undefine
         snprintf(get_erstr(), 32768, "\"%s\": Cannot undefine internal function \"%s\"", cmdname, (char *)cur->data);
         *get_ercode() = TOKEN_ERROR_INTERNAL;
         printFileError(t);
         if(*get_ercode()) return(*get_ercode());
      }      
      cur = cur->next;
   }
   return(TOKEN_SUCCESS);

}


int tokenCmdArg(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Argument */

   if(tokenGetText(t)) {
      file_write_line(fo, "$$");
      file_write_line(fo, cmdname);
      file_write_char(fo, ';');
   }
   return(TOKEN_SUCCESS);

}


/*** Commands: runtime debugging ***/


int tokenCmdDumpCmd(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Dump-Commands */

   slinknode *cur = t->commands;
   tokencmd *c;
   int spaces;

   file_write_line(fo, "Token command data dump\n");
   file_write_line(fo, "------------------------------------------------------------------------\n");
   while(cur) {
      c = (tokencmd *)cur->data;
      file_write_line(fo, c->cmdname);
      spaces = 36 - sscan_null(c->cmdname);
      while(spaces--) file_write_char(fo, ' ');
      file_write_char(fo, c->flags & TOKEN_CMD_IS_CUSTOM ? 'C' : ' ');
      file_write_char(fo, c->flags & TOKEN_CMD_DO_NOT_EXEC_ARGS ? 'X' : ' ');
      file_write_char(fo, c->flags & TOKEN_CMD_IGNORE_IF_BLOCKS ? 'I' : ' ');
      file_write_char(fo, ' ');
      file_write_char(fo, ' ');
      file_write_char(fo, ' ');
      file_write_char(fo, ' ');
      if(c->flags & TOKEN_CMD_IS_CUSTOM) file_write_line(fo, (char *)c->param);
      file_write_char(fo, '\n');
      cur = cur->next;
   }   
   file_write_line(fo, "------------------------------------------------------------------------\n");
   return(TOKEN_SUCCESS);

}


int tokenCmdDumpVar(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Dump-Variables */

   slinknode *cur = t->variables;
   tokenvar *c;
   int spaces;

   file_write_line(fo, "Token variable data dump\n");
   file_write_line(fo, "------------------------------------------------------------------------\n");
   while(cur) {
      c = (tokenvar *)cur->data;
      file_write_line(fo, c->name);
      spaces = 36 - sscan_null(c->name);
      while(spaces--) file_write_char(fo, ' ');
      file_write_line(fo, c->value);
      file_write_char(fo, '\n');
      cur = cur->next;
   }   
   file_write_line(fo, "------------------------------------------------------------------------\n");
   return(TOKEN_SUCCESS);

}


int tokenCmdDumpState(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Dump-State */

   file_write_line(fo, "Token variable state dump\n");
   file_write_line(fo, "------------------------------------------------------------------------\n");
   if(t->text > 0) file_write_line(fo, "Flowing text is not being echoed\n");
   else file_write_line(fo, "Flowing text is being echoed\n");
   if(t->squash > 0) file_write_line(fo, "Whitespace is being squashed in flowing text\n");
   else file_write_line(fo, "Whitespace is not being squashed in flowing text\n");
   if(t->newline > 0) file_write_line(fo, "Newlines in flowing text are not being echoed\n");
   else file_write_line(fo, "Newlines in flowing text are being echoed\n");
   file_write_line(fo, "------------------------------------------------------------------------\n");
   return(TOKEN_SUCCESS);

}


int tokenCmdDumpVersions(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Dump-Versions */

   long *vp = get_token_versions();
   char buf[512];
   file_write_line(fo, "Token version info dump\n");
   file_write_line(fo, "------------------------------------------------------------------------\n");
   while(*vp) {
      file_write_line(fo, tokenStrFromVersion(buf, *vp));
      file_write_line(fo, "\n");
      vp++;
   }
   file_write_line(fo, "------------------------------------------------------------------------\n");
   return(TOKEN_SUCCESS);

}


/*** Commands: Math Functions ***/


#ifndef NO_MATH_LIB
int tokenCmdAdd(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Add */

   double result = 0;
   const slinknode *cur = cmdargs;
   while(cur) {
      result += (atof((char *)cur->data));
      cur = cur->next;
   }
   mathWriteNumber(fo, result);
   return(TOKEN_SUCCESS);

}


int tokenCmdSubtract(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Subtract */

   double result = 0;
   const slinknode *cur = cmdargs;
   while(cur) {
      if(cur == cmdargs) result = atof((char *)cur->data);
      else result += (atof((char *)cur->data));
      cur = cur->next;
   }
   mathWriteNumber(fo, result);
   return(TOKEN_SUCCESS);

}


int tokenCmdMultiply(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Multiply */

   double result = 1;
   const slinknode *cur = cmdargs;
   while(cur) {
      result *= (atof((char *)cur->data));
      cur = cur->next;
   }
   mathWriteNumber(fo, result);
   return(TOKEN_SUCCESS);

}


int tokenCmdDivide(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Divide */

   double result = 1;
   const slinknode *cur = cmdargs;
   while(cur) {
      if(cur == cmdargs) result = atof((char *)cur->data);
      else result /= (atof((char *)cur->data));
      cur = cur->next;
   }
   mathWriteNumber(fo, result);
   return(TOKEN_SUCCESS);

}


int tokenCmdNegate(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Negate */

   if(cmdargs) mathWriteNumber(fo, -atof((char *)cmdargs->data));
   else mathWriteNumber(fo, 0);
   return(TOKEN_SUCCESS);

}


int tokenCmdExp(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Exponential (base e) */

   if(cmdargs) mathWriteNumber(fo, exp(atof((char *)cmdargs->data)));
   else mathWriteNumber(fo, 1);
   return(TOKEN_SUCCESS);

}


int tokenCmdLn(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Logarithm (base e) */

   if(cmdargs) mathWriteNumber(fo, log(atof((char *)cmdargs->data)));
   else mathWriteNumber(fo, 0);
   return(TOKEN_SUCCESS);

}
#endif // math lib support


/*** Version Commands ***/


int tokenCmdGetVersion(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Get-Version */

   if(cmdargs) {
      snprintf(get_erstr(), 32768, "%s: This command does not take any arguments", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   mathWriteNumber(fo, t->version);
   return(TOKEN_SUCCESS);
   
}



int tokenCmdSetVersion(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Set-Version */

   if(!cmdargs) {
      snprintf(get_erstr(), 32768, "%s: This command requires 1 argument; none were specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_FEW_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
      return(TOKEN_SUCCESS);
   }
   if(cmdargs->next) {
      snprintf(get_erstr(), 32768, "%s: Too many arguments specified; this command accepts 1, but %d were specified.", cmdname, slink_count(cmdargs));
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   if(!tokenSetVersion(t, atoi((char *)cmdargs->data))) {
      snprintf(get_erstr(), 32768, "%s: Version failure attempting to set version to %d", cmdname, atoi((char *)cmdargs->data));
      *get_ercode() = TOKEN_ERROR_VERSION;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   return(TOKEN_SUCCESS);
   
}


int tokenCmdSupportedVersion(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Supported-Version */

   if(!cmdargs) {
      snprintf(get_erstr(), 32768, "%s: This command requires 1 argument; none were specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_FEW_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
      return(TOKEN_SUCCESS);
   }
   if(cmdargs->next) {
      snprintf(get_erstr(), 32768, "%s: Too many arguments specified; this command accepts 1, but %d were specified.", cmdname, slink_count(cmdargs));
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   mathWriteNumber(fo, tokenVersionSupported(atoi((char *)cmdargs->data)));
   return(TOKEN_SUCCESS);
   
}


int tokenCmdGetRealVersion(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Get-Real-Version */

   if(cmdargs) {
      snprintf(get_erstr(), 32768, "%s: This command does not take any arguments", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   mathWriteNumber(fo, *get_token_versions());
   return(TOKEN_SUCCESS);
   
}


int tokenCmdVersionID(tokeniser *t, file *fi, file *fo, const char *cmdname, const slinknode *cmdargs, void *param) {
/* Token-Command-Version-ID */

   if(!cmdargs) {
      snprintf(get_erstr(), 32768, "%s: This command requires 1 argument; none were specified.", cmdname);
      *get_ercode() = TOKEN_ERROR_TOO_FEW_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
      return(TOKEN_SUCCESS);
   }
   if(cmdargs->next) {
      snprintf(get_erstr(), 32768, "%s: Too many arguments specified; this command accepts 1, but %d were specified.", cmdname, slink_count(cmdargs));
      *get_ercode() = TOKEN_ERROR_TOO_MANY_ARGS;
      if((*get_ercode() = printFileError(t))) return(*get_ercode());
   }
   mathWriteNumber(fo, tokenVersionFromStr((char *)cmdargs->data));
   return(TOKEN_SUCCESS);
   
}


