/* LIBTOKEN                                                 Version 0.1.2
   Copyright (C) 1998   Justin Smith             justins@ugcs.caltech.edu
   ----------------------------------------------------------------------

   This  program  is free software; you can redistribute it and/or modify
   it  under  the terms of the GNU General Public License as published by
   the  Free  Software  Foundation;  either  version 2 of the License, or   
   (at your option) any later version.

   This  program  is  distributed  in  the  hope  that it will be useful,
   but  WITHOUT  ANY  WARRANTY;  without  even  the  implied  warranty of 
   MERCHANTABILITY   or  FITNESS  FOR  A  PARTICULAR  PURPOSE.   See  the  
   GNU General Public License for more details.

   You  should  have  received  a  copy of the GNU General Public License
   along  with  this  program,  in the file COPYING. If not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA

   ----------------------------------------------------------------------
*/


#define __LIBTOKEN_SOURCE
#include <libjtoken.h>
#include <libtoken_int.h>
#include <libtoken_supp.h>
#include <libtoken_parser.h>
#include <libjlink.h>
#include <libjmem.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef NO_MATH_LIB
#include <math.h>
#endif  // math lib?


/*** Command execution ***/


int parseCommand(tokeniser *t, file *fi, file *fo, const char *cmdname, slinknode *cmdargs) {
/* Parse-Command
   This function executes a command, given the name of the command and its
   arguments. This function is normally only called by parseTokens(). This
   function will execute command arguments if needed, and will decide whether
   or not the command should actually be executed. */
   
   slinknode *newargs = NULL;
   slinknode *cur;
   file *fbufi;
   file *fbufo;
   char *s;
   tokencmd *c;
   tokensavedstate ss;

   // Get the command in question.
   if(!(c = cmdExists(t, cmdname))) {
      if(ifExec(t)) {
         snprintf(get_erstr(), 32768, "Command not found - \"%s\"", cmdname);
         *get_ercode() = TOKEN_ERROR_COMMAND_NOT_FOUND;
         printFileError(t);
         if(*get_ercode()) return(*get_ercode());
      }
      return(TOKEN_SUCCESS);
   }

   // Determine whether to process this command, based on if blocks
   if(!ifCmdExec(t, c)) return(TOKEN_SUCCESS);

   // Debug output
   #ifdef LIBTOKEN_DEBUG_TRACE
   printf("%s(%d): %s(", fName(fi), fLine(fi), cmdname);
   cur = cmdargs;
   while(cur) {
      printf("%s", (char *)cur->data);
      cur = cur->next;
      if(cur) printf(", ");
   }
   printf(")\n");
   #endif // trace

   // First action: expand arguments by running the command substitutions. 
   if(ifExec(t) && !cmdDoNotExecArgs(c)) {
      cur = cmdargs;
      while(cur) {
         s = (char *)malloc(TOKEN_ARG_LENGTH + 1);
         *s = 0;
         if(*(char *)cur->data) {
            fbufi = file_create_from_source(cmdname, 'r', (char *)cur->data, sscan_null(cur->data));
            fbufo = file_create_from_source(cmdname, 'w', s, TOKEN_ARG_LENGTH);
            tokenSaveState(&ss, t);
            if((*get_ercode() = parseTokens(t, fbufi, fbufo))) {
               file_destroy(&fbufi);
               file_destroy(&fbufo);
               slink_release(&newargs);
               tokenRestoreState(t, &ss);
               free(s);
               return(*get_ercode());
            }
            tokenRestoreState(t, &ss);
            file_destroy(&fbufi);
            file_destroy(&fbufo);
         } else scopy(s, cur->data);
         slink_append_head(&newargs, s, free);
         cur = cur->next;
      }
      cmdargs = newargs;
   }

   // Run the command using the supplied function
   if(c->exfunc) {
      // Execute some user code for this one
      if((*get_ercode() = c->exfunc(t, fi, fo, cmdname, cmdargs, c->param))) {
         slink_release(&newargs);
         return(*get_ercode());
      }
   } else file_write_line(fo, cmdname);
   
   // Finish up here; we're done.
   slink_release(&newargs);
   return(TOKEN_SUCCESS);
   
}
