/* LIBTOKEN                                                 Version 0.1.2
   Copyright (C) 1998   Justin Smith             justins@ugcs.caltech.edu
   ----------------------------------------------------------------------

   This  program  is free software; you can redistribute it and/or modify
   it  under  the terms of the GNU General Public License as published by
   the  Free  Software  Foundation;  either  version 2 of the License, or   
   (at your option) any later version.

   This  program  is  distributed  in  the  hope  that it will be useful,
   but  WITHOUT  ANY  WARRANTY;  without  even  the  implied  warranty of 
   MERCHANTABILITY   or  FITNESS  FOR  A  PARTICULAR  PURPOSE.   See  the  
   GNU General Public License for more details.

   You  should  have  received  a  copy of the GNU General Public License
   along  with  this  program,  in the file COPYING. If not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA

   ----------------------------------------------------------------------
*/


#define __LIBTOKEN_SOURCE
#include <libjtoken.h>
#include <libjlink.h>
#include <libjmem.h>
#include <stdlib.h>
#ifndef NO_MATH_LIB
#include <math.h>
#endif  // math lib?


/*** Expression Evaluation ***/


bool evalExpression(const char *expr) {
/* Evaluate-Expression 
   This function will evaluate the character expression as indicated.
   Valid expression forms are listed in the documentation. All command
   execution and variable substitution must have already taken place
   when the expression is passed to this function. This function will
   not evaluate individual components of the expression, and it will
   not unescape characters. Returns true if the expression evaluates
   to true, and false otherwise. */

   char  str1[TOKEN_ARG_LENGTH + 1] = "";
   char  str2[TOKEN_ARG_LENGTH + 1] = "";
   char *strp = str1;
   const char *ep = expr;
   int state = EXPR_STATE_INIT;
   int action = EXPR_STRING;
   bool r = false;

   while(*ep) {
      switch(state) {
         
         // EXPR_STATE_INIT
         // Searching for the first non-whitespace character.
         //
         // -> EXPR_STATE_STRING    First non-whitespace character found
         // -> EXPR_STATE_STRING_Q  Quoted string found
         // -> EXPR_STATE_INIT      Whitespace character found.
         case EXPR_STATE_INIT:
            switch(*ep) {
               case ' ':
               case '\t':
               case '\n':
                  break;
                  
               case '\"':
                  state = EXPR_STATE_STRING_Q;
                  break;
                  
               default:
                  *(strp++) = *ep;
                  state = EXPR_STATE_STRING;
                  break;
            }
            break;
            
         // EXPR_STATE_STRING
         // Reading a string in.
         // -> EXPR_STATE_STRING    Normal character
         // -> EXPR_STATE_STRING    Symbol processed, next string
         case EXPR_STATE_STRING:
            if(action == EXPR_STRING) {
               switch(*ep) {
                  case '=':
                     if(*(ep + 1) == '=') {
                        // == operator
                        ep++;
                        action = EXPR_EQUATE;
                        strp--;
                        while(strp >= str1 && (*strp == ' ' || *strp == '\t' || *strp == '\n')) strp--; 
                        *(++strp) = 0;
                        strp = str2;
                        state = EXPR_STATE_INIT;
                     } else *(strp++) = *ep;
                     break;
                     
                  case '!':
                     if(*(ep + 1) == '=') {
                        // != operator
                        ep++;
                        action = EXPR_NOT_EQUATE;
                        strp--;
                        while(strp >= str1 && (*strp == ' ' || *strp == '\t' || *strp == '\n')) strp--; 
                        *(++strp) = 0;
                        strp = str2;
                        state = EXPR_STATE_INIT;
                     } else *(strp++) = *ep;
                     break;
                  
                  default:
                     *(strp++) = *ep;
                     break;
                     
               }
            } else *(strp++) = *ep;
            break;
            
         // EXPR_STATE_STRING_Q
         // Quoted string.
         // -> EXPR_STATE_SYMBOL       End of quotes, look for a symbol.
         // -> EXPR_STATE_STRING_Q     Normal character
         case EXPR_STATE_STRING_Q:
            switch(*ep) {
               case '\"':
                  *strp = 0;
                  state = EXPR_STATE_SYMBOL;
                  break;

               default:
                  *(strp++) = *ep;
                  break;
            }
            break;
            
         // EXPR_STATE_SYMBOL
         // Looking for a symbol.
         // -> EXPR_STATE_SYMBOL       Symbol not found.
         // -> EXPR_STATE_STRING       Symbol found.
         case EXPR_STATE_SYMBOL:
            if(action == EXPR_STRING) {
               switch(*ep) {
                  case '=':
                     if(*(ep + 1) == '=') {
                        // == operator
                        ep++;
                        action = EXPR_EQUATE;
                        strp = str2;
                        state = EXPR_STATE_INIT;
                     }
                     break;
                     
                  case '!':
                     if(*(ep + 1) == '=') {
                        // != operator
                        ep++;
                        action = EXPR_NOT_EQUATE;
                        strp = str2;
                        state = EXPR_STATE_INIT;
                     }
                     break;
                  
               }
            }
            break;
                 
      }
      ep++;
   }   
   *strp = 0;

   // trim first string
   if(state == EXPR_STATE_STRING && action == EXPR_STRING) {
      strp--;
      while(strp >= str1 && (*strp == ' ' || *strp == '\t' || *strp == '\n')) strp--; 
      *(++strp) = 0;
   }
   
   // trim second string
   if(state == EXPR_STATE_STRING && action != EXPR_STRING) {
      strp--;
      while(strp >= str2 && (*strp == ' ' || *strp == '\t' || *strp == '\n')) strp--; 
      *(++strp) = 0;
   }
   
   switch(action) {
      case EXPR_STRING:
         r = *str1;
         #ifdef LIBTOKEN_DEBUG_EVAL
         printf("eval: \"%s\" not null? %c\n", str1, r ? 'T' : 'F');
         #endif
         break;
         
      case EXPR_EQUATE:
         r = !(scmp(str1, str2) ^ 0xffffffff);
         #ifdef LIBTOKEN_DEBUG_EVAL
         printf("eval: \"%s\" == \"%s\" ? %c\n", str1, str2, r ? 'T' : 'F');
         #endif
         break;

      case EXPR_NOT_EQUATE:
         r = scmp(str1, str2) ^ 0xffffffff;
         #ifdef LIBTOKEN_DEBUG_EVAL
         printf("eval: \"%s\" != \"%s\" ? %c\n", str1, str2, r ? 'T' : 'F');
         #endif
         break;
   }
   
   return(r);     
      
}


