/* LIBTOKEN                                                 Version 0.1.2
   Copyright (C) 1998   Justin Smith             justins@ugcs.caltech.edu
   ----------------------------------------------------------------------

   This  program  is free software; you can redistribute it and/or modify
   it  under  the terms of the GNU General Public License as published by
   the  Free  Software  Foundation;  either  version 2 of the License, or   
   (at your option) any later version.

   This  program  is  distributed  in  the  hope  that it will be useful,
   but  WITHOUT  ANY  WARRANTY;  without  even  the  implied  warranty of 
   MERCHANTABILITY   or  FITNESS  FOR  A  PARTICULAR  PURPOSE.   See  the  
   GNU General Public License for more details.

   You  should  have  received  a  copy of the GNU General Public License
   along  with  this  program,  in the file COPYING. If not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA

   ----------------------------------------------------------------------
*/


#define __LIBTOKEN_SOURCE
#include <libjtoken.h>
#include <libtoken_int.h>
#include <libtoken_supp.h>
#include <libtoken_command.h>
#include <libjlink.h>
#include <libjmem.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef NO_MATH_LIB
#include <math.h>
#endif  // math lib?


/*** Token Parsing ***/


inline void initParsedCommand(parsedcommand *c, tokeniser *t, file *fi, file *fo) {
/* Initialise-Parsed-Command */
   
   c->t = t;
   c->fi = fi;
   c->fo = fo;
   c->cmdname = (char *)malloc(TOKEN_CMD_LENGTH + 1);
   c->cmdnamep = c->cmdname;
   c->cmdarg = (char *)malloc(TOKEN_ARG_LENGTH + 1);
   c->cmdargp = c->cmdarg;
   c->cmdvar = (char *)malloc(TOKEN_VAR_LENGTH + 1);
   c->cmdvarp = c->cmdvar;
   *c->cmdname = 0;
   *c->cmdarg = 0;
   *c->cmdvar = 0;
   c->cmdargs = NULL;
   c->states = NULL;
   c->nestlevel = 0;

}


inline void newParsedCommand(parsedcommand *c) {
/* New-Parsed-Command */

   c->cmdnamep = c->cmdname;
   c->cmdargp = c->cmdarg;
   c->cmdvarp = c->cmdvar;
   *c->cmdname = 0;
   *c->cmdarg = 0;
   *c->cmdvar = 0;
   slink_release(&c->cmdargs);
   slink_release(&c->states);
   c->nestlevel = 0;

}


inline void freeParsedCommand(parsedcommand *c) {
/* Free-Parsed-Command */

   slink_release(&c->cmdargs);
   slink_release(&c->states);
   free(c->cmdname);
   free(c->cmdarg);
   free(c->cmdvar);

}


inline void parsedCommandAppendCmd(parsedcommand *c, char ch) {
/* Parsed-Command-Append-Command-Character */

   if((c->cmdnamep - c->cmdname) < TOKEN_CMD_LENGTH) {
      *c->cmdnamep = ch;
      *(++(c->cmdnamep)) = 0;
   }

}


inline void parsedCommandAppendArg(parsedcommand *c, char ch) {
/* Parsed-Command-Append-Argument-Character */

   if((c->cmdargp - c->cmdarg) < TOKEN_ARG_LENGTH) {
      *c->cmdargp = ch;
      *(++(c->cmdargp)) = 0;
   }

}


inline void parsedCommandAppendVar(parsedcommand *c, char ch) {
/* Parsed-Command-Append-Variable-Character */

   if((c->cmdvarp - c->cmdvar) < TOKEN_ARG_LENGTH) {
      *c->cmdvarp = ch;
      *(++(c->cmdvarp)) = 0;
   }

}


inline void parsedCommandAddArgument(parsedcommand *c) {
/* Parsed-Command-Add-Argument */

   if(c->cmdargp != c->cmdarg) {
      c->cmdargp--;
      while(c->cmdargp>= c->cmdarg && (*c->cmdargp == ' ' || *c->cmdargp == '\t' || *c->cmdargp == '\n')) c->cmdargp--;
      *(++(c->cmdargp)) = 0;
   }
   c->cmdargp = (char *)malloc((c->cmdargp - c->cmdarg) + 1);
   scopy(c->cmdargp, c->cmdarg);
   slink_append_head(&c->cmdargs, c->cmdargp, free);
   c->cmdargp = c->cmdarg;
   *c->cmdarg = 0;
   
}


inline void parsedCommandAddArgumentNoTrim(parsedcommand *c) {
/* Parsed-Command-Add-Argument-No-Trim */

   c->cmdargp = (char *)malloc((c->cmdargp - c->cmdarg) + 1);
   scopy(c->cmdargp, c->cmdarg);
   slink_append_head(&c->cmdargs, c->cmdargp, free);
   c->cmdargp = c->cmdarg;
   *c->cmdarg = 0;
   
}


inline void parsedCommandExpandVar(parsedcommand *c) {
/* Parsed-Command-Expand-Variable-Name */

   c->cmdvarp = (char *)tokenGetVariable(c->t, c->cmdvar);
   while(*c->cmdvarp) parsedCommandAppendArg(c, *(c->cmdvarp++));
   c->cmdvarp = c->cmdvar;
   *c->cmdvar = 0;

}


inline void parsedCommandANSIEscape(parsedcommand *c) {
/* Parsed-Command-ANSI-Escape-Sequence
   Process an ANSI escape sequence at the current file in pointer, and
   write to the current argument. */

   int num = 0;
   char ch;
   
   if(!file_read_char(c->fi, &ch)) return;
   switch(ch) {
      case '\0':
         parsedCommandAppendArg(c, 0);
         break;

      case '\n':
         // Do NOT emit a newline.. escape sequence.
         break;
         
      // Numbers 
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
         num = ch - '0';
         if(!file_peek_char(c->fi, &ch)) break;
         if(ch >= '0' && ch <= '7') {
            file_read_char(c->fi, &ch);
            num = (num << 3) | (ch - '0');
            if(!file_peek_char(c->fi, &ch)) break;
            if(ch >= '0' && ch <= '3') {
               file_read_char(c->fi, &ch);
               num = (num << 3) | (ch - '0');
            }
         }
         parsedCommandAppendArg(c, (unsigned char)num);
         break;
         
      // General escapes
      case 'a':
         parsedCommandAppendArg(c, '\a');
         break;
      case 'b':
         parsedCommandAppendArg(c, '\b');
         break;
      case 'f':
         parsedCommandAppendArg(c, '\f');
         break;
      case 'n':
         parsedCommandAppendArg(c, '\n');
         break;
      case 'r':
         parsedCommandAppendArg(c, '\r');
         break;
      case 't':
         parsedCommandAppendArg(c, '\t');
         break;
      case 'v':
         parsedCommandAppendArg(c, '\v');
         break;
         
      // Hexadecimal numbers case
      case 'x':
         if(!file_peek_char(c->fi, &ch)) break;
         if(ch >= '0' && ch <= '9') {
            file_read_char(c->fi, &ch);
            num = ch - '0';
            if(!file_peek_char(c->fi, &ch)) break;
            if(ch >= '0' && ch <= '9') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - '0');
            } 
            if(ch >= 'a' && ch <= 'f') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - 'a' + 10);
            }
            if(ch >= 'A' && ch <= 'F') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - 'A' + 10);
            }
            parsedCommandAppendArg(c, (unsigned char)num);
         } else if(ch >= 'a' && ch <= 'f') {
            file_read_char(c->fi, &ch);
            num = ch - 'a' + 10;
            if(!file_peek_char(c->fi, &ch)) break;
            if(ch >= '0' && ch <= '9') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - '0');
            } 
            if(ch >= 'a' && ch <= 'f') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - 'a' + 10);
            }
            if(ch >= 'A' && ch <= 'F') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - 'A' + 10);
            }
            parsedCommandAppendArg(c, (unsigned char)num);
         } else if(ch >= 'A' && ch <= 'F') {
            file_read_char(c->fi, &ch);
            num = ch - 'A' + 10;
            if(!file_peek_char(c->fi, &ch)) break;
            if(ch >= '0' && ch <= '9') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - '0');
            } 
            if(ch >= 'a' && ch <= 'f') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - 'a' + 10);
            }
            if(ch >= 'A' && ch <= 'F') {
               file_read_char(c->fi, &ch);
               num = (num << 4) | (ch - 'A' + 10);
            }
            parsedCommandAppendArg(c, (unsigned char)num);
         }
         break;
      
      // This will catch everything else
      default:
         parsedCommandAppendArg(c, ch);
         break;
         
   }

}


inline void parsedCommandPushState(parsedcommand *c, int state) {
/* Parsed-Command-Push-State */

   int *i = (int *)malloc(sizeof(int));
   *i = state;
   slink_prepend_head(&c->states, i, free);
   c->nestlevel++;

}


inline int parsedCommandPopState(parsedcommand *c) {
/* Parsed-Command-Pop-State */

   int i = (*(int *)c->states->data);
   slink_free(&c->states, true);
   c->nestlevel--;
   return(i);

}


int parseTokens(tokeniser *t, file *fi, file *fo) {
/* Parse-Tokens
   This function parses tokens from fi and writes them to fo. This function
   is the function used to "execute" files and also individual memory buffers
   (ie, arguments and whatnot) */
   
   char  ch;
   int state = TOKEN_STATE_TEXT;
   int commlevel = 0;
   parsedcommand c;
   
   // Debug code
   #ifdef LIBTOKEN_DEBUG_INCLUDE
   printf("Writing %s -> %s ...\n", fName(fi), fName(fo));
   #endif // debug
   
   // Add a note that we are parsing <fi> (used in error logging).
   slink_prepend_head(&t->nestedfiles, fi, NULL);
   
   // Initialise the command structure.
   initParsedCommand(&c, t, fi, fo);

   // Start reading characters...
   while(file_read_char(fi, &ch)) {
      #ifdef LIBTOKEN_DEBUG_SUICIDAL
      printf("%c: state = %2d, nestlevel = %2d, commlevel = %2d\n", ch, state, c.nestlevel, commlevel);
      #endif // debug
      switch(state) {
      
         // TOKEN_STATE_TEXT
         // Normal, flowing text which will be echoed verbatim to the output
         // file. Will search for the first character of the "$$" escape 
         // sequence which initiates a command.
         //
         // -> TOKEN_STATE_COMMAND_NAME   Expect a command name.
         // -> TOKEN_STATE_COMMENT        Beginning of a comment block sequence
         case TOKEN_STATE_TEXT:
            // This state only exists at the base level.
         
            if(ch == '$') {
               // Possible command escape sequence?
               if(file_peek_char(fi, &ch) && ch == '$') {
                  // Escape sequence! Yay! Advance the file pointer to the
                  // first character of the command name and setup to process
                  // a new command.
                  file_read_char(fi, &ch);
                  // Check to see if this is a comment escape sequence
                  if(file_peek_char(fi, &ch) && ch == '<') {
                     // Beginning a comment block. Crap.
                     file_read_char(fi, &ch);
                     commlevel = 1;
                     state = TOKEN_STATE_COMMENT;
                  } else if(file_peek_char(fi, &ch) && ch == '>') {
                     // Closing a comment block. What comment block? :)
                     file_read_char(fi, &ch);
                  } else {
                     // Beginning a new command sequence
                     newParsedCommand(&c);
                     state = TOKEN_STATE_CMD_NAME;
                  }
               } else {
                  // Just a lonely dollar sign. Emit the symbol and continue.
                  if(t->text <= 0 && ifExec(t)) file_write_char(fo, '$');
               }
            } else {
               // Just another ASCII character to emit.
               if(t->text <= 0 && ifExec(t)) {
                  if(ch != '\n' || t->newline <= 0) {
                     // Newlines allowed or this isn't a newline
                     if(t->squash <= 0) file_write_char(fo, ch);
                     else {
                        if(ch == ' ' || ch == '\t') {
                           // Space or tab
                           if(!t->sawspace) {
                              t->sawspace = 1;
                              file_write_char(fo, ' ');
                           }
                        } else {
                           // Nonspace and nontab
                           t->sawspace = 0;
                           file_write_char(fo, ch);
                        }
                     }
                  }
               }
            }
            break;
            
         // TOKEN_STATE_CMD_NAME
         // We're getting the name of a command now. This sequence can
         // be terminated by a '(' symbol or a ';' symbol. This sequence
         // can generate an error on a ')' symbol, a ',' symbol, a '\"' 
         // symbol, a '\'' symbol, or a '$' symbol.
         //
         // -> TOKEN_STATE_ARGUMENT    On a '(' symbol, expect an argument
         // -> TOKEN_STATE_TEXT        On termination of command, ';' symbol
         // -> TOKEN_STATE_TEXT        On an error in processing.
         case TOKEN_STATE_CMD_NAME:
            if(!c.nestlevel) switch(ch) {
               // This is a base-level command
               
               case ')':
               case ',':
               case '$':
               case '\"':
               case '\'':
                  // Erroneous symbols. Return to text processing after
                  // indicating an error has occurred.
                  snprintf(get_erstr(), 32768, "Invalid symbol '%c' in partial command name \"%s\"", ch, c.cmdname);
                  *get_ercode() = TOKEN_ERROR_INVALID_SYMBOL;
                  printFileError(t);
                  if(*get_ercode()) {
                     freeParsedCommand(&c);
                     return(*get_ercode());
                  }
                  state = TOKEN_STATE_TEXT;
                  break;
                  
               case '(':
                  // Beginning of arguments
                  state = TOKEN_STATE_ARGUMENT;
                  break;
                  
               case ';':
                  // End of command
                  if((*get_ercode() = parseCommand(t, fi, fo, c.cmdname, c.cmdargs))) {
                     freeParsedCommand(&c); 
                     return(*get_ercode());
                  }
                  state = TOKEN_STATE_TEXT;
                  break;
                  
               case ' ':
               case '\t':
               case '\n':
                  // Whitespace. Ignore.
                  break;
                  
               default:
                  // Normal command character
                  parsedCommandAppendCmd(&c, ch);
                  break;
                  
            } else switch(ch) {
               // This is a nested command
               
               case '(':
                  // Beginning of arguments
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_ARGUMENT;
                  break;
                  
               case ';':
                  // We are a nested command. Just decrement the nesting 
                  // level and hopefully we'll eventually return to the
                  // original command.
                  parsedCommandAppendArg(&c, ch);
                  state = parsedCommandPopState(&c);
                  break;
                  
               default:
                  // Normal command character
                  parsedCommandAppendArg(&c, ch);
                  break;
                  
            }
            break;
            
         // TOKEN_STATE_ARGUMENT
         // This state is set when we are at the beginning of a new
         // argument. In this state, we will ignore spaces, tabs and
         // newlines (consider them as whitespace), consider ',' to
         // indicate that this argument is empty, consider ')' to
         // terminate the argument list, and any other character to
         // begin the argument. Note that quotes will be special-cased
         // here; if an argument is in quotes then the ENTIRE argument
         // must be in quotes. Note that if the argument contains any
         // of the following characters then it MUST be quoted:
         // ';' '(' ')' ',' '$' '\"' '\'' leading or trailing whitespace 
         // or '#' (that is not part of a variable expansion)
         //
         // -> TOKEN_STATE_IN_ARG      Normal character to begin an argument
         // -> TOKEN_STATE_IN_ARGV     Beginning an argument with a variable.
         // -> TOKEN_STATE_IN_ARG_Q    Quoted argument
         // -> TOKEN_STATE_ARGUMENT    ',' encountered, null (empty) argument
         // -> TOKEN_STATE_END_ARGS    ')' encountered, null (empty) final arg
         // -> TOKEN_STATE_TEXT        Invalid character encountered
         case TOKEN_STATE_ARGUMENT:
            if(!c.nestlevel) switch(ch) {
               // This is a top-level command
               
               case '$':
                  // We might be looking at another escape sequence
                  state = TOKEN_STATE_IN_ARG;
                  parsedCommandAppendArg(&c, ch);
                  if(file_peek_char(fi, &ch) && ch == '$') {
                     // An escape sequence! Yay!
                     parsedCommandAppendArg(&c, ch);
                     file_read_char(fi, &ch);
                     if(file_peek_char(fi, &ch) && (ch == '<' || ch == '>')) {
                        // Comment sequence only.
                        parsedCommandAppendArg(&c, ch);
                        file_read_char(fi, &ch);
                     } else {
                        // Must be a command sequence. Begin the nesting 
                        // levels!
                        parsedCommandPushState(&c, state);
                        state = TOKEN_STATE_CMD_NAME;
                     }
                  }
                  break;

               case '\"':
                  // Begin a quoted argument.
                  state = TOKEN_STATE_IN_ARG_Q;
                  break;
                  
               case ',':
                  // Null (empty) argument.
                  parsedCommandAddArgument(&c);
                  break;
                  
               case ')':
                  // Null (empty) argument, end of arguments.
                  parsedCommandAddArgument(&c);
                  state = TOKEN_STATE_END_ARGS;
                  break;
                  
               case ' ':
               case '\t':
               case '\n':
                  // Leading whitespace; ignore.
                  break;
                  
               case '#':
                  // Beginning of a variable name.
                  state = TOKEN_STATE_IN_ARGV;
                  break;
                  
               default:
                  // Normal character argument. Process as usual.
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_IN_ARG;
                  break;
                  
            } else switch(ch) {
               // This is a nested command
               
               case '$':
                  // We might be looking at another escape sequence
                  state = TOKEN_STATE_IN_ARG;
                  parsedCommandAppendArg(&c, ch);
                  if(file_peek_char(fi, &ch) && ch == '$') {
                     // An escape sequence! Yay!
                     parsedCommandAppendArg(&c, ch);
                     file_read_char(fi, &ch);
                     if(file_peek_char(fi, &ch) && (ch == '<' || ch == '>')) {
                        // Comment sequence only.
                        parsedCommandAppendArg(&c, ch);
                        file_read_char(fi, &ch);
                     } else {
                        // Must be a command sequence. Begin the nesting 
                        // levels!
                        parsedCommandPushState(&c, state);
                        state = TOKEN_STATE_CMD_NAME;
                     }
                  }
                  break;

               case '\"':
                  // Begin a quoted argument.
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_IN_ARG_Q;
                  break;
                  
               case ')':
                  // Null (empty) argument, end of arguments.
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_END_ARGS;
                  break;
                  
               default:
                  // Normal character argument. Process as usual.
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_IN_ARG;
                  break;
                  
            }
            break;
               
         // TOKEN_STATE_IN_ARG
         // We're processing an individual argument right now. Remember,
         // the characters disallowed in an argument are '(', ')', ',',
         // '$', '\"', '\'', ';'. ',' or ')' will signal the end of the
         // argument. Whitespace is allowed inside the argument but leading
         // and trailing whitespace is truncated. Be prepared to intercept
         // '#' as the beginning of a variable name.
         //
         // -> TOKEN_STATE_IN_ARG     Normal character
         // -> TOKEN_STATE_IN_ARGV    A variable is actually beginning now.
         // -> TOKEN_STATE_ARGUMENT   ',' encountered
         // -> TOKEN_STATE_END_ARGS   ')' encountered
         // -> TOKEN_STATE_TEXT       Invalid character encountered
         case TOKEN_STATE_IN_ARG:
            if(!c.nestlevel) switch(ch) {
               // Top level command
                  
               case '$':
                  // We might be looking at another escape sequence
                  parsedCommandAppendArg(&c, ch);
                  if(file_peek_char(fi, &ch) && ch == '$') {
                     // An escape sequence! Yay!
                     parsedCommandAppendArg(&c, ch);
                     file_read_char(fi, &ch);
                     if(file_peek_char(fi, &ch) && (ch == '<' || ch == '>')) {
                        // Comment sequence only.
                        parsedCommandAppendArg(&c, ch);
                        file_read_char(fi, &ch);
                     } else {
                        // Must be a command sequence. Begin the nesting 
                        // levels!
                        parsedCommandPushState(&c, state);
                        state = TOKEN_STATE_CMD_NAME;
                     }
                  }
                  break;

               case ',':
                  // End of this argument, begin the next argument.
                  parsedCommandAddArgument(&c);
                  state = TOKEN_STATE_ARGUMENT;
                  break;
                  
               case ')':
                  // End of argument list.
                  parsedCommandAddArgument(&c);
                  state = TOKEN_STATE_END_ARGS;
                  break;
                  
               case '#':
                  // Beginning of a variable name.
                  state = TOKEN_STATE_IN_ARGV;
                  break;
                  
               default:
                  // Normal character
                  parsedCommandAppendArg(&c, ch);
                  break;
                  
            } else switch(ch) {
               // Nested level command
                  
               case '$':
                  // We might be looking at another escape sequence
                  parsedCommandAppendArg(&c, ch);
                  if(file_peek_char(fi, &ch) && ch == '$') {
                     // An escape sequence! Yay!
                     parsedCommandAppendArg(&c, ch);
                     file_read_char(fi, &ch);
                     if(file_peek_char(fi, &ch) && (ch == '<' || ch == '>')) {
                        // Comment sequence only.
                        parsedCommandAppendArg(&c, ch);
                        file_read_char(fi, &ch);
                     } else {
                        // Must be a command sequence. Begin the nesting 
                        // levels!
                        parsedCommandPushState(&c, state);
                        state = TOKEN_STATE_CMD_NAME;
                     }
                  }
                  break;

               case ',':
                  // End of this argument, begin the next argument.
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_ARGUMENT;
                  break;
                  
               case ')':
                  // End of argument list.
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_END_ARGS;
                  break;
                  
               default:
                  // Normal character
                  parsedCommandAppendArg(&c, ch);
                  break;
                  
            }
            break;
            
         // TOKEN_STATE_IN_ARGV
         // We're processing an individual argument right now. This
         // is a variable name in the argument. Remember, of course, that
         // the characters disallowed in an argument are '(', ')', ',',
         // '$', '\"', '\'', ';'. ',' or ')' will signal the end of the
         // argument. Whitespace is allowed inside the argument but leading
         // and trailing whitespace is truncated. Look for a '#' to terminate
         // the variable name; we will also accept ',' or ')' to terminate
         // the variable name and the argument.
         //
         // -> TOKEN_STATE_IN_ARG     Normal character
         // -> TOKEN_STATE_IN_ARGV    Still the variable name...
         // -> TOKEN_STATE_ARGUMENT   ',' encountered
         // -> TOKEN_STATE_END_ARGS   ')' encountered
         // -> TOKEN_STATE_TEXT       Invalid character encountered
         case TOKEN_STATE_IN_ARGV:
            // This must always be a top level command
            switch(ch) {
                  
               case ',':
                  // We need to substitute the variable value now.
                  parsedCommandExpandVar(&c);
                  // End of this argument, begin the next argument.
                  parsedCommandAddArgument(&c);
                  state = TOKEN_STATE_ARGUMENT;
                  break;
                  
               case ')':
                  // We need to substitute the variable value now.
                  parsedCommandExpandVar(&c);
                  // End of argument list.
                  parsedCommandAddArgument(&c);
                  state = TOKEN_STATE_END_ARGS;
                  break;
                  
               case '#':
                  // End of vararg
                  parsedCommandExpandVar(&c);
                  state = TOKEN_STATE_IN_ARG;
                  break;
                  
               default:
                  // Normal character
                  parsedCommandAppendVar(&c, ch);
                  break;
                  
            }
            break;
            
         // TOKEN_STATE_IN_ARG_Q
         // In a quoted argument. Escape sequences are allowed in here,
         // and this state will continue until an unescaped quote character
         // is encountered. Wheeee!!! Won't this be fun? :)
         //
         // -> TOKEN_STATE_IN_ARG_Q    For normal char, ANSI escape, ...
         // -> TOKEN_STATE_IN_ARGV_Q   For variable arguments.
         // -> TOKEN_STATE_END_ARG_Q   Once the term " is found, but before ,)
         case TOKEN_STATE_IN_ARG_Q:
            if(!c.nestlevel) switch(ch) {
               // Top level argument.
               
               case '$':
                  // We might be looking at another escape sequence
                  parsedCommandAppendArg(&c, ch);
                  if(file_peek_char(fi, &ch) && ch == '$') {
                     // An escape sequence! Yay!
                     parsedCommandAppendArg(&c, ch);
                     file_read_char(fi, &ch);
                     if(file_peek_char(fi, &ch) && (ch == '<' || ch == '>')) {
                        // Comment sequence only.
                        parsedCommandAppendArg(&c, ch);
                        file_read_char(fi, &ch);
                     } else {
                        // Must be a command sequence. Begin the nesting 
                        // levels!
                        parsedCommandPushState(&c, state);
                        state = TOKEN_STATE_CMD_NAME;
                     }
                  }
                  break;

               case '\\':
                  // ANSI escape sequence beginning. Let's just process
                  // the entire sequence here, neh? :)
                  parsedCommandANSIEscape(&c);
                  break;

               case '#':
                  // Nonescaped hash. Begin variable name.
                  state = TOKEN_STATE_IN_ARGV_Q;
                  break;

               case '\"':
                  // Non-escaped quote. End of quote-string.
                  state = TOKEN_STATE_END_ARG_Q;
                  break;
                  
               default:
                  // Normal character.
                  parsedCommandAppendArg(&c, ch);
                  break;
                  
            } else switch(ch) {
               // Nested argument.
               
               case '$':
                  // We might be looking at another escape sequence
                  parsedCommandAppendArg(&c, ch);
                  if(file_peek_char(fi, &ch) && ch == '$') {
                     // An escape sequence! Yay!
                     parsedCommandAppendArg(&c, ch);
                     file_read_char(fi, &ch);
                     if(file_peek_char(fi, &ch) && (ch == '<' || ch == '>')) {
                        // Comment sequence only.
                        parsedCommandAppendArg(&c, ch);
                        file_read_char(fi, &ch);
                     } else {
                        // Must be a command sequence. Begin the nesting 
                        // levels!
                        parsedCommandPushState(&c, state);
                        state = TOKEN_STATE_CMD_NAME;
                     }
                  }
                  break;

               case '\"':
                  // Non-escaped quote. End of quote-string.
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_END_ARG_Q;
                  break;
                  
               default:
                  // Normal character.
                  parsedCommandAppendArg(&c, ch);
                  break;
                  
            }
            break;
            
         // TOKEN_STATE_IN_ARGV_Q
         // In a quoted argument. Escape sequences are allowed in here,
         // and this state will continue until an unescaped quote character
         // is encountered. Wheeee!!! Won't this be fun? :)
         //
         // -> TOKEN_STATE_IN_ARG_Q    After the variable is finished, ...
         // -> TOKEN_STATE_IN_ARGV_Q   Keep getting the varname.
         // -> TOKEN_STATE_END_ARG_Q   Once the term " is found, but before ,)
         case TOKEN_STATE_IN_ARGV_Q:
            // This state can only be reached at the top level command.
            
            switch(ch) {
               case '\\':
                  // ANSI escape sequence beginning. Let's just process
                  // the entire sequence here, neh? :)
                  parsedCommandANSIEscape(&c);
                  break;

               case '\"':
                  // We need to substitute the variable value now.
                  parsedCommandExpandVar(&c);
                  // Non-escaped quote. End of quote-string.
                  state = TOKEN_STATE_END_ARG_Q;
                  break;
                  
               case '#':
                  // We need to substitute the variable value now.
                  parsedCommandExpandVar(&c);
                  state = TOKEN_STATE_IN_ARG_Q;
                  break;
                  
               default:
                  // Normal character, for use in a variable name.
                  parsedCommandAppendVar(&c, ch);
                  break;
                  
            }
            break;
            
         // TOKEN_STATE_END_ARG_Q
         // End of a quoted argument. The only characters allowed here
         // are the ',' or the ')' characters, or whitespace.
         // 
         // -> TOKEN_STATE_ARGUMENT    ',' found
         // -> TOKEN_STATE_END_ARGS    ')' found
         // -> TOKEN_STATE_TEXT        error found
         case TOKEN_STATE_END_ARG_Q:
            if(!c.nestlevel) switch(ch) {
               // Top level command

               case ',':
                  // next argument ...
                  parsedCommandAddArgumentNoTrim(&c);
                  state = TOKEN_STATE_ARGUMENT;
                  break;
                  
               case ')':
                  // End of arguments
                  parsedCommandAddArgumentNoTrim(&c);
                  state = TOKEN_STATE_END_ARGS;
                  break;
                  
               case ' ':
               case '\t':
               case '\n':
                  // Whitespace
                  break;
                  
               default:
                  // Erroneous symbols. Return to text processing after
                  // indicating an error has occurred.
                  parsedCommandAppendArg(&c, ch);
                  snprintf(get_erstr(), 32768, "\"%s\": Invalid symbol ending command argument \"%s\"", c.cmdname, c.cmdarg);
                  *get_ercode() = TOKEN_ERROR_INVALID_SYMBOL;
                  printFileError(t);
                  if(*get_ercode()) {
                     freeParsedCommand(&c);
                     return(*get_ercode());
                  }
                  state = TOKEN_STATE_TEXT;
                  break;
                  
            } else switch(ch) {
               // Nested command

               case ',':
                  // next argument ...
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_ARGUMENT;
                  break;
                  
               case ')':
                  // End of arguments
                  parsedCommandAppendArg(&c, ch);
                  state = TOKEN_STATE_END_ARGS;
                  break;
                  
               default:
                  parsedCommandAppendArg(&c, ch);
                  break;
                  
            }
            break;
            
         // TOKEN_STATE_END_ARGS
         // Ignores whitespace. Once a ';' is encountered, the command
         // substitution will be executed.
         case TOKEN_STATE_END_ARGS:
            if(!c.nestlevel) switch(ch) {
               // Top level command

               case ';':
                  // Command substitution here.
                  if((*get_ercode() = parseCommand(t, fi, fo, c.cmdname, c.cmdargs))) {
                     freeParsedCommand(&c);
                     return(*get_ercode());
                  }
                  state = TOKEN_STATE_TEXT;
                  break;
                 
               case ' ':
               case '\t':
               case '\n':
                  // Whitespace ignored.
                  break;
                  
               default:
                  // Erroneous symbols. Return to text processing after
                  // indicating an error has occurred.
                  snprintf(get_erstr(), 32768, "\"%s\": Invalid symbol '%c' after argument list and before terminating semicolon", c.cmdname, ch);
                  *get_ercode() = TOKEN_ERROR_INVALID_SYMBOL;
                  printFileError(t);
                  if(*get_ercode()) {
                     freeParsedCommand(&c);
                     return(*get_ercode());
                  }
                  state = TOKEN_STATE_TEXT;
                  break;
                  
            } else switch(ch) {
               // Nested command

               case ';':
                  // Command substitution here.
                  parsedCommandAppendArg(&c, ch);
                  state = parsedCommandPopState(&c);
                  break;
                 
               default:
                  parsedCommandAppendArg(&c, ch);
                  break;
                  
            }
            break;
            
         // TOKEN_STATE_COMMENT
         // Normal, flowing text which will NOT be echoed verbatim to output
         // file. Will search for the first character of the "$$" escape 
         // sequence which initiates a command or (hopefully) the end of the
         // comment block.
         //
         // -> TOKEN_STATE_TEXT           Return to echoed text.
         // -> TOKEN_STATE_COMMENT        Continue the comment block sequence
         case TOKEN_STATE_COMMENT:
            // This can only be encountered at the top level
            
            if(ch == '$') {
               // Possible command escape sequence?
               if(file_peek_char(fi, &ch) && ch == '$') {
                  // Escape sequence! Yay! 
                  file_read_char(fi, &ch);
                  // Check to see if this is a comment escape sequence
                  if(file_peek_char(fi, &ch) && ch == '<') {
                     // Beginning another comment block! argh...
                     file_read_char(fi, &ch);
                     commlevel++;
                  } else if(file_peek_char(fi, &ch) && ch == '>') {
                     // Closing a comment block.
                     file_read_char(fi, &ch);
                     commlevel--;
                     if(commlevel <= 0) state = TOKEN_STATE_TEXT;
                  }
               }
            }
            break;
            
      }
   }
   
   // If we're not on a terminus state then we screwed up.
   if(state != TOKEN_STATE_TEXT && state != TOKEN_STATE_COMMENT) {
      snprintf(get_erstr(), 32768, "Unexpected end of file (state %d)", state);
      *get_ercode() = TOKEN_ERROR_EOF;
      printFileError(t);
      if(*get_ercode()) {
         freeParsedCommand(&c);
         return(*get_ercode());
      }
   } 

   #ifdef LIBTOKEN_DEBUG_INCLUDE
   printf("Wrote %s(%d lines) -> %s ...\n", fName(fi), fLine(fi), fName(fo));
   #endif // debug
   
   // Release the argument linked list, if needed
   freeParsedCommand(&c);
   file_flush(fo);
   slink_free(&t->nestedfiles, false);
   return(TOKEN_SUCCESS);
   
}
