/* LIBTOKEN                                                 Version 0.1.2
   Copyright (C) 1998   Justin Smith             justins@ugcs.caltech.edu
   ----------------------------------------------------------------------

   This  program  is free software; you can redistribute it and/or modify
   it  under  the terms of the GNU General Public License as published by       
   the  Free  Software  Foundation;  either  version 2 of the License, or
   (at your option) any later version.

   This  program  is  distributed  in  the  hope  that it will be useful,
   but  WITHOUT  ANY  WARRANTY;  without  even  the  implied  warranty of
   MERCHANTABILITY   or  FITNESS  FOR  A  PARTICULAR  PURPOSE.   See  the
   GNU General Public License for more details.       

   You  should  have  received  a  copy of the GNU General Public License       
   along  with  this  program,  in the file COPYING. If not, write to the
   Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA

   ----------------------------------------------------------------------
*/


#define __LIBTOKEN_SOURCE
#include <libjtoken.h>
#include <libtoken_int.h>
#include <libjlink.h>
#include <libjmem.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef NO_MATH_LIB
#include <math.h>
#endif  // math lib?


/*** Support Functions ***/


int printFileError(const tokeniser *t) {
/* Print-File-Error
   This function prints out a simple error message and indicates where
   the error occurred. */

   #ifndef LIBTOKEN_NO_OUTPUT
   slinknode *cur = t->nestedfiles;
   if(cur) {
      fprintf(stderr, "In file %s(%d)", file_name((file *)cur->data), file_line((file *)cur->data));
      cur = cur->next;
      if(cur) fprintf(stderr, ",\n");
      while(cur) {
         fprintf(stderr, "Included from %s(%d)", file_name((file *)cur->data), file_line((file *)cur->data));
         cur = cur->next;
         if(cur) fprintf(stderr, ",\n");
      }
   } else fprintf(stderr, "(no file)");         
   fprintf(stderr, ": Error %d: %s\n\n", *get_ercode(), get_erstr());
   #endif
   if(t->erfunc && (*get_ercode() = t->erfunc(t->nestedfiles ? (file *)t->nestedfiles->data : NULL, *get_ercode(), get_erstr()))) return(*get_ercode());
   return(*get_ercode() = TOKEN_SUCCESS);

}


tokencmd *cmdExists(tokeniser *t, const char *cmdname) {
/* Command-Exists 
   Returns a pointer to the matching command token if the command exists,
   and NULL otherwise. */

   slinknode *cur = t->commands;
   while(cur) {
      if(!(scmp(((tokencmd *)cur->data)->cmdname, cmdname) ^ 0xffffffff)) return((tokencmd *)cur->data);
      cur = cur->next;
   }
   return(NULL);

}


bool cmdIgnoreIfBlocks(const tokencmd *c) {
/* Command-Ignores-If-Blocks 
   Returns true if the specified command is executed even if it would
   normally not be executed because of an if block. In particular, 
   if/elif/else/endif commands must ignore if blocks to properly
   determine the nesting level. Commands which ignore if blocks should
   take extreme care to make sure they do not have any effects when
   they are not supposed to by checking the ifExec() function. False is
   returned if the command does not exist. */

   if(c) return(c->flags & TOKEN_CMD_IGNORE_IF_BLOCKS);
   else return(false);

}


bool cmdDoNotExecArgs(const tokencmd *c) {
/* Command-Do-Not-Exec-Args
   Returns true if the command specified has requested its arguments
   NOT be executed before they are passed to the command. The command
   is free to do its own argument execution if it wishes. Returns false
   if the command wishes its arguments to be automatically executed,
   or if the command does not exist. */

   if(c) return(c->flags & TOKEN_CMD_DO_NOT_EXEC_ARGS);
   else return(false);

}


bool cmdIsCustom(const tokencmd *c) {
/* Command-Is-Custom 
   Returns true if the command specified is a custom command defined
   by a $$define() function. Internal commands and commands which are
   compiled in should never set this flag, as setting this flag allows
   a $$define() call to override the command. */

   if(c) return(c->flags & TOKEN_CMD_IS_CUSTOM);
   else return(false);

}


bool ifExec(const tokeniser *t) {
/* If-Executable 
   Returns true if we are currently allowed to execute commands and echo
   normal text. We are not allowed to execute normal commands inside an
   if block unless the condition evaluates as true. */

   if(t->iff) return(((tokeniff *)t->iff->data)->exec);
   else return(true);

}


bool ifCmdExec(const tokeniser *t, const tokencmd *c) {
/* If-Command-Executable 
   Returns true if the specified command is allowed to execute. This
   function will check and see if the command is allowed to ignore
   if blocks. */

   if(t->iff) {
      if(((tokeniff *)t->iff->data)->exec) return(true);
      if(cmdIgnoreIfBlocks(c)) return(true);
      return(false);
   } else return(true);

}


tokenvar *varExists(tokeniser *t, const char *varname) {
/* Variable-Exists 
   Returns a pointer to the matching variable token if the command exists,
   and NULL otherwise. */

   slinknode *cur = t->variables;
   while(cur) {
      if(!(scmp(((tokenvar *)cur->data)->name, varname) ^ 0xffffffff)) return((tokenvar *)cur->data);
      cur = cur->next;
   }
   return(NULL);

}


bool intVar(tokenvar *v) {
/* Variable-Is-Internal */

   return(v->flags & TOKEN_VAR_INTERNAL);

}


void tokenCommandValidate(char *d, const char *s) {
/* Token-Command-Validate */

   char *ds = d;
   while(*s && (ds - d) < TOKEN_VAR_LENGTH) {
      if(*s >= 'a' && *s <= 'z') *(d++) = *s;
      if(*s >= 'A' && *s <= 'Z') *(d++) = *s;
      if(*s >= '0' && *s <= '9') *(d++) = *s;
      if(*s == '_') *(d++) = *s;
      if(*s == '.') *(d++) = *s;
      if(*s == '/') *(d++) = *s;
      s++;
   }
   *d = 0;

}


bool tokenDestroyCommand(void *data) {
/* Token-Destroy-Command */

   tokencmd *c = (tokencmd *)data;
   if(!c) return(true);
   free(c->cmdname);
   free(c);
   return(true);
   
}


void tokenVariableValidate(char *d, const char *s) {
/* Token-Variable-Validate */

   char *ds = d;
   while(*s && (ds - d) < TOKEN_VAR_LENGTH) {
      if(*s >= 'a' && *s <= 'z') *(d++) = *s;
      if(*s >= 'A' && *s <= 'Z') *(d++) = *s;
      if(*s >= '0' && *s <= '9') *(d++) = *s;
      if(*s == '_') *(d++) = *s;
      if(*s == '.') *(d++) = *s;
      s++;
   }
   *d = 0;

}


bool tokenDestroyVariable(void *data) {
/* Token-Destroy-Variable */

   tokenvar *v = (tokenvar *)data;
   if(!v) return(true);
   free(v->name);
   free(v->value);
   free(v);
   return(true);
   
}


void tokenSaveState(tokensavedstate *ss, tokeniser *t) {
/* Token-Save-State
   Saves the token state and resets certain token defaults */
   
   ss->text = t->text;
   ss->newline = t->newline;
   ss->squash = t->squash;
   ss->sawspace = 0;
   ss->iff = t->iff;
   t->text = 0;
   t->newline = 0;
   t->squash = 0;
   t->sawspace = 0;
   t->iff = NULL;
   
}


void tokenRestoreState(tokeniser *t, tokensavedstate *ss) {
/* Token-Restore-State */
   
   slink_release(&t->iff);
   t->text = ss->text;
   t->newline = ss->newline;
   t->squash = ss->squash;
   t->sawspace = ss->sawspace;
   t->iff = ss->iff;
   
}


#ifndef NO_MATH_LIB
void mathWriteNumber(file *fo, double n) {
/* Math-Write-Number */

   unsigned int mag = 1;      // What magnitude are we on right now?
   long num = (long)n;

   // Handling for negative numbers
   if (num < 0) {
      file_write_char(fo, '-');
      num = -num;
   }

   // Print all digits
   while (num / (mag * 10) > 0) mag *= 10;

   // Start copying off the digits :)
   while (mag) {
      file_write_char(fo, (num / mag) % 10 + '0');
      num %= mag;
      mag /= 10;
   }

}
#endif // math lib?
